 /*
 * Copyright (c) 2001-2003, David Janssens
 * Copyright (c) 2002-2003, Yannick Verschueren
 * Copyright (c) 2003-2005, Francois Devaux and Antonin Descampe
 * Copyright (c) 2005, Herv� Drolon, FreeImage Team
 * Copyright (c) 2002-2005, Communications and remote sensing Laboratory, Universite catholique de Louvain, Belgium
 * Copyright (c) 2005-2006, Dept. of Electronic and Information Engineering, Universita' degli Studi di Perugia, Italy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
/**
 * Reed-Solomon coding and decoding
 * Phil Karn (karn@ka9q.ampr.org) September 1996
 * 
 * This file is derived from the program "new_rs_erasures.c" by Robert
 * Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari Thirumoorthy
 * (harit@spectra.eng.hawaii.edu), Aug 1995
 *
 * I've made changes to improve performance, clean up the code and make it
 * easier to follow. Data is now passed to the encoding and decoding functions
 * through arguments rather than in global arrays. The decode function returns
 * the number of corrected symbols, or -1 if the word is uncorrectable.
 *
 * This code supports a symbol size from 2 bits up to 16 bits,
 * implying a block size of 3 2-bit symbols (6 bits) up to 65535
 * 16-bit symbols (1,048,560 bits). The code parameters are set in rs.h.
 *
 * Note that if symbols larger than 8 bits are used, the type of each
 * data array element switches from char to unsigned int. The
 * caller must ensure that elements larger than the symbol range are
 * not passed to the encoder or decoder.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "p25/ecc/rs.h"

//#define r_MM	6
//#define r_NN	63
int r_KK;

/* 1 + x + x^6 */
int r_pp[r_MM+1] = { 1, 1, 0, 0, 0, 0, 1 };

int r_alpha_to[r_NN+1];
int r_index_of[r_NN+1];

int A0 = 0;
int B0 = 1;

/* Generator polynomial g(x)
 * Degree of g(x) = 2*TT
 * has roots @**B0, @**(B0+1), ... ,@^(B0+2*TT-1)
 */
/*int r_gg[r_NN - r_KK + 1];*/
int *r_gg;
int gg_36_20[r_NN - 1];
int gg_24_16[r_NN - 1];
int gg_24_12[r_NN - 1];

/* Compute x % r_NN, where r_NN is 2**r_MM - 1,
 * without a slow divide
 */
static /*inline*/ int
modnn(int x)
{
	while (x >= r_NN) {
		x -= r_NN;
		x = (x >> r_MM) + (x & r_NN);
	}
	return x;
}

/*#define	min(a,b)	((a) < (b) ? (a) : (b))*/

#define	CLEAR(a,n) {\
	int ci;\
	for(ci=(n)-1;ci >=0;ci--)\
		(a)[ci] = 0;\
	}

#define	COPY(a,b,n) {\
	int ci;\
	for(ci=(n)-1;ci >=0;ci--)\
		(a)[ci] = (b)[ci];\
	}
#define	COPYDOWN(a,b,n) {\
	int ci;\
	for(ci=(n)-1;ci >=0;ci--)\
		(a)[ci] = (b)[ci];\
	}

void init_rs()
{
	generate_gf_rs();
    
    r_gg = gg_36_20; r_KK = 47; gen_poly_rs();
    r_gg = gg_24_16; r_KK = 55; gen_poly_rs();
    r_gg = gg_24_12; r_KK = 51; gen_poly_rs();
}

/* generate GF(2**m) from the irreducible polynomial p(X) in p[0]..p[m]
   lookup tables:  index->polynomial form   r_alpha_to[] contains j=alpha**i;
                   polynomial form -> index form  r_index_of[j=alpha**i] = i
   alpha=2 is the primitive element of GF(2**m)
   HARI's COMMENT: (4/13/94) r_alpha_to[] can be used as follows:
        Let @ represent the primitive element commonly called "alpha" that
   is the root of the primitive polynomial p(x). Then in GF(2^m), for any
   0 <= i <= 2^m-2,
        @^i = a(0) + a(1) @ + a(2) @^2 + ... + a(m-1) @^(m-1)
   where the binary vector (a(0),a(1),a(2),...,a(m-1)) is the representation
   of the integer "r_alpha_to[i]" with a(0) being the LSB and a(m-1) the MSB. Thus for
   example the polynomial representation of @^5 would be given by the binary
   representation of the integer "r_alpha_to[5]".
                   Similarily, r_index_of[] can be used as follows:
        As above, let @ represent the primitive element of GF(2^m) that is
   the root of the primitive polynomial p(x). In order to find the power
   of @ (alpha) that has the polynomial representation
        a(0) + a(1) @ + a(2) @^2 + ... + a(m-1) @^(m-1)
   we consider the integer "i" whose binary representation with a(0) being LSB
   and a(m-1) MSB is (a(0),a(1),...,a(m-1)) and locate the entry
   "r_index_of[i]". Now, @^r_index_of[i] is that element whose polynomial 
    representation is (a(0),a(1),a(2),...,a(m-1)).
   NOTE:
        The element r_alpha_to[2^m-1] = 0 always signifying that the
   representation of "@^infinity" = 0 is (0,0,0,...,0).
        Similarily, the element r_index_of[0] = A0 always signifying
   that the power of alpha which has the polynomial representation
   (0,0,...,0) is "infinity".
 
*/

void generate_gf_rs(void)
{
	register int i, mask;

	mask = 1;
	r_alpha_to[r_MM] = 0;
	for (i = 0; i < r_MM; i++) {
		r_alpha_to[i] = mask;
		r_index_of[r_alpha_to[i]] = i;
		/* If r_pp[i] == 1 then, term @^i occurs in poly-repr of @^r_MM */
		if (r_pp[i] != 0)
			r_alpha_to[r_MM] ^= mask;	/* Bit-wise EXOR operation */
		mask <<= 1;	/* single left-shift */
	}
	r_index_of[r_alpha_to[r_MM]] = r_MM;
	/*
	 * Have obtained poly-repr of @^r_MM. Poly-repr of @^(i+1) is given by
	 * poly-repr of @^i shifted left one-bit and accounting for any @^r_MM
	 * term that may occur when poly-repr of @^i is shifted.
	 */
	mask >>= 1;
	for (i = r_MM + 1; i < r_NN; i++) {
		if (r_alpha_to[i - 1] >= mask)
			r_alpha_to[i] = r_alpha_to[r_MM] ^ ((r_alpha_to[i - 1] ^ mask) << 1);
		else
			r_alpha_to[i] = r_alpha_to[i - 1] << 1;
		r_index_of[r_alpha_to[i]] = i;
	}
	r_index_of[0] = -1;
	r_alpha_to[r_NN] = 0;
}


/*
 * Obtain the generator polynomial of the TT-error correcting, length
 * r_NN=(2**r_MM -1) Reed Solomon code from the product of (X+@**(B0+i)), i = 0,
 * ... ,(2*TT-1)
 *
 * Examples:
 *
 * If B0 = 1, TT = 1. deg(g(x)) = 2*TT = 2.
 * g(x) = (x+@) (x+@**2)
 *
 * If B0 = 0, TT = 2. deg(g(x)) = 2*TT = 4.
 * g(x) = (x+1) (x+@) (x+@**2) (x+@**3)
 */
void gen_poly_rs()
{
	register int i, j;

	r_gg[0] = r_alpha_to[B0];
	r_gg[1] = 1;		/* g(x) = (X+@**B0) initially */
	for (i = 2; i <= r_NN - r_KK; i++) {
		r_gg[i] = 1;
		/*
		 * Below multiply (r_gg[0]+r_gg[1]*x + ... +r_gg[i]x^i) by
		 * (@**(B0+i-1) + x)
		 */
		for (j = i - 1; j > 0; j--)
			if (r_gg[j] != 0)
				r_gg[j] = r_gg[j - 1] ^ r_alpha_to[modnn((r_index_of[r_gg[j]]) + B0 + i - 1)];
			else
				r_gg[j] = r_gg[j - 1];
		/* r_gg[0] can never be zero */
		r_gg[0] = r_alpha_to[modnn((r_index_of[r_gg[0]]) + B0 + i - 1)];
	}
	/* convert r_gg[] to index form for quicker encoding */
	for (i = 0; i <= r_NN - r_KK; i++) {
//		printf("r_gg[%d] = %d\n", i, r_gg[i]);
		r_gg[i] = r_index_of[r_gg[i]];
	}
}


/*
 * take the string of symbols in data[i], i=0..(k-1) and encode
 * systematically to produce r_NN-r_KK parity symbols in bb[0]..bb[r_NN-r_KK-1] data[]
 * is input and bb[] is output in polynomial form. Encoding is done by using
 * a feedback shift register with appropriate connections specified by the
 * elements of r_gg[], which was generated above. Codeword is   c(X) =
 * data(X)*X**(r_NN-r_KK)+ b(X)
 */
int encode_rs(r_dtype *data, r_dtype *enc_data)
{
	register int i, j;
	int feedback;
    r_dtype *bb;
	
	bb = (r_dtype *) malloc(r_NN-r_KK);

	CLEAR(bb,r_NN-r_KK);
	for (i = r_KK - 1; i >= 0; i--) {
#if (r_MM != 8)
		if(data[i] > r_NN) {
			free(bb);
			return -1;	/* Illegal symbol */
		}
#endif
		feedback = r_index_of[data[i] ^ bb[r_NN - r_KK - 1]];
		if (feedback != -1) {	/* feedback term is non-zero */
			for (j = r_NN - r_KK - 1; j > 0; j--)
				if (r_gg[j] != -1)
					bb[j] = bb[j - 1] ^ r_alpha_to[modnn(r_gg[j] + feedback)];
				else
					bb[j] = bb[j - 1];
			bb[0] = r_alpha_to[modnn(r_gg[0] + feedback)];
		} else {	/* feedback term is zero. encoder becomes a
				 * single-byte shifter */
			for (j = r_NN - r_KK - 1; j > 0; j--)
				bb[j] = bb[j - 1];
			bb[0] = 0;
		}
	}

	for (i = 0; i < r_NN-r_KK; i++) {
		enc_data[i] = bb[i];
    }

	free(bb);
	return 0;
}

int encode_rs_36_20(r_dtype *data, r_dtype *enc)
{
    int res,N,K;
    r_dtype d[47];
    r_dtype cw[r_NN];

    r_gg = gg_36_20;
    r_KK = 47;
    N = 36;
    K = 20;

    memset(d,0,r_KK);
    memcpy(d,data,K);

    res = encode_rs(d,cw);
    if(!res) {
        memcpy(enc,cw,N-K);
        memcpy(enc+(N-K),data,K);
    }

    return res;
}

int encode_rs_24_16(r_dtype *data, r_dtype *enc)
{
    int res,N,K;
    r_dtype d[55];
    r_dtype cw[r_NN];

    r_gg = gg_24_16;
    r_KK = 55;
    N = 24;
    K = 16;

    memset(d,0,r_KK);
    memcpy(d,data,K);

    res = encode_rs(d,cw);
    if(!res) {
        memcpy(enc,cw,N-K);
        memcpy(enc+(N-K),data,K);
    }

    return res;
}

int encode_rs_24_12(r_dtype *data, r_dtype *enc)
{
    int res,N,K;
    r_dtype d[51];
    r_dtype cw[r_NN];

    r_gg = gg_24_12;
    r_KK = 51;
    N = 24;
    K = 12;

    memset(d,0,r_KK);
    memcpy(d,data,K);

    res = encode_rs(d,cw);
    if(!res) {
        memcpy(enc,cw,N-K);
        memcpy(enc+(N-K),data,K);
    }

    return res;
}


/*
 * Performs ERRORS+ERASURES decoding of RS codes. If decoding is successful,
 * writes the codeword into data[] itself. Otherwise data[] is unaltered.
 *
 * Return number of symbols corrected, or -1 if codeword is illegal
 * or uncorrectable.
 * 
 * First "no_eras" erasures are declared by the calling program. Then, the
 * maximum # of errors correctable is t_after_eras = floor((r_NN-r_KK-no_eras)/2).
 * If the number of channel errors is not greater than "t_after_eras" the
 * transmitted codeword will be recovered. Details of algorithm can be found
 * in R. Blahut's "Theory ... of Error-Correcting Codes".
 */
int eras_dec_rs(r_dtype *data, int *eras_pos, int no_eras)
{
	int deg_lambda, el, deg_omega;
	int i, j, r;
	int u,q,tmp,num1,num2,den,discr_r;
	int recd[r_NN];
	/* Err+Eras Locator poly and syndrome poly */
	/*int lambda[r_NN-r_KK + 1], s[r_NN-r_KK + 1];
	int b[r_NN-r_KK + 1], t[r_NN-r_KK + 1], omega[r_NN-r_KK + 1];
	int root[r_NN-r_KK], reg[r_NN-r_KK + 1], loc[r_NN-r_KK];*/
	int lambda[r_NN + 1], s[r_NN + 1];
	int b[r_NN + 1], t[r_NN + 1], omega[r_NN + 1];
	int root[r_NN], reg[r_NN + 1], loc[r_NN];
	int syn_error, count;

	/* data[] is in polynomial form, copy and convert to index form */
	for (i = r_NN-1; i >= 0; i--){
#if (r_MM != 8)
		if(data[i] > r_NN)
			return -1;	/* Illegal symbol */
#endif
		recd[i] = r_index_of[data[i]];
	}
	/* first form the syndromes; i.e., evaluate recd(x) at roots of g(x)
	 * namely @**(B0+i), i = 0, ... ,(r_NN-r_KK-1)
	 */
	syn_error = 0;
	for (i = 1; i <= r_NN-r_KK; i++) {
		tmp = 0;
		for (j = 0; j < r_NN; j++)
			if (recd[j] != A0)	/* recd[j] in index form */
				tmp ^= r_alpha_to[modnn(recd[j] + (B0+i-1)*j)];
		syn_error |= tmp;	/* set flag if non-zero syndrome =>
					 * error */
		/* store syndrome in index form  */
		s[i] = r_index_of[tmp];
	}
	if (!syn_error) {
		/*
		 * if syndrome is zero, data[] is a codeword and there are no
		 * errors to correct. So return data[] unmodified
		 */
		return 0;
	}
	CLEAR(&lambda[1],r_NN-r_KK);
	lambda[0] = 1;
	if (no_eras > 0) {
		/* Init lambda to be the erasure locator polynomial */
		lambda[1] = r_alpha_to[eras_pos[0]];
		for (i = 1; i < no_eras; i++) {
			u = eras_pos[i];
			for (j = i+1; j > 0; j--) {
				/* JC - Testing for out of bound indexes */
				if (lambda[j-1] > sizeof(r_index_of)) {
					printf("Index out of bounds:  lambda[j-1] = %d, sizeof(r_index_of) = %d\n", 
						lambda[j-1], sizeof(r_index_of));
					return -1;
				}

				tmp = r_index_of[lambda[j - 1]];
				if(tmp != A0)
					lambda[j] ^= r_alpha_to[modnn(u + tmp)];
			}
		}
#ifdef ERASURE_DEBUG
		/* find roots of the erasure location polynomial */
		for(i=1;i<=no_eras;i++)
			reg[i] = r_index_of[lambda[i]];
		count = 0;
		for (i = 1; i <= r_NN; i++) {
			q = 1;
			for (j = 1; j <= no_eras; j++)
				if (reg[j] != A0) {
					reg[j] = modnn(reg[j] + j);
					q ^= r_alpha_to[reg[j]];
				}
			if (!q) {
				/* store root and error location
				 * number indices
				 */
				root[count] = i;
				loc[count] = r_NN - i;
				count++;
			}
		}
		if (count != no_eras) {
			printf("\n lambda(x) is WRONG\n");
			return -1;
		}
#ifndef NO_PRINT
		printf("\n Erasure positions as determined by roots of Eras Loc Poly:\n");
		for (i = 0; i < count; i++)
			printf("%d ", loc[i]);
		printf("\n");
#endif
#endif
	}
	for(i=0;i<r_NN-r_KK+1;i++)
		b[i] = r_index_of[lambda[i]];

	/*
	 * Begin Berlekamp-Massey algorithm to determine error+erasure
	 * locator polynomial
	 */
	r = no_eras;
	el = no_eras;
	while (++r <= r_NN-r_KK) {	/* r is the step number */
		/* Compute discrepancy at the r-th step in poly-form */
		discr_r = 0;
		for (i = 0; i < r; i++){
			if ((lambda[i] != 0) && (s[r - i] != A0)) {
				discr_r ^= r_alpha_to[modnn(r_index_of[lambda[i]] + s[r - i])];
			}
		}
		discr_r = r_index_of[discr_r];	/* Index form */
		if (discr_r == A0) {
			/* 2 lines below: B(x) <-- x*B(x) */
			COPYDOWN(&b[1],b,r_NN-r_KK);
			b[0] = A0;
		} else {
			/* 7 lines below: T(x) <-- lambda(x) - discr_r*x*b(x) */
			t[0] = lambda[0];
			for (i = 0 ; i < r_NN-r_KK; i++) {
				if(b[i] != A0)
					t[i+1] = lambda[i+1] ^ r_alpha_to[modnn(discr_r + b[i])];
				else
					t[i+1] = lambda[i+1];
			}
			if (2 * el <= r + no_eras - 1) {
				el = r + no_eras - el;
				/*
				 * 2 lines below: B(x) <-- inv(discr_r) *
				 * lambda(x)
				 */
				for (i = 0; i <= r_NN-r_KK; i++)
					b[i] = (lambda[i] == 0) ? A0 : modnn(r_index_of[lambda[i]] - discr_r + r_NN);
			} else {
				/* 2 lines below: B(x) <-- x*B(x) */
				COPYDOWN(&b[1],b,r_NN-r_KK);
				b[0] = A0;
			}
			COPY(lambda,t,r_NN-r_KK+1);
		}
	}

	/* Convert lambda to index form and compute deg(lambda(x)) */
	deg_lambda = 0;
	for(i=0;i<r_NN-r_KK+1;i++){
		lambda[i] = r_index_of[lambda[i]];
		if(lambda[i] != A0)
			deg_lambda = i;
	}
	/*
	 * Find roots of the error+erasure locator polynomial. By Chien
	 * Search
	 */
	COPY(&reg[1],&lambda[1],r_NN-r_KK);
	count = 0;		/* Number of roots of lambda(x) */
	for (i = 1; i <= r_NN; i++) {
		q = 1;
		for (j = deg_lambda; j > 0; j--)
			if (reg[j] != A0) {
				reg[j] = modnn(reg[j] + j);
				q ^= r_alpha_to[reg[j]];
			}
		if (!q) {
			/* store root (index-form) and error location number */
			root[count] = i;
			loc[count] = r_NN - i;
			count++;
		}
	}

#ifdef DEBUG
	printf("\n Final error positions:\t");
	for (i = 0; i < count; i++)
		printf("%d ", loc[i]);
	printf("\n");
#endif
	if (deg_lambda != count) {
		/*
		 * deg(lambda) unequal to number of roots => uncorrectable
		 * error detected
		 */
		return -1;
	}
	/*
	 * Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) (modulo
	 * x**(r_NN-r_KK)). in index form. Also find deg(omega).
	 */
	deg_omega = 0;
	for (i = 0; i < r_NN-r_KK;i++){
		tmp = 0;
		j = (deg_lambda < i) ? deg_lambda : i;
		for(;j >= 0; j--){
			if ((s[i + 1 - j] != A0) && (lambda[j] != A0))
				tmp ^= r_alpha_to[modnn(s[i + 1 - j] + lambda[j])];
		}
		if(tmp != 0)
			deg_omega = i;
		omega[i] = r_index_of[tmp];
	}
	omega[r_NN-r_KK] = A0;

	/*
	 * Compute error values in poly-form. num1 = omega(inv(X(l))), num2 =
	 * inv(X(l))**(B0-1) and den = lambda_pr(inv(X(l))) all in poly-form
	 */
	for (j = count-1; j >=0; j--) {
		num1 = 0;
		for (i = deg_omega; i >= 0; i--) {
			if (omega[i] != A0)
				num1  ^= r_alpha_to[modnn(omega[i] + i * root[j])];
		}
		num2 = r_alpha_to[modnn(root[j] * (B0 - 1) + r_NN)];
		den = 0;

		/* lambda[i+1] for i even is the formal derivative lambda_pr of lambda[i] */
		for (i = min(deg_lambda,r_NN-r_KK-1) & ~1; i >= 0; i -=2) {
			if(lambda[i+1] != A0)
				den ^= r_alpha_to[modnn(lambda[i+1] + i * root[j])];
		}
		if (den == 0) {
#ifdef DEBUG
			printf("\n ERROR: denominator = 0\n");
#endif
			return -1;
		}
		/* Apply error to data */
		if (num1 != 0) {
			data[loc[j]] ^= r_alpha_to[modnn(r_index_of[num1] + r_index_of[num2] + r_NN - r_index_of[den])];
		}
	}
	return count;
}

int decode_rs_36_20(r_dtype *data, int *eras_pos, int no_eras)
{
    int res,N,K;
    r_dtype cw[r_NN];

    r_gg = gg_36_20;
    r_KK = 47;
    N = 36;
    K = 20;

    memset(cw,0,r_NN);
    memcpy(cw,data,N);

    res = eras_dec_rs(cw,eras_pos,no_eras);
    if(res >= 0) {
        memcpy(data,cw,K);
    }

    return res;
}

int decode_rs_24_16(r_dtype *data, int *eras_pos, int no_eras)
{
    int res,N,K;
    r_dtype cw[r_NN];

    r_gg = gg_24_16;
    r_KK = 55;
    N = 24;
    K = 16;

    memset(cw,0,r_NN);
    memcpy(cw,data,N);

    res = eras_dec_rs(cw,eras_pos,no_eras);
    if(res >= 0) {
        memcpy(data,cw,K);
    }

    return res;
}

int decode_rs_24_12(r_dtype *data, int *eras_pos, int no_eras)
{
    int res,N,K;
    r_dtype cw[r_NN];

    r_gg = gg_24_12;
    r_KK = 51;
    N = 24;
    K = 12;

    memset(cw,0,r_NN);
    memcpy(cw,data,N);

    res = eras_dec_rs(cw,eras_pos,no_eras);
    if(res >= 0) {
        memcpy(data,cw,K);
    }

    return res;
}

