///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	bit_fifo.c
 * @author	Cal(it)2
 * @author	Anthony Nwokafor
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2011, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "p25/format/bit_fifo.h"
#include "p25/util/logger.h"


/** 
 * @details
 *
 * @param size expected size in bits of the fifo, the allocated
 * 		 fifo size might be larger
 *
 * @return Pointer to a bit_fifo structure, or NULL if failed
 */		 
struct bit_fifo *bf_alloc(int size)
{
	int nbytes;
	struct bit_fifo *bf;

	if (size == 0)
		size = BF_SIZE_DFT;
	if (size > BF_SIZE_MAX)
		size = BF_SIZE_MAX;

	nbytes = size/8 + 1;
	bf = (struct bit_fifo *) malloc(sizeof(struct bit_fifo) + nbytes);
	if (bf == NULL) {
		ERROR("alloc bit fifo failed, size = %d\n",size);
		return NULL;
	}

	bf->byte_size = nbytes;
	bf->head_index = 0;
	bf->head_offset = 0;
	bf->tail_index = bf->byte_size - 1;
	bf->tail_offset = 7;
	bf->bytes = (char *) (bf + 1);

	return bf;
}

/** 
 * @details
 *
 * @param bf The bit fifo to be freed
 */		 
void bf_free(struct bit_fifo *bf)
{
	free(bf);
}

/** 
 * @details
 *
 * @param bf The bit fifo to be reset
 */		 
void bf_reset(struct bit_fifo *bf)
{
	bf->head_index = 0;
	bf->head_offset = 0;
	bf->tail_index = bf->byte_size - 1;
	bf->tail_offset = 7;
}

#define incr_index_offset(idx, off, nbytes)	do {			\
	(off) = ((off) + 1) % 8;								\
	(idx) = ((off) == 0) ? ((idx)+1) % (nbytes) : (idx);	\
} while (0)

/**
 * @details
 *
 * @return Return 1 if bit fifo is empty, otherwise
 * 			  return 0
 */
int bf_empty(struct bit_fifo *bf)
{
	int index, offset;

	index = bf->tail_index;
	offset = bf->tail_offset;

	incr_index_offset(index, offset, bf->byte_size);
	return index == bf->head_index && offset == bf->head_offset;
}

/**
 * @details
 *
 * @return Return 1 if bit fifo is full, otherwise 
 * 			 return 0
 */
int bf_full(struct bit_fifo *bf)
{
	return bf->tail_index == bf->head_index &&
		bf->tail_offset == bf->head_offset;
}

/**
 * @details 
 *
 * @return Return the number of bits in a bit fifo
 */
int bf_size(struct bit_fifo *bf)
{
	int size;
	size = (bf->head_index - bf->tail_index) * 8 +
		bf->head_offset - bf->tail_offset - 1;
	return (size >= 0) ? size : size + bf->byte_size * 8; 
}

/**
 * @details
 *
 * @return Return the number of bits can be put in the bit fifo
 */
int bf_space(struct bit_fifo *bf)
{
	return bf->byte_size * 8 - 1 - bf_size(bf);
}

/**
 * @details
 */
void __bf_put(struct bit_fifo *bf, char b)
{
	bf->bytes[bf->head_index] &= ~(1<<bf->head_offset);
	bf->bytes[bf->head_index] |= (b & 0x01) << bf->head_offset;
	incr_index_offset(bf->head_index, bf->head_offset, bf->byte_size);
}

/**
 * @details
 */
char __bf_get(struct bit_fifo *bf)
{
	incr_index_offset(bf->tail_index, bf->tail_offset, bf->byte_size);
	return (bf->bytes[bf->tail_index] >> bf->tail_offset) & 0x01;
}

/** 
 * @details 
 *
 * @param bf   	The destination bit fifo
 * @param bits  Bits array, each bit is saved in a byte
 * @param n 	Number of bits to be put into fifo
 *
 * @return 		The number of bits put into the fifo
 */
int bf_put_n(struct bit_fifo *bf, char *bits, int n)
{
	int i;

	for (i = 0; i < n && !bf_full(bf); i++)
		__bf_put(bf, bits[i]);
	if (i != n)
		ERROR("n = %d, i = %d\n", n, i);
	return i;	
}

/** 
 * @details 
 *
 * @param bf   	The source bit fifo
 * @param bits	Bits array, each bit is saved in a byte
 * @param n		Number of bits to be put into fifo
 * 
 * @return The number of bits got from the fifo
 */
int bf_get_n(struct bit_fifo *bf, char *bits, int n)
{
	int i;

	for (i = 0; i < n && !bf_empty(bf); i++)
		bits[i] = __bf_get(bf);
	if (i != n)
		ERROR("n = %d, i = %d\n", n, i);
	return i;	
}

/**
 * @details 
 *
 * @param bf	Bit fifo
 * @param bytes Buffer head
 * @param bit_len Length of bits to import
 *
 * @return The number of bits imported
 */
int bf_import(struct bit_fifo *bf, char *bytes, int bit_len)
{
	int i, j;
	char b;

	for (i = 0; i < bit_len/8; i++)
		for (j = 0; j < 8; j++) {
			if (bf_full(bf))
				goto out;
			b = (bytes[i] >> j) & 0x01;
			__bf_put(bf, b);
		}

	for (j = 0; j < bit_len%8; j++) {
		if (bf_full(bf))
			goto out;
		b = (bytes[i] >> j) & 0x01;
		__bf_put(bf, b);
	}
out:
	return i*8+j;
}

/**
 * @details 
 *
 * @param bf	Bit fifo
 * @param bytes	Buffer head
 * @param bit_len Length of bits to export
 *
 * @return The number of bits exported
 */
int bf_export(struct bit_fifo *bf, char *bytes, int bit_len)
{
	int i, j;
	char b;

	for (i = 0; i < bit_len/8; i++) {
		b = 0;
		for (j = 0; j < 8; j++) {
			if (bf_empty(bf))
				goto out;
			b |= __bf_get(bf) << j;
		}
		bytes[i] = b;
	}

	if (i == bit_len/8)
		return bit_len;

	for (j = 0, b = 0; j < bit_len%8; j++) {
		if (bf_empty(bf))
			goto out;
		b |= __bf_get(bf) << j;
	}
out:
	bytes[i] = b;
	return i*8+j;
}

/**
 * @details Dump the contents of a bit fifo to stdout.
 */
void bf_dump(struct bit_fifo *bf)
{
	struct bit_fifo bf1;
	char b;
	int id;

	memcpy(&bf1, bf, sizeof(bf1));
	printf("size=%d, space=%d", bf_size(bf), bf_space(bf));
	printf(" bs=%d, hi=%d, ho=%d, ti=%d, to=%d", bf->byte_size,
		bf->head_index, bf->head_offset, bf->tail_index,
		bf->tail_offset);
	id = 0;
	while (!bf_empty(&bf1)) {
		if (id % 16 == 0)
			printf("\n%06d  ", id);
		else if (id % 8 ==0)
			printf("  ");

		b = __bf_get(&bf1);
		printf("%2d", b);
		id++;
	}
	printf("\n");
}
