///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	filter.cpp
 * @author	Cal(it)2
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2011, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
////////////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cstdio>
#include <string.h>

#include "dsp/filter.hpp"

using namespace libdsp;

/*
 * @brief Initialize a FIR filter
 *
 * @param coef: the coefficiants of the FIR filter
 * @param n: 	 length of the coefficiants
 *
 * @return A pointer to a fir structure, or NULL if failed
 */
struct libdsp::fir *libdsp::fir_alloc(int *coef, int n)
{
	struct fir *f;

	f = (struct fir *)malloc(sizeof(struct fir) + 2*n*sizeof(int));
	if (!f) {
		printf("fir_alloc malloc failed\n");
		return 0;
	}

	f->order = n;
	f->coef = (int *)(f + 1);
	f->regs = f->coef + n;

	memcpy(f->coef, coef, n*sizeof(int));
	memset(f->regs, 0, n*sizeof(int));

	return f;
}

/*
 * @brief Destroy a FIR filter
 */
void libdsp::fir_free(struct fir *f)
{
	free(f);
}

/*
 * @brief Set the FIR coefficients to zeros
 */
void libdsp::fir_reset(struct fir *f)
{
	memset(f->regs, 0, f->order*sizeof(int));
}

/*
 * @brief Filter a data sequence
 *
 * @param f: the FIR filter
 * @param x: the current value in the data sequence
 *
 * @return The current filter output
 */
int libdsp::fir_seq(struct fir *f, int x)
{
	int i;
	int sum;

	/* shift registers */
	for (i = f->order-1; i >= 1; i--)
		f->regs[i] = f->regs[i-1];
	f->regs[0] = x;

	/* compute sum */
	for (i = 0, sum = 0; i < f->order; i++)
		sum += f->coef[i] * f->regs[i];

	return sum;
}
