///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	bit_stream.hpp
 * @author	Cal(it)2
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2011, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
////////////////////////////////////////////////////////////////////////////////

/* FIXME: optimize bit stream */
#include <cstdio>
#include <cstdlib>
#include <pthread.h>
#include <semaphore.h>
#include <limits.h>
#include <iostream>

#include "../../intfc.hpp"
#include "../../libsys.hpp"
#include "p25/format/bit_fifo.h"
#include "p25/format/bs_packet.h"

#define TX_BS_BUF_LEN8		BLOB_LEN8*2
#define RX_BS_BUF_LEN8		BLOB_LEN8*2

pthread_mutex_t mutex_bf_in;
pthread_mutex_t mutex_bf_out;

static struct bit_fifo *bf_in;
static struct bit_fifo *bf_out;
static int output_flush;
static int last_packet;
static pthread_t inThread;
static pthread_t outThread;
static sem_t input_bf_full_sem;
static sem_t input_bf_empty_sem;
static sem_t output_bf_full_sem;
static sem_t output_bf_empty_sem;


static void *input_task(void *arg)
{
	int ret, test, bf_isfull, bufsize;
	struct bs_packet pkt;

	while (1) {
		pthread_mutex_lock( &mutex_bf_in );
		bf_isfull = bf_full(bf_in) || 
			(bf_space(bf_in) < BLOB_LEN8);
		pthread_mutex_unlock( &mutex_bf_in );

		// Block on semaphore if buffer is full.
		while (bf_isfull){
			//printf("input_task:  bf_isfull, waiting\n");
			if (sem_wait(&input_bf_full_sem) < 0) {
				ERROR("input_task:  error in sem_wait\n");
				continue;
			}
			pthread_mutex_lock( &mutex_bf_in );
			bf_isfull = bf_full(bf_in) || 
				(bf_space(bf_in) < BLOB_LEN8);
			pthread_mutex_unlock( &mutex_bf_in );
		}
		//printf("input_task:  !bf_isfull, continuing\n");

		// Receive data from GPP.
		ret = dsp::intfc::read_gppi((char *)&pkt, 
			sizeof(struct bs_packet));
		if (ret < 0)
			perror("input_task:  read_gppi failed\n");

		// Put received data into buffer.
		pthread_mutex_lock( &mutex_bf_in );
		ret = bf_import(bf_in, pkt.data, pkt.blen);
		bufsize = bf_size(bf_in);
		pthread_mutex_unlock( &mutex_bf_in );

		if (ret != pkt.blen)
			perror("input_task:  bf import failed\n");

		if (bufsize > 0) {
			// Notify the consumer that the buffer is no longer empty.
			//printf("input_task:  bufsize > 0, sem_post(bf_empty)\n");
			if (sem_post(&input_bf_empty_sem) < 0) {
				perror("error in sem post\n");
				continue;
			}
		}
	}
}

/*
 * @brief Read bits from GPP
 * 
 * @param bits: array of bits, each bit is represented by a byte
 * @param n:	 number of bits to read
 *
 * @return The number of bits read or -1 if error
 */
int bs_input(char *bits, int n)
{
	int ret, bf_isempty, bufspace;

	if (n > TX_BS_BUF_LEN8) {
		printf("bs_input len too long %d\n", n);
		n = TX_BS_BUF_LEN8;
	}

	// Block on semaphore if buffer is empty.
	pthread_mutex_lock( &mutex_bf_in );
	bf_isempty = bf_empty(bf_in) || (bf_size(bf_in) < n);
	pthread_mutex_unlock( &mutex_bf_in );

	while (bf_isempty){
		//printf("bs_input:  bf_isempty, waiting\n");
		if (sem_wait(&input_bf_empty_sem) < 0) {
			perror("bs_input:  error in sem_wait\n");
			continue;
		}

		pthread_mutex_lock( &mutex_bf_in );
		bf_isempty = bf_empty(bf_in) || (bf_size(bf_in) < n);
		pthread_mutex_unlock( &mutex_bf_in );
	}
	//printf("bs_input:  !bf_isempty, continuing\n");

	// Retrieve requested data from buffer.
	pthread_mutex_lock( &mutex_bf_in );
	ret = bf_get_n(bf_in, bits, n);
	bufspace = bf_space(bf_in);
	pthread_mutex_unlock( &mutex_bf_in );

	if (ret != n)
		printf("bs_input:  bf get failed\n");

	if (bufspace > 0) {
		// Notify the producer that the buffer is no longer full.
		//printf("bs_input:  bf_space(bf_in) > 0, sem_post(bf_full)\n");
		if (sem_post(&input_bf_full_sem) < 0) {
			perror("bs_input:  error in sem_post\n");
			return 0;
		}
	}

	return ret;
}

static void *output_task(void *arg)
{
	int ret, bf_isempty, bufspace;
	struct bs_packet pkt;
	//short * ext_hdr;

	while (1) {
		// Block on semaphore if buffer is empty.
		pthread_mutex_lock( &mutex_bf_out );
		bf_isempty = bf_empty(bf_out) ||
			(bf_size(bf_out) < BLOB_LEN8);
		pthread_mutex_unlock( &mutex_bf_out );

		while (bf_isempty && !output_flush) {
			//printf("output_task:  bf_isempty, waiting\n");
			if (sem_wait(&output_bf_empty_sem) < 0) {
				perror("error in sem_wait\n");
				continue;
			}

			pthread_mutex_lock( &mutex_bf_out );
			bf_isempty = bf_empty(bf_out) ||
				(bf_size(bf_out) < BLOB_LEN8);
			pthread_mutex_unlock( &mutex_bf_out );
		}
		//printf("output_task:  !bf_isempty, continuing\n");

		// Retrieve data from buffer.
		pthread_mutex_lock( &mutex_bf_out );
		ret = bf_export(bf_out, pkt.data, BLOB_LEN8);
		bufspace = bf_space(bf_out);
		pthread_mutex_unlock( &mutex_bf_out );
		pkt.blen = ret;
		pkt.hdr = BS_PKT_DATA;

		if (output_flush && last_packet) {
			printf("Sending Last Packet\n");
			/*
			// Write special header in MSB of int.
			// Note: For little Endian systems this is position [0] and
			// for Big Endian systems this is postition [1]
			ext_hdr = (short *)&pkt.blen;
			ext_hdr[1] = 255;
			printf("ext_hdr = %d\n",pkt.blen);
			*/
			pkt.hdr = BS_PKT_LAST;
			printf("pkt.hdr = %d\n",pkt.hdr);
		}

		// Write retrieved data to GPP.
		ret = dsp::intfc::write_gppi((char *)&pkt, sizeof(struct bs_packet));
		if (ret < 0)
			printf("write gpp failed\n");

		// Reset output flush state.
		output_flush = 0;
		last_packet = 0;

		// Notify the producer that the buffer is no longer full.
		if (bufspace > 0) {
			//printf("output_task:  bf_space(bf_out) > 0, sem_post(bf_empty)\n");
			if (sem_post(&output_bf_full_sem) < 0) {
				perror("error in sem_post\n");
				continue;
			}
		}
	}
}

/*
 * @brief Write bits to GPP
 * 
 * @param bits: array of bits, each bit is represented by a byte
 * @param n:	 number of bits to write
 *
 * @return The number of bits write or -1 if error
 */
int bs_output(char *bits, int n)
{
	int ret, bf_isfull, bufsize;

	/*
	for (ret = 0; ret < n; ret++)
		printf("bo %d\n", bits[ret]);
	*/

	if (n > RX_BS_BUF_LEN8) {
		printf("bs_output len too long %d\n", n);
		n = RX_BS_BUF_LEN8;
	}

	pthread_mutex_lock( &mutex_bf_out );
	bf_isfull = bf_full(bf_out) ||
		(bf_space(bf_out) < n);
	pthread_mutex_unlock( &mutex_bf_out );

	// Block on semaphore if buffer is full.
	while (bf_isfull){
		//printf("bs_output:  bf_isfull, waiting\n");
		if (sem_wait(&output_bf_full_sem) < 0) {
			ERROR("bs_output:  error in sem_wait\n");
			continue;
		}
		pthread_mutex_lock( &mutex_bf_out );
		bf_isfull = bf_full(bf_out) ||
			(bf_space(bf_out) < BLOB_LEN8);
		pthread_mutex_unlock( &mutex_bf_out );
	}
	//printf("bs_output:  !bf_isfull, continuing\n");

	// Put payload into buffer.
	pthread_mutex_lock( &mutex_bf_out );
	ret = bf_put_n(bf_out, bits, n);
	bufsize = bf_size(bf_out);
	pthread_mutex_unlock( &mutex_bf_out );

	if (ret != n)
		printf("bf put failed\n");

	// Notify the consumer that the buffer is no longer empty.
	if (bufsize > 0) {
		//printf("bs_output:  bf_size(bf_out) > 0, sem_post(bf_empty)\n");
		if (sem_post(&output_bf_empty_sem) < 0) {
			perror("error in sem_post\n");
			return 0;
		}
	}

	return ret;
}

/*
 * @brief Flush the remaining bits in the bit
 *                   stream to GPP
 */

void bs_output_flush(int mode)
{
	int bufsize;

	pthread_mutex_lock( &mutex_bf_out );
	bufsize = bf_size(bf_out);
	pthread_mutex_unlock( &mutex_bf_out );

	if (bufsize > 0) {
		output_flush = 1;
		last_packet = mode;
		//printf("Flushing baseband->packetizer FIFO\n");
		if (sem_post(&output_bf_empty_sem) < 0) {
			perror("error in sem_post\n");
			return;
		}
	}
}

/* 
 * @brief Initialize bit stream
 *
 * @return 0 if success, -1 if failed
 */
int bs_init(void)
{
    int ret;
	pthread_attr_t attr;

	output_flush = 0;
	last_packet = 0;
	bf_in = bf_alloc(TX_BS_BUF_LEN8);
	bf_out = bf_alloc(RX_BS_BUF_LEN8);
	if (!bf_in || !bf_out) {
		printf("bf_alloc failed\n");
		return -1;
	}

	pthread_mutex_init(&mutex_bf_in, NULL);
	pthread_mutex_init(&mutex_bf_out, NULL);

	sem_init(&input_bf_full_sem, 0, 0);
	sem_init(&input_bf_empty_sem, 0, 0);
	sem_init(&output_bf_full_sem, 0, 0);
	sem_init(&output_bf_empty_sem, 0, 0);

    //attrs.priority = TSK_MAXPRI;
    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, /*PTHREAD_STACK_MIN + 0x2000*/ 131072);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    ret = pthread_create(&inThread, &attr, input_task, NULL);
    if (ret) ERROR("Failed to start bit stream input thread\n");

    pthread_attr_init(&attr);
    pthread_attr_setstacksize(&attr, /*PTHREAD_STACK_MIN + 0x2000*/ 131072);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    ret = pthread_create(&outThread, &attr, output_task, NULL);
    if (ret) ERROR("Failed to start bit stream output thread\n");

	return 0;
}

/*
 * @brief Close bit stream
 */
void bs_exit(void)
{
	sem_destroy(&input_bf_full_sem);
	sem_destroy(&input_bf_empty_sem);
	sem_destroy(&output_bf_full_sem);
	sem_destroy(&output_bf_empty_sem);
	pthread_mutex_destroy(&mutex_bf_in);
	pthread_mutex_destroy(&mutex_bf_out);
	bf_free(bf_in);
	bf_free(bf_out);
}
