///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	main.cpp
 * @brief	main() of the DSP module.
 * @details Being the program entry point, it sets up the necessary objects
 * 			and the spawns two threads, one each for Tx and Rx for our
 * 			half-duplex P25 SDR implementation.
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <cstring>
#include <cstdlib>
#include "p25/ecc/bch.h"
#include "p25/comm/ipcomm.h"
#include "p25/format/bit_stream.h"

#include "format/int_stream.h"
#include "libsys.hpp"
#include "intfc.hpp"
#include "modem.hpp"

using namespace std;

int sys_status; // global Tx/Rx switch

/**
 * @brief 	This channel is used to send control signals to the Radio(USRP)
 * 			to control its switching. The Tx thread sends these signals.
 */
char *uhd_ctrl_chan_path;

/**
 * @brief 	This channel is used to receive control signals from the Radio(USRP)
 * 			to control its switching. The Tx thread sends these signals.
 */
char *ctrl_chan_modem_baseband_path;

//The DSP modules interface at the Radio side.
//char *uhdi_source_fifo; 
//char *uhdi_sink_fifo;


int main(int argc, char *argv[])
{

int ret, rx_arg, tx_arg, term_arg;
pthread_t rx_thread, tx_thread, termthread;
pthread_attr_t tx_attr, rx_attr, tm_attr;
void *status;
char *tx_sock, *rx_sock, *uhd_tx_spath , *uhd_rx_spath;

if (argc < 7){
	cout<< "Usage: " << argv[0] << " <tx_port> <rx_port> <uhd_tx_port> <uhd_rx_port> <uhd_ctrl_chan_path> <ctrl_chan_modem_baseband_path>" << '\n';
	exit(-1);

}

//GPP interface
tx_sock = argv[1]; 
rx_sock = argv[2];

//Use if using sockets instead of FIFO pipes for UHD-DSP interface
uhd_tx_spath = argv[3];
uhd_rx_spath = argv[4];

//uhdi_sink_fifo = NULL;
//uhdi_source_fifo = NULL;

uhd_ctrl_chan_path = argv[5];
ctrl_chan_modem_baseband_path = argv[6];

///////////////////////////////////////////////////////////////////////////////
//Initialize everything
///////////////////////////////////////////////////////////////////////////////

//Interfaces
ret = dsp::intfc::intfc_init(tx_sock, rx_sock, uhd_tx_spath, uhd_rx_spath);

//System Init. Does nothing for now.
ret = sys_init();
if (ret < 0) {
	std::cout << "sys_init failed\n";
	exit(-1);
}

//Initialize the bit FIFO structures
ret = bs_init();
if (ret < 0) {
	std::cout << "bs_init failed\n";
	sys_exit(sys_status);
	exit(-1);
}

libdsp::bch_init();

//Always go into Rx state while system starts
sys_status = STAT_RX;

class dsp::modem::rx rx_chain; //Rx object
class dsp::modem::tx tx_chain; //Tx object


rx_chain.rx_init();
tx_chain.tx_init(&rx_chain);

//Initiate Rx_thread
pthread_attr_init(&rx_attr);
pthread_attr_setstacksize(&rx_attr, 131072);
ret = pthread_create(&rx_thread, &rx_attr, rx_task_start, (void *)&rx_chain);
if (ret)ERROR("Failed to create Rx thread, cannot receive...");


//Initiate Tx_thread
pthread_attr_init(&tx_attr);
pthread_attr_setstacksize(&tx_attr, 131072);
ret = pthread_create(&tx_thread, &tx_attr, tx_task_start, (void *)&tx_chain);
if (ret)ERROR("Failed to create Tx thread, cannot transmit.");


pthread_join(rx_thread, &status);

tx_chain.tx_exit();
rx_chain.rx_exit();
bs_exit();

sys_exit(sys_status);
std::cout << "Done!\n\n";


return 0;
}



