///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	modem.hpp
 * @brief	Declaration of the two main classes of the dsp::modem namespace;
 * 			rx and tx.
 *
 * @details At 30,000 feet, their jobs are to convert the bitstream
 * 			generated by the GPP packetizer moduleto the 4-level C4FM
 * 			analog waveform to be fed to the Radio module for Frequency
 * 			Modulation and vice versa. Their methods thus implement various
 * 			DSP functions of P25 mod/demod.
 *
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#ifndef _MODEM_H
#define _MODEM_H

#include <time.h>

#define SYMBOL_RATE			4800	/* baseband symbol rate	*/
#define SYMBOL_OSR			10		/* baseband signal oversample rate */
#define FIX_SHIFT			12		/* fix point shift left in computing */

#define abs(x)				((x) > 0 ? (x) : -(x))
#define bit2sym(s, b1, b2)	((s) = (1 - 2*(b1)) * (1 + 2*(b2)))
#define sym2bit(b1, b2, s, shift)		\
							((b1) = (s) < 0, (b2) = abs(s) >= (2<<shift))


namespace dsp{ 

namespace modem{

/**
 * @brief The rx class implements the Rx chain of the P25 SDR implementation.
 */
class rx{ //Rx class

public:

/**
 *  Initialize the rx variables and filter structures for the rx thread
 */
int rx_init(void);

/**
 * rx_stop - Stop Rx when transmit starts
 */
void rx_stop(void);

/**
 * rx_start - Start Rx, when Tx stops.
 */
void rx_start(void);


/**
 * @brief rx_exit - shutdown RX functionality
 */
void rx_exit(void);


/**
 * @brief 	rx_task() starts the actual Rx operaion.
 *
 * @details It calls various helper functions from within, to implement the
 * 			(averaging) filtering,  Rx correlation of the sync frame, bch \
 * 			decoding of the NID and symbol to bit conversion. This is the Rx
 * 			thread handler function.
 */
void rx_task(void);

}; 

/*!The tx class implements the Tx chain of the P25 SDR implementation.	
*/

class tx{  //Tx class

rx *rx_chain;

public:

/**
 * @brief	tx_init initializes variables and filter structures for the tx thread.
 * 			It also sets up the control channel b/w DSP and Radio/UHD.
 *
 * @param 	rx the rx object representing the Rx chain.
 *			This parameter is needed for access to the rx_start() and rx_stop()
 *			methods, from within the tx object.
 * @return 	Returns 0 on success, -1 otherwise.
*/
int tx_init(rx*);

/**
 * tx_exit - shut down the TX functionality
 */
void tx_exit(void);

/**
 * tx_start - start TX
 */
void tx_start(void);

/**
 * tx_stop - top TX
 */
void tx_stop(void);

/**
 * rx_modem_baseband_ctrl - receive control message from modem
 */
void rx_modem_baseband_ctrl();

/*!tx_task() starts the actual Tx operaion. It calls various helper functions 
*from within to implement the filtering/shaping, bit to symbol conversion and 
*modulation. This is the Tx thread handler function.
*/
void tx_task(void );

};

} //end namespace modem

} //end namespace dsp
#endif
