///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	rx.cpp
 * @brief	Member definitions of the dsp::modem::rx
 * 			class, and its helper functions.
 * @details	Much of this code has been ported to linux from the original code
 * 			written by Wenhua Zhao for the SDR4000 TI DSP  platform.
 *
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////


#include <cstdio>
#include <cstdlib>
#include <math.h>
#include <pthread.h>
#include <string.h>
#include "p25/constants.h"

#include "libsys.hpp"
#include "modem.hpp"
#include "intfc.hpp"
#include "format/int_stream.h"
#include "format/bit_stream.h"
#include "dsp/filter.hpp"
#include "dsp/p25_frame.hpp"

///////////////////////////////////////
//Declare constants
///////////////////////////////////////
#define CORR_THRESHOLD			120
#define CORR_CACHE_SIZE			18
#define CORR_MIN_HOLD			9
#define SYNC_SYMBOL_LEN			24
#define SYNC_SAMPLE_LEN			(SYNC_SYMBOL_LEN * SYMBOL_OSR)
#define SAMPLE_REGS_SIZE		(SYNC_SAMPLE_LEN + CORR_CACHE_SIZE)

using namespace libdsp;


///////////////////////////////////////
/**
 * @brief rx chain global variables
 */
///////////////////////////////////////

static int average[SYMBOL_OSR] = {
	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1
};

static int sync[SYNC_SYMBOL_LEN] = {
	 3,	 3,	 3,	 3,
	 3,	-3,	 3,	 3,
	-3,	-3,	 3,	 3,
	-3,	-3,	-3,	-3,
	 3,	-3,	 3,	-3,
	-3,	-3,	-3,	-3
};

static struct fir *fir_average;

static int sample_regs[SAMPLE_REGS_SIZE];
static int sample_regs_head;
static int sample_sync_offset;

static int corr_cache[CORR_CACHE_SIZE];
static int corr_index = 0;

static int rx_state;

#define RX_SYNC				0
#define RX_DATA				1

#define sample_regs_shift(x)	do {									\
	sample_regs_head = (sample_regs_head + 1) % SAMPLE_REGS_SIZE;		\
	sample_regs[sample_regs_head] = x;									\
} while (0)

#define sample_regs_get_symbol(x)	do {								\
	x = sample_regs[(sample_regs_head - sample_sync_offset +			\
					 SAMPLE_REGS_SIZE) % SAMPLE_REGS_SIZE];				\
	sample_sync_offset -= SYMBOL_OSR;									\
} while (0)


///////////////////////////////////////////////////////////////////////////////
/**
 *@ brief rx class helper functions
 */
///////////////////////////////////////////////////////////////////////////////

/**
 * Estimate the DC offset
 * @return the dc offset
 */
static int estimate_dc_offset(void)
{
	float fnsum = 0.0;
	int i, idx, sum;

	sum = 0;
	idx = sample_sync_offset;
	for (i = 0; i < 22; i++) {
		sum += sample_regs[(sample_regs_head - idx +
							SAMPLE_REGS_SIZE) % SAMPLE_REGS_SIZE];
		idx -= SYMBOL_OSR;
	}

	fnsum = (float)sum/22.0;
	//printf("sum/22 = %f, rounded value = %d\n", (float)sum / 22.0, (int)floor(fnsum + 0.5));

	return (int)floor(fnsum + 0.5);
}

/**
 * Calculate correlation between sync and registered samples
 *
 * @return the correlation
 */
static int sample_regs_corr(void)
{
	int i, j;
	int sum = 0;

	for (i = 0; i < SYNC_SYMBOL_LEN; i++) {
		/* Note:  The order of elements in sync is reversed with respect 
		          to the order of the circular buffer sample_regs.  We
		          multiply by SYMBOL_OSR since we are performing the
		          correlation on symbols, not samples.  We add
		          SAMPLE_REGS_SIZE prior to taking the modulus to ensure
		          that the value (an index for the sample_regs array) is
		          a positive integer, e.g. (-1 + 10) % 10 = 9.
		 */
		j = (sample_regs_head - (SYNC_SYMBOL_LEN - 1 - i)*SYMBOL_OSR +
			 SAMPLE_REGS_SIZE) % SAMPLE_REGS_SIZE;
		sum += sync[i] * sample_regs[j] >> FIX_SHIFT;
	}
	return sum;
}

/*
 * Find the index of the largest data value (integer) in an array.
 *
 * @param array		data array
 * @param length	length of the array
 *
 * @return the index of the maximum data, or -1 if error
 */
static int max_index(int *array, int length)
{
	int i, index = -1;
	int max = CORR_THRESHOLD;

	for (i = 0; i < length; i++)
		if (array[i] >= max) {
			max = array[i];
			index = i;
		}
	return index;
}

/**
 * Find from the the corr_cache, the maximum correlation above threshold.
 *
 * @param corr the new correlation value to save to corr_cache
 *
 * @return the index of the maximum valid correlation
 */
static int find_max_corr(int corr)
{
	int offset;

	if (corr < CORR_THRESHOLD && corr_index < CORR_MIN_HOLD) {
		corr_index = 0;
		return -1;
	}

	corr_cache[corr_index++] = corr;
	if (corr < CORR_THRESHOLD || corr_index == CORR_CACHE_SIZE) {
		offset = corr_index - 1 - max_index(corr_cache, corr_index);
		corr_index = 0;
		return offset;
	}
	return -1;
}

///////////////////////////////////////////////////////////////////////////////
/**
 * @brief rx class member definitions
 */
///////////////////////////////////////////////////////////////////////////////



/** 
 * @brief	Initialize RX functionality
 */
int dsp::modem::rx::rx_init(void)
{
	fir_average = fir_alloc(average, SYMBOL_OSR);
	if (!fir_average) {
		printf("alloc fir_average failed\n");
		return -1;
	}

	memset(sample_regs, 0, sizeof(sample_regs));
	sample_regs_head = 0;
	sample_sync_offset = -1;

	rx_state = RX_SYNC;
	return 0;
}


/**
 * @brief	Shutdown RX functionality
 */
void dsp::modem::rx::rx_exit(void)
{
	fir_free(fir_average);
}

/**
 * @brief	Start RX mode
 */
void dsp::modem::rx::rx_start(void)
{
	fir_reset(fir_average);
	memset(sample_regs, 0, sizeof(sample_regs));
	sample_regs_head = 0;
	sample_sync_offset = -1;

	rx_state = RX_SYNC;
	dsp::intfc::uhdi_rx_en(1);
	printf("Rx Started\n");

}


/**
 * @brief 	Implement the Rx thread
 *
 * 			The thread is controlled by input data availability
 * 			at the UHD/Radio interface. The thread continues execution until
 * 			there is data (float stream) available from the Radio, otherwise
 * 			it waits on input from Radio.
 */
void dsp::modem::rx::rx_task()
{
	int x, prevx, dc_offset, sym;
	char b[2], hdr[HDR_BLEN];
	int frame_offset, frame_length;

    int count = 0;
    clock_t start;
    clock_t end;
    clock_t diff;

	INFO("RX task running\n");

/* rx debug dumps
#ifdef DEBUG_DUMP
	FILE *rx_dco_file = fopen("rx_dco_dump.dat", "wb");
	FILE *rx_x1_file = fopen("rx_x1_dump.dat", "wb");
	FILE *rx_x2_file = fopen("rx_x2_dump.dat", "wb");
#endif
*/

	//FILE *testfile = fopen("c4_samples.dat", "wb");
	//FILE *testfile1 = fopen("c4_dc_offsets.dat", "wb");
	//FILE *testfile2 = fopen("c4_sym2bits.dat", "wb");

loop: //infinite loop that switches between two states RX_DATA and RX_SYNC
	
	/* obtain a new sample (16-bit integer) from RX input buffer */
	is_input(&x);
	x = (short)x;

	/* average to cancel out tx pre-compensate filter */
	x = fir_seq(fir_average, x) / SYMBOL_OSR;
	
/* rx debug dumps
#ifdef DEBUG_DUMP
	fwrite((char *)&x, 1, sizeof(int), rx_x1_file);
#endif
*/
	
	/* put filtered sample into FIFO */
	sample_regs_shift(x);

	switch (rx_state) {
	case RX_DATA://Sync frame is found
		sample_sync_offset++;
		if (sample_sync_offset < 0)
			break;

		if (frame_offset == 0) {//Estimate the DC offset
			dc_offset = estimate_dc_offset();
			//printf("dc_offset is %d\n", dc_offset);
		}

		/* obtain symbol from FIFO */
		sample_regs_get_symbol(x);

		//fprintf(testfile, "%d\n", x);
		//fprintf(testfile1, "%d\n", dc_offset);
			
/* rx debug dumps
#ifdef DEBUG_DUMP
		fwrite((char *)&x, 1, sizeof(int), rx_x2_file);
		fwrite((char *)&dc_offset, 1, sizeof(int), rx_dco_file);
#endif
*/
			
		x -= dc_offset;			/* remove dc offset */

		//sym2bit(b[0], b[1], x, FIX_SHIFT); // Convert from symbols to bits

		b[0] = (x < 0);
		b[1] = (abs(x) >= (2<<FIX_SHIFT));

		//fprintf(testfile2, "%d %d %d\n", b[0], b[1], x);

//		printf("x / 4096 = %f, b[0] = %d, b[1] = %d\n", x / 4096.0, b[0], b[1]);

		/* attempt to decode frame header */
		if (frame_offset < HDR_BLEN) {
			hdr[frame_offset++] = b[0];
			hdr[frame_offset++] = b[1];
			if (frame_offset == HDR_BLEN) {
				/* successfully retrieved frame header, decode to obtain frame length */
				frame_length = frame_header_decode(hdr);
				if (frame_length <= HDR_BLEN ) {
					/* invalid frame length, setting rx state to sync search */
					rx_state = RX_SYNC;
					//printf("RX Frame Null at %d\n", (int)time(NULL));
				} else {
					/* send header to packetizer */
					bs_output(hdr, HDR_BLEN);
					//printf("Sent HDR to packetizer\n");
				}
			}
		} else {
			/* send contents of frame to packetizer */
			bs_output(b, 2);
			frame_offset += 2;
			if (frame_offset == frame_length) {
				/* end of frame, setting rx state to sync search */
				if (frame_length == TDU_BLEN || frame_length == TDU_LCW_BLEN) {
					printf("Got TDU\n");
					bs_output_flush(1);
				} else
					bs_output_flush(0);

				rx_state = RX_SYNC;
				//printf("RX Frame Completed at %d ms, len = %d\n",
				//		(int) time(NULL), frame_length);
			}
		}
		break;
	case RX_SYNC: //Search for RX Sync, if found change state to RX_DATA to start receiving P25 frames

      /* for delay testing
	    if ((count % 10000000) == 0) {
            end = clock();
            diff = end - start;
            INFO("Clock Ticks = %f, Seconds = %f \n", (double)diff, ((double)diff/CLOCKS_PER_SEC));
            count = 0;
            start = clock();
        }
        count++;
       */
			
		/* compute correlation between sync and current samples in FIFO register */
        	x = sample_regs_corr();
			
		/* cache current correlation value, find the maximum correlation value in the cache,
		   and set that value as the sample sync offset.
		 */ 
		sample_sync_offset = find_max_corr(x);
        if (sample_sync_offset > 0) { // Found sync frame, start decoding the Rx frame data
			sample_sync_offset += (SYNC_SYMBOL_LEN - 1) * SYMBOL_OSR;

			frame_length = 0;
			frame_offset = 0;

			rx_state = RX_DATA;
			//printf("RX SYNC Found at %d ms\n", (int) time(NULL));
		}
		break;
	default:
		break;
	}

	goto loop;
}

/**
 * Stop Rx
 */
void dsp::modem::rx::rx_stop(){
	dsp::intfc::uhdi_rx_en(0);
	printf("Rx Stopped\n");
}




