///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	tx.cpp
 * @brief	Member definitions of the dsp::modem::tx class, and its helper
 * 			functions.
 * @details	Part of this code has been ported to linux from the original code
 * 			written by Wenhua Zhao for the SDR4000 TI DSP  platform.
 *
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009, The Regents of the University of California
 * All Rights Reserve
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#include <cstdio>
#include <time.h>
#include <cstring>
#include <iostream>
#include "p25/comm/ipcomm.h"
#include "p25/constants.h"

#include "libsys.hpp"
#include "intfc.hpp"
#include "modem.hpp"
#include "format/int_stream.h"
#include "format/bit_stream.h"
#include "dsp/filter.hpp"
#include "dsp/p25_frame.hpp"



// Define constants
#define PULSE_SHAPE_LEN		159
#define PADDING_LEN			PULSE_SHAPE_LEN

/**
 * @brief	The DSP modem to Radio control channel.
 *
 * 			Used to control the radio. Typically it switches the Radio (USRP)
 * 			between transmit and receive modes.
 */
P25::Comm::IPComm* uhd_modem_ctrl_chan;
P25::Comm::IPComm* ctrl_chan_modem_baseband;

extern char *uhd_ctrl_chan_path;
extern char *ctrl_chan_modem_baseband_path;

using namespace libdsp;


//////////////////////////////////////////
/**
 * @brief tx chain global variables
 */
//////////////////////////////////////////


static int pulse_shape[PULSE_SHAPE_LEN] = { //Coefficients for the shaping filter
	0,
	0,
	0,
	0,
	-1,
	-1,
	0,
	1,
	3,
	3,
	2,
	-2,
	-6,
	-9,
	-9,
	-4,
	5,
	16,
	23,
	23,
	12,
	-8,
	-35,
	-64,
	-88,
	-102,
	-103,
	-90,
	-65,
	-29,
	14,
	62,
	109,
	151,
	183,
	199,
	195,
	169,
	119,
	50,
	-33,
	-124,
	-213,
	-290,
	-345,
	-369,
	-356,
	-303,
	-211,
	-85,
	65,
	227,
	385,
	521,
	618,
	662,
	641,
	551,
	390,
	168,
	-101,
	-396,
	-691,
	-956,
	-1160,
	-1277,
	-1281,
	-1155,
	-891,
	-491,
	33,
	659,
	1355,
	2085,
	2806,
	3473,
	4046,
	4485,
	4762,
	4856,
	4762,
	4485,
	4046,
	3473,
	2806,
	2085,
	1355,
	659,
	33,
	-491,
	-891,
	-1155,
	-1281,
	-1277,
	-1160,
	-956,
	-691,
	-396,
	-101,
	168,
	390,
	551,
	641,
	662,
	618,
	521,
	385,
	227,
	65,
	-85,
	-211,
	-303,
	-356,
	-369,
	-345,
	-290,
	-213,
	-124,
	-33,
	50,
	119,
	169,
	195,
	199,
	183,
	151,
	109,
	62,
	14,
	-29,
	-65,
	-90,
	-103,
	-102,
	-88,
	-64,
	-35,
	-8,
	12,
	23,
	23,
	16,
	5,
	-4,
	-9,
	-9,
	-6,
	-2,
	2,
	3,
	3,
	1,
	0,
	-1,
	-1,
	0,
	0,
	0,
	0
};

static struct fir *fir_pulse_shape;


/**
 *  @brief Modulate a symbol.
 *
 *  Also performs interpolation from 4.8ksps to 48ksps. After modulation,
 *  the symbols are output to the radio module.
 *
 *  @param sym the symbol to modulate
 */
static void modulate(int sym)
{
	int j;
	int x;

	for (j = 0; j < SYMBOL_OSR; j++) {
		/* interpolate with zeros */
		x = (j == 0) ? sym : 0;

		/* pulse shape filter */
		x = fir_seq(fir_pulse_shape, x);

		/* send out data */
	   	is_output(&x);
	}
}

////////////////////////////////////////////////////////////////////
/**
 * @brief tx class member definitions
 */
////////////////////////////////////////////////////////////////////

// tx_init - initialize the TX functionality
int dsp::modem::tx::tx_init(dsp::modem::rx *rx_obj)
{
	//setup control channel b/w DSP modem and UHD radio
	uhd_modem_ctrl_chan = new P25::Comm::IPComm(uhd_ctrl_chan_path);
	uhd_modem_ctrl_chan->start_connection(P25::Comm::IPComm::TXMode);

	ctrl_chan_modem_baseband = new P25::Comm::IPComm(ctrl_chan_modem_baseband_path);
	if (ctrl_chan_modem_baseband->start_connection(P25::Comm::IPComm::RXMode) != 0) {
		ERROR("Error starting connection to ctrl_chan_modem_baseband!\n");
		return 0;
	}
	printf("opened %s\n", ctrl_chan_modem_baseband_path);

	fir_pulse_shape = fir_alloc(pulse_shape, PULSE_SHAPE_LEN);
	rx_chain = rx_obj;

	if (!fir_pulse_shape) {
		INFO("alloc fir_pulse_shape failed\n");
		return -1;
	}
	return 0;
}

// tx_exit - shut down the TX functionality
void dsp::modem::tx::tx_exit(void)
{
	fir_free(fir_pulse_shape);
}

// tx_stop - top TX
void dsp::modem::tx::tx_stop(void)
{
	INFO("Tx Stopped at %d ms\n", (int)time(NULL));

	is_output_flush();
	int res;
	unsigned char signal[1];
	
//set radio back to receive mode
	signal[0]= 0xEE;
	res=uhd_modem_ctrl_chan->send((char *)signal, sizeof(signal));
	if (res!=sizeof(signal))
			perror("Control Channel: Could not signal Tx Stop");

//not required
	dsp::intfc::uhdi_tx_en(0);
}

// tx_start - start TX
void dsp::modem::tx::tx_start()
{
	unsigned char signal[1];
	int res;

//set radio to transmit mode
	signal[0]= 0xBB;
	res=uhd_modem_ctrl_chan->send((char *)signal, sizeof(signal));
	if (res!=sizeof(signal))
		perror("Control Channel: Could not signal Tx Start");
	fir_reset(fir_pulse_shape);

	//not required
	dsp::intfc::uhdi_tx_en(1);

	INFO("Tx Started at %d ms\n", (int)time(NULL));

}

void dsp::modem::tx::rx_modem_baseband_ctrl()
{
	unsigned char signal[1];
	int res;

	do {
		res = ctrl_chan_modem_baseband->receive((char *)signal, sizeof(signal));
		if (res != sizeof(signal)) {
			ERROR("Control Channel: Error receiving baseband ACK ctrl\n");
		} else {
			if (signal[0] == 0xAA)
				INFO("Baseband ACK received\n");
		}
	} while(signal[0] != 0xAA);
}

/**
 * @brief	tx_task - Implement the Tx thread
 *
 * 			The thread is controlled by input data
 * 			availability at the GPP interface. The thread continues execution
 * 			until there is data (bitstream) available from the GPP, otherwise
 * 			it waits on input from GPP (waits at the bs_input() function call).
 */
void dsp::modem::tx::tx_task() {
	int res;
	int sym;
	char b[2], hdr[HDR_BLEN];
 	int frame_length=0, i;
 	int count = 0;
	clock_t start;
	clock_t end;
	clock_t diff;
	
	INFO("TX task running\n");

/* tx debug dumps
#ifdef DEBUG_DUMP
	FILE *tx_fmfile = fopen("tx_dsp_dump.dat", "wb");
	if (tx_fmfile == NULL)
		ERROR("Error opening up tx_gpp_dump.dat!\n");
#endif
*/

/*
	char *buf = new char[pkt.blen/8];
	for (int i = 0; i < pkt.blen/8; i++) {
		unsigned char a = pkt.data[i];
		a = ((a & 0xAA) >> 1) | ((a & 0x55) << 1);
		a = ((a & 0xCC) >> 2) | ((a & 0x33) << 2);
		a = ((a & 0xF0) >> 4) | ((a & 0x0F) << 4);

		buf[i] = a;
	}
*/
	//struct bit_fifo *bf_in_tx = bf_alloc(HDR_BLEN);

/* tx debug dumps
#ifdef DEBUG_DUMP
	int cnt;
#endif
*/

loop: //infinite loop to transmit data, controlled by data availability at GPP (waits at bs_input() )
    /*
    if ((count % 100) == 0) {
        end = clock();
        diff = end - start;
        INFO("Clock Ticks = %f, Seconds = %f \n", (double)diff, ((double)diff/CLOCKS_PER_SEC));
        count = 0;
        start = clock();
    }
    count++;*/

	res = bs_input(hdr, HDR_BLEN);
	if (res < HDR_BLEN) {
		ERROR("header retrieve failed, res = %d, HDR_BLEN = %d\n", res, HDR_BLEN);
		sys_status = STAT_ERR;
		goto loop;
	}

	if (sys_status != STAT_TX) {
		sys_status = STAT_TX;
		rx_chain->rx_stop();
		tx_start();
	}

	frame_length = frame_header_encode(hdr);
	if (frame_length <= HDR_BLEN) {
		INFO("header encode failed\n");
		sys_status = STAT_ERR;
		goto loop;
	}

/* tx debug dumps
#ifdef DEBUG_DUMP
	cnt = fwrite(hdr, 1, HDR_BLEN, tx_fmfile);
	fflush(tx_fmfile);
	if (cnt != HDR_BLEN)
		ERROR("Error writing to tx_dsp_file!\n");
#endif
*/
	for (i = 0; i < HDR_BLEN; i += 2) {
		bit2sym(sym, hdr[i], hdr[i+1]);
		modulate(sym);
	}

	for (i = HDR_BLEN; i < frame_length; i += 2) {
		bs_input(b, 2);

/* tx debug dumps
#ifdef DEBUG_DUMP
	cnt = fwrite(b, 1, 2, tx_fmfile);
	fflush(tx_fmfile);
	if (cnt != 2)
		ERROR("Error writing to tx_dsp_file!\n");
#endif
*/
		bit2sym(sym, b[0], b[1]);
		modulate(sym);
	}

	/* FIXME: use bit stream control instead of frame_length to
	 * terminate transmission */
	if (frame_length == TDU_BLEN || frame_length == TDU_LCW_BLEN) {
		for (i = 0; i < PADDING_LEN; i++)
			modulate(0);

		sys_status = STAT_RX;
	
		/*	
		INFO("Waiting for Baseband ACK message\n");
		rx_modem_baseband_ctrl();
		*/

		tx_stop();
		rx_chain->rx_start();
	}

	goto loop;
}
