///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	Freq_Mod.hpp
 * @brief	Implementation of Frequency Modulator and Demodulator class
 * 			functions.
 *
 * @author	Cal(it)2
 * @author  Minsoo Kang
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <cmath>
#include "Freq_Mod.hpp"

#define PI 3.141592653589793
#define sine_table_len 2048

const float AMP_SCALE = .0625;


float Freq_Mod::sum = 0;
float Freq_Demod::prev_phase = 0;

using namespace std;

// Freq_Mod Constructor
// rate: sampling rate (= 48ksps)
// delta: frequency deviation
Freq_Mod::Freq_Mod(double tx_rate, unsigned int tx_delta)
{
	rate = tx_rate;
	delta = tx_delta;
	sum = 0;
}

// Freq_Mod Destructor
Freq_Mod::~Freq_Mod()
{
}

// FM Modulation function
// sum: keep the last phase information between function calls
void Freq_Mod::fm_mod(float input[], float real_in[], float imag_in[], size_t len_input) {
	input[0] = input[0]*delta/rate + sum;

	for (int i=1; i<len_input; i++) {
		input[i] = fmod(input[i-1] + input[i]*delta/rate, 1);
	}
	sum = input[len_input-1];

	for(int i=0; i < len_input; i++) {
		real_in[i] = AMP_SCALE*cos(2*PI*input[i]);
		imag_in[i] = AMP_SCALE*sin(2*PI*input[i]);
	}
}

void Freq_Mod::reset() {
	sum = 0;
}



// FM Demod
Freq_Demod::Freq_Demod(double rx_rate, unsigned int rx_delta)
{	
	rate = rx_rate;
	delta = rx_delta;
	prev_phase = 0;
}

Freq_Demod::~Freq_Demod()
{
}


size_t Freq_Demod::comp(float in) {
	if (in > PI)
		return 1;
	else return 0;
}

void Freq_Demod::fm_demod(float real_in[], float imag_in[], float output[], size_t len_input) 
{
	float cur_phase, phase_diff, f;

	for (size_t i=0; i<len_input; i++) {
		cur_phase = atan2(imag_in[i], real_in[i]);
		phase_diff = cur_phase - prev_phase;
		//f= phase_diff - Freq_Demod::comp(phase_diff)*2*PI + Freq_Demod::comp(-phase_diff)*2*PI;
		f = phase_diff - (phase_diff > PI)*2*PI + (phase_diff < -PI)*2*PI;
		output[i] = f/2/PI/delta*rate;
		prev_phase = cur_phase;
	}

/*
	// PLL Implementation of FM demodulator.
	float *d1 = new float[len_input];
	float *d2 = new float[len_input];
	float *diff = new float[len_input];
	float *vco = new float[len_input];
	float c, s;

	const float alpha = 0.99;
	const float K = 0.02;

	d1[0] = 0; 
	d2[0] = 0; 
	diff[0] = 0; 

	output[0] = real_in[0];
	vco[0] = 0;
	for (int i = 1; i < len_input; i++) {
		c = cos(vco[i-1]);
		s = sin(vco[i-1]);
		d1[i] = c*real_in[i] - s*imag_in[i];
		d2[i] = c*imag_in[i] + s*real_in[i];

		diff[i] = atan(d1[i]/d2[i]);

		output[i] = diff[i]*K + output[i-1]*alpha;
		vco[i] = vco[i-1] + output[i];
	}

	for (int i = 0; i < len_input; i++) {
		output[i] = output[i]/2/PI/delta*rate;
	}
*/
}

void Freq_Demod::reset() {
	prev_phase = 0;
}


// Sine wave

gen_sine::gen_sine() {

	sine_table = new float[sine_table_len];
	
	static const double tau = 2*PI;
	for (size_t i = 0; i < sine_table_len; i++)
        	sine_table[i] = float(std::sin((tau*i)/sine_table_len));
}
gen_sine::~gen_sine() {
}



float gen_sine::sine(float x){
	return sine_table[size_t(x*sine_table_len)%sine_table_len];
}

void gen_sine::gen_sine_wave(float sine_buf[], double rate, double wave_freq, float tx_ampl, size_t spb){
	cps = wave_freq/rate;

	for (size_t n = 0; n < spb; n++){
       		sine_buf[n] = tx_ampl*gen_sine::sine(theta);
            	
             	theta += cps;
        }
	
	theta = std::fmod(theta, 1);
}





