///////////////////////////////////////////////////////////////////////////////
/**
 * @file 	radio.hpp
 * @brief	Declarations for the radio module.
 *
 * @author	Cal(it)2
 * @author  Sivaramakrishnan Chandrasekaran <bscrk29@gmail.com>
 * @author  Jeffrey Cuenco
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
///////////////////////////////////////////////////////////////////////////////

#ifndef __RADIO__
#define __RADIO__

/**
 * @brief	Tx thread for the radio.
 *
 * @details	Operation is controlled by signals from DSP/Baseband
 *			module via control channel
 */
void *uhd_tx(void *);

/**
 * @brief	Rx thread for the radio.
 *
 * @details	Operation is controlled by signals from DSP/Baseband
 *			module via control channel
 */
void *uhd_rx(void *);

/**
 * @brief	Control signal thread
 *
 * @details	Responds to control signals
 *			received via the control channel, sent by the DSP/Baseband
 *			module.
 */
void *control_sig(void *);

/**
 * @brief	Frequency-change control signal thread 
 *
 * @details	Responds to control signals received via the control channel, 
 *			sent by the Packetizer component.
 */
void *freq_sig(void *);

/**
 * @brief			RF Carrier (TX and RX) frequency change function
 * @details			Sets both the TX and RX frequencies for the RF front end.
 *					This function is equivalent to calling both set_txfreq() 
 *					and set_rxfreq() passing in the same frequency value.
 * @param cfreq		The carrier frequency in Hz
 */
void change_cfreq(double cfreq);

/**
 * @brief			TX RF frequency change function.
 * @details			Sets the TX frequency for the RF front end.
 * @param cfreq		The carrier frequency in Hz
 */
void set_txfreq(double cfreq);

/**
 * @brief			RX RF frequency change function.
 * @details			Sets the RX frequency for the RF front end.
 * @param cfreq		The carrier frequency in Hz
 */
void set_rxfreq(double cfreq);

/**Constants for radio operation.
 * \brief	The constants defined below set the parameters for Frequency
 *	  		Modulation and  Up/Down sampling.
 */
#define CHUNK_LEN 480
#define DELTA 600
//#define FM_RATE 200000	//48000  // FM sampling rate = 48ksps
#define FM_RATE 48000  // FM sampling rate = 48ksps
#define UP_SAMPL_RATE 25 // 200ksps/48ksps = 25/6
#define DOWN_SAMPL_RATE 6
#define TX_NUM_SAMPS_ONE_SIDE 4
#define RX_NUM_SAMPS_ONE_SIDE 40
#define TX_FIR_LENGTH 201
#define RX_FIR_LENGTH 481
#define RS_LEN 87


#endif
