/**
 * @file    resample.hpp
 * @brief   Class definition for resampling operation.  This operation 
 *          resamples a sequence at upsample_rate/downsample_rate times 
 *          the original sampling rate, using a polyphase filter 
 *          implementation.
 * @author  Zhongren Cao <zcao@soe.ucsd.edu>
 *
 * @verbatim
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *      02/17/2011          Zhongren Cao        
 *
 * ************************************************************************** 
 *  Copyright (c) 2011 	The Regents of the University of California
 *
 *             	Zhongren Cao (zcao@ucsd.edu)
 *				Wireless System and Engineering Group
 *				UCSD Division
 *				California Institute for Telecommunications and
 *						Information Technology (Calit2)
 *  All Rights Reserved
 *			
 *  Permission to copy, modify and distribute any part of this software for 
 *  educational, research, United States Government, and non-profit purposes, 
 *  without fee, and without a written agreement is hereby granted, provided 
 *  that the above copyright notice, this paragraph and the following three 
 *  paragraphs appear in all copies.
 *
 *  Those desiring to incorporate this software into commercial products or 
 *  use for commercial purposes should contact the Technology Transfer & 
 *  Intellectual Property Services, University of California, San Diego, 
 *  9500 Gilman Drive, Mail Code 0910, La Jolla,  CA 92093-0910,
 *  Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL: invent@ucsd.edu.
 *
 *  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA OR ANY CONTRIBUTOR BE 
 *  LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 *  CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA OR ANY
 *  CONTRIBUTORHAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  THIS SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 *  CALIFORNIA OR ANY CONTRIBUTOR HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 *  SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF CALIFORNIA 
 *  OR ANY CONTRIBUTORS MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY 
 *  KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *  WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE 
 *  USE OF THIS SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 **********************************************************************************
 * @endverbatim
 */

//#include <boost/format.hpp>
#include <iostream>
#include <fstream>
#include "resample.hpp"

/*
	Resample_OP(unsigned short, unsigned short, unsigned short, size_t, float *);
	Class constructor. Populate the upsample_rate, downsample_rate, the number
	of one side samples for interpolation, the FIR filter length and coeffients.
	Input:
		u -- upsampling rate
		d -- downsampling rate
		n -- number of one side samples for interpolation
		l -- filter length
		fir -- FIR filter coeffiences
*/
Resample_OP::Resample_OP(unsigned short u, unsigned short d, unsigned short n, size_t l, float *fir){	
	
	int rr,cc;
	int count;
	
	upsample_rate = u;
	downsample_rate = d;
	num_oneside_sample = n;
	fir_length = l;
	i_stamp = -u;
	o_stamp = 0;
	t_thsd = (n-1)*u;
	
	if ((2*num_oneside_sample*upsample_rate+1)!=fir_length){ 	//Sanity Check
		std::cout << "The length of the filter is not appropriate!!" << std::endl;
	}
	else{
		std::cout << "The length of the filter is correct!!" << std::endl;
		
		num_row = upsample_rate;
		num_col = 2*num_oneside_sample+1;
		count = 1-num_row;
		fir_coef_twoD = new float*[num_row];
		for(rr=0;rr<num_row;rr++)
			fir_coef_twoD[rr] = new float[num_col];
		for(cc=0;cc<num_col;cc++){
			for(rr=(num_row-1);rr>=0;rr--){
				if (count<0){
					fir_coef_twoD[rr][cc]=0;
				}
				else{
					fir_coef_twoD[rr][cc]=fir[count];
				}
				count++;	
			}
		}
		reg = new float[num_col];
		for(cc=0;cc<num_col;cc++){
			reg[cc]=0;
		}
	}	
}

void Resample_OP::print_2d(){
	int rr,cc;
	for(rr=0;rr<num_row;rr++){
		for(cc=0;cc<num_col;cc++){
//			std::cout << boost::format("%1$#10.6f_") % fir_coef_twoD[rr][cc];
		}
		std::cout << std::endl;
	}
}

unsigned short Resample_OP::resample(float input, float *output){

	int cc, t_dis, tmp;
	unsigned short num_output = 0, state;
	float sum;
	
	/* Shift the internal registers */
	for(cc=0;cc<(num_col-1);cc++){
		reg[cc] = reg[cc+1];
	}
	reg[num_col-1] = input;
	/* Advance input time stamp */
	i_stamp = i_stamp + upsample_rate;
	
	tmp = o_stamp;
	while((i_stamp-tmp)>t_thsd){
		num_output ++;
		tmp += downsample_rate;
	}	
	
	if (num_output > 0){
		for(tmp=0;tmp<num_output;tmp++){
			output[tmp] = 0;
			state = o_stamp % upsample_rate;
			for(cc=0;cc<(num_col-1);cc++){
				output[tmp]+=reg[cc]*fir_coef_twoD[state][cc];
			}
			o_stamp += downsample_rate;
		}
	}
	
	while(o_stamp >= upsample_rate){
		o_stamp -= upsample_rate;
		i_stamp -= upsample_rate;
	}	
	return num_output;
}
