/**
 * @file    resample.hpp
 * @brief   Class definition for resampling operation.  This operation 
 *          resamples a sequence at upsample_rate/downsample_rate times 
 *          the original sampling rate, using a polyphase filter 
 *          implementation.
 * @author  Zhongren Cao <zcao@soe.ucsd.edu>
 *
 * @verbatim
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 *      02/17/2011          Zhongren Cao        
 *
 * ************************************************************************** 
 *  Copyright (c) 2011 	The Regents of the University of California
 *
 *             	Zhongren Cao (zcao@ucsd.edu)
 *				Wireless System and Engineering Group
 *				UCSD Division
 *				California Institute for Telecommunications and
 *						Information Technology (Calit2)
 *  All Rights Reserved
 *			
 *  Permission to copy, modify and distribute any part of this software for 
 *  educational, research, United States Government, and non-profit purposes, 
 *  without fee, and without a written agreement is hereby granted, provided 
 *  that the above copyright notice, this paragraph and the following three 
 *  paragraphs appear in all copies.
 *
 *  Those desiring to incorporate this software into commercial products or 
 *  use for commercial purposes should contact the Technology Transfer & 
 *  Intellectual Property Services, University of California, San Diego, 
 *  9500 Gilman Drive, Mail Code 0910, La Jolla,  CA 92093-0910,
 *  Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL: invent@ucsd.edu.
 *
 *  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA OR ANY CONTRIBUTOR BE 
 *  LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 *  CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA OR ANY
 *  CONTRIBUTORHAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  THIS SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 *  CALIFORNIA OR ANY CONTRIBUTOR HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 *  SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF CALIFORNIA 
 *  OR ANY CONTRIBUTORS MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY 
 *  KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *  WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE 
 *  USE OF THIS SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 **********************************************************************************
 * @endverbatim
 */
 
 
#ifndef __RESAMPLE_HPP__
#define __RESAMPLE_HPP__

/** 
 * The upsample_rate and downsample_rate must be positive integers.  
 */
class Resample_OP{
public:
	/**
	 * @brief The up-sampling rate and down-sampling rate of the resample operation.
	 */
	unsigned short upsample_rate, downsample_rate;
	
	/**
	 * @brief The new samples are interpolated using the old samples on each side.
	 *        The number of old samples used on each side is stored in this variable.
	 */
	unsigned short num_oneside_sample;
	
	/**
	 * @brief Resample_OP(unsigned short, unsigned short, unsigned short, size_t, float *);
	 *	
	 *	      Class constructor. Populate the upsample_rate, downsample_rate, the number
	 *        of one side samples for interpolation, the FIR filter length and coeffients.
	 */
	Resample_OP(unsigned short, unsigned short, unsigned short, size_t, float *);
	
	/**
	 * @brief The main operation function, working on each new input sample of the
	 *        sequence to be resampled. 
	 * 
	 * @param input	a new input sample of the sequence
	 * @param output pointer to the array to store the interpolated samples obtained
	 *               with the input
	 *
	 * @return the number of resample points given the single input. Depending 
	 *         on the relationship between upsample_rate and downsample_rate, for 
	 *         each sequence input, there may be zero, one, or multiple resampled 
	 *         points.
	 */
	unsigned short resample(float input, float *output);
	
	/**
	 *	@brief Prints to screen the resampling FIR filter coefficients in the 2D array 
	 *         used by the resampling object. This is for debug purposes.
	 */
	void print_2d();
	
private:
	/**
	 *	@brief The length of the FIR filter used for the resampling operation. It should
	 *	       equal to (2*num_oneside_sample*upsample_rate+1). If the length of the 
	 *	       supplied filter doesn't conform to the above formula, the class constructor
	 *	       quits without initialization.
	 */
	unsigned int fir_length;
	
	/**
	 *	@brief To reduce complexity and eliminate unnecessary computation, the resampling 
	 *	       is implemented using polyphase. Each resampled point only needs 
	 *		   2*num_oneside_sample or 2*num_oneside_sample+1 
	 *	       old sequence samples to interpolate. 
	 *
	 *         The coefficients of the supplied FIR filter is arranged in an 2-D array, 
	 *         fir_coef_twoD, with 
	 *		      num_row = upsample_rate;
	 *		      num_col = 2*num_oneside_sample+1
	 */
	unsigned short num_row, num_col;	
	float **fir_coef_twoD;
	
	/**
	 *	@brief reg is an array used by member function resample(float, float *&) to 
	 *	       remember previous input samples within the interpolation window. 
	 */
	float *reg;
	
	/**
	 *	@brief Time stamp and threshold used by the member function resample().
	 *	       i_stamp is the relative time stamp of current input; o_stamp is the relative
	 *	       time stamp of the new resampling point; if i_stamp-o_stamp > t_thsd, at 
	 *	       least one new resampled point is obtainable.
	 */
	int i_stamp, o_stamp, t_thsd;
};

# endif /*__RESAMPLE_HPP__*/