//-------------------------------------------------------------------------
/**
 * @file 	main.cpp
 * @brief	Starting point of program
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        Author              Description of Change
 * 		1.0 Feb 22 2011 Anthony Nwokafor    C++ Model Structure created
 * @endverbatim
 */ 
//-------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "controller.h"
#include "p25/format/bit_manipulate.h"
#include "p25/format/bit_stream.h"
#include "p25/util/logger.h"
#include "p25/util/sys.h"
#include "p25/comm/ipcomm.h"

#include <string.h>

//-------------------------------------------------------------------------
/**
 * @brief	Entry point of P25 GPP program
 * @details	Responsible for initializing the system components
 * 			on the GPP side and loading the Controller which in turn loads the
 * 			rest of the radio components.
 *
 * @param[in]	argc	Number of arguments passed to program
 * @param[in]	argv	arguments passed to program
 * @return				returns 0 on normal exit, -1 if error occurred
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */  
//-------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    int res;
    int ui_port;
    char* vc_sport;
    char* vc_rxfile;
    char* tx_spath;
    char* rx_spath;
    char* fc_spath;
    int r_mode;

    if (argc < 8) {
    	INFO("Usage: %s <vc55_serial_port> <vc_rx_file> <ui_port> <out_unix_sock> <in_unix_sock> <gpp_uhd_control_channel> <radio_mode>\n",argv[0]);
    	exit(1);
    }

    // Vocoder serial communication port
    vc_sport = argv[1];

    // Vocoder filetest RX file. Used only when Radio is in TEST_MODE
    vc_rxfile = argv[2];

    // graphical UI TCP communication port
    ui_port = atoi(argv[3]);

    // UNIX domain communication socket path for transmitter
    tx_spath = argv[4];

    // UNIX domain communication socket path for receiver
    rx_spath = argv[5];

    // UNIX domain communication socket path for RF frequency changes
    fc_spath = argv[6];

    // Radio mode
    r_mode = atoi(argv[7]);

    // initialize radio system
    res = sys_init(tx_spath, rx_spath, fc_spath);
    if (res < 0) {
        ERROR("sys init error\n");
        sys_exit();
        return -1;
    }  

    // initialize bit stream formatter
    res = bs_init();
    if (res < 0) {
        ERROR("bs init error\n");
        sys_exit();
        return -1;
    }

    // Instantiate Controller and hand over control of radio
    Controller cntrl(ui_port, vc_sport, vc_rxfile, r_mode);
    cntrl.start();

    return 0;
}
