//-------------------------------------------------------------------------
/**
 * @file 	ui.h
 * @brief	UserInterface Class Definition
 *
 * @author	Cal(it)2
 * @author  Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 *
 * Change History:
 * 		Rev Date        	Author              Description of Change
 * @endverbatim
 */
//-------------------------------------------------------------------------

#ifndef _USERINTERFACE_H
#define _USERINTERFACE_H

#include <pthread.h>
#include "p25/config.h"
#include "controller.h"

class UserInterface
{
private:
	/** @brief	File descriptor that */
	pthread_t		uiThread;

	/** @brief	Pointer to Controller object that created instance of this class. */
	Controller*		mController;


	/**
	 * @brief 	Graphical User Interface (GUI) communication port.
	 * @details Current implementation uses a TCP socket.
	 */
	int		ui_port;

	/** @brief	File descriptor for incoming TCP connections. */
	int		ui_connFd;

	/** @brief	File descriptor for TCP server socket. */
	int		ui_sock;

	/** @brief	Keeps track of whether or not a GUI is connected. */
	int		ui_connected;

	
	/** @brief	Receives data from GUI via TCP socket connection. */
	int		receive_data(char* buf, int len);

	/** @brief	Sends data to GUI via TCP socket connection. */
	void	send_data(char* buf, int len);

	/** @brief	Updates p25 waveform properties given input waveform property struct. */
	void	update(UInterface ui_cntrl);

	/** @brief	Prepares the UserInterface instance for GUI communication. */
	void	initialize();

	/** @brief	Gracefully ends GUI communication and shuts down the UserInterface instance. */
	void	shutdown();
public:
	/** @brief	Constructor for UserInterface Class. */
	UserInterface(Controller* mController, int ui_port);

	/** @brief	Begin listening for GUI connections and relaying p25 waveform properties. */
	void	run_ui();

	/** @brief	Waits for main loop thread termination. */
	void	join();

	/** @brief	Updates cache of received p25 waveform properties decoded from received 
				p25 frames and sends a copy to the GUI.
	 */
	void	r_ui_update(void);

	/** @brief	Starts main loop thread. */
	static void* start_thread(void *obj);
};

#endif
