/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#include <string>
#include <iostream>
#include <stdio.h>
#include "ModemDevice.h"

/* user includes */
#include "p25/format/bit_stream.h"
#include "p25/format/packets.h"
#include "p25/util/sys.h"

const char id[] = "ModemDevice";

#define INFO(fmt, args...)							\
	fprintf(stdout, "%20s: [info] %s() " fmt, id, __FUNCTION__, ##args); \
	fflush(stdout);

#define ERROR(fmt, args...)							\
	fprintf(stderr, "%20s: [error] %s() " fmt, id, __FUNCTION__, ##args); \
	fflush(stderr);

/**
 * @details Modem class constructor for OSSIE implementation.
 *
 * @author
 */
ModemDevice_i::ModemDevice_i(const char *uuid, omni_condition *condition) :
    Resource_impl(uuid), component_running(condition)
{
    dataIn_0 = new standardInterfaces_i::realChar_p("modem_in");
    dataOut_0 = new standardInterfaces_i::realChar_u("modem_out");
    ui_dataIn = new standardInterfaces_i::realChar_p("ui_in");
    ui_dataOut = new standardInterfaces_i::realChar_u("ui_out");
//    start();
}

/**
 * @details Modem class destructor.
 */
ModemDevice_i::~ModemDevice_i(void)
{
    delete dataIn_0;
    delete dataOut_0;
}

// Static function for omni thread
void ModemDevice_i::Run( void * data )
{
    ((ModemDevice_i*)data)->ProcessData();
}

// Static function for omni thread
void ModemDevice_i::bs_Run( void * data )
{
    ((ModemDevice_i*)data)->bs_ProcessData();
}

void ModemDevice_i::ui_Run( void * data )
{
	((ModemDevice_i*)data)->ui_ProcessData();
}

CORBA::Object_ptr ModemDevice_i::getPort( const char* portName ) throw (
    CORBA::SystemException, CF::PortSupplier::UnknownPort)
{
    DEBUG(3, ModemDevice, "getPort() invoked with " << portName);

    CORBA::Object_var p;

    p = dataIn_0->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = dataOut_0->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = ui_dataIn->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = ui_dataOut->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    /*exception*/
    throw CF::PortSupplier::UnknownPort();
}

/**
 * @details Initiates socket connections for UHD, bit streams, sets the stack size, allocates
 *			buffers and starts all threads for Modem component.
 *
 * @author
 */
void ModemDevice_i::start() throw (CORBA::SystemException,
    CF::Resource::StartError)
{
	int res;

	char tx_spath[] = "dsp_gpp_sock\0";
	char rx_spath[] = "gpp_dsp_sock\0";
	char fc_spath[] = "uhd_ctrl_sock\0";

    chdir("/home/sdr-overo/dev/usrp_br/run");

    // initialize radio system
    res = sys_init(tx_spath, rx_spath, fc_spath);
    if (res < 0) {
        ERROR("sys init error\n");
        sys_exit();
        stop();
        //return -1;

        throw CF::Resource::StartError(CF::CFNOTSET, "sys init error.\n");
    }

    // initialize bit stream formatter
    res = bs_init();
    if (res < 0) {
        //ERROR("bs init error\n");
        sys_exit();
        stop();
        //return -1;

        throw CF::Resource::StartError(CF::CFNOTSET, "bs init error.\n");
    }

    omni_thread::stacksize(2*1024*1024);

	omni_mutex_lock  l(processing_mutex);
	if( false == thread_started )
	{
		thread_started = true;
		// Create the thread for the writer's processing function
		processing_thread = new omni_thread(Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		processing_thread->start();
	}

	omni_mutex_lock  ll(bs_processing_mutex);
	if( false == bs_thread_started )
	{
		bs_thread_started = true;
		// Create the thread for the writer's processing function
		bs_processing_thread = new omni_thread(bs_Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		bs_processing_thread->start();
	}

	omni_mutex_lock  lll(ui_processing_mutex);
	if( false == ui_thread_started )
	{
		ui_thread_started = true;
		// Create the thread for the writer's processing function
		ui_processing_thread = new omni_thread(ui_Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		ui_processing_thread->start();
	}
}

/**
 * @details Stops all threads, frees buffers, and stop bit streams.
 *
 * @author
 */
void ModemDevice_i::stop() throw (CORBA::SystemException, CF::Resource::StopError)
{
    DEBUG(3, ModemDevice, "stop() invoked");
	omni_mutex_lock l(processing_mutex);
	thread_started = false;

	omni_mutex_lock ll(bs_processing_mutex);
	bs_thread_started = false;

	bs_exit();
	sys_exit();
}

/**
 * @details
 * @author
 */
void ModemDevice_i::releaseObject() throw (CORBA::SystemException,
    CF::LifeCycle::ReleaseError)
{
    DEBUG(3, ModemDevice, "releaseObject() invoked");

    component_running->signal();
}

/**
 * @details
 * @author
 */
void ModemDevice_i::initialize() throw (CF::LifeCycle::InitializeError,
    CORBA::SystemException)
{
    DEBUG(3, ModemDevice, "initialize() invoked");
}

/**
 * @details
 * @author
 */
void ModemDevice_i::query( CF::Properties & configProperties ) throw (CORBA::SystemException, CF::UnknownProperties)
{
	if( configProperties.length() == 0 )
	{
		configProperties.length( propertySet.length() );
		for( int i = 0; i < propertySet.length(); i++ )
		{
			configProperties[i].id = CORBA::string_dup( propertySet[i].id );
			configProperties[i].value = propertySet[i].value;
		}
		return;
	} else {
		for( int i = 0; i < configProperties.length(); i++ ) {
			for( int j = 0; j < propertySet.length(); j++ ) {
				if( strcmp(configProperties[i].id, propertySet[j].id) == 0 ) {
					configProperties[i].value = propertySet[j].value;
				}
			}
		}
	} // end if-else
}

/**
 * @details
 * @author
 */
void ModemDevice_i::configure(const CF::Properties& props)
throw (CORBA::SystemException,
    CF::PropertySet::InvalidConfiguration,
    CF::PropertySet::PartialConfiguration)
{
    DEBUG(3, ModemDevice, "configure() invoked");

    static int init = 0;
    if( init == 0 ) {
        if( props.length() == 0 ) {
            std::cout << "configure called with invalid props.length - " << props.length() << std::endl;
            return;
        }
        propertySet.length(props.length());
        for( int j=0; j < props.length(); j++ ) {
            propertySet[j].id = CORBA::string_dup(props[j].id);
            propertySet[j].value = props[j].value;
        }
        init = 1;
    }

    std::cout << "props length : " << props.length() << std::endl;

    for ( int i = 0; i <props.length(); i++)
    {
        std::cout << "Property id : " << props[i].id << std::endl;

    }
}

/**
 * @details Gets data from TX device, and places in FIFO for further processing by 
 *			baseband and radio components.
 *
 * @author Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
void ModemDevice_i::ProcessData()
{
    DEBUG(3, ModemDevice, "ProcessData() invoked");

    int ret;
    char *buf;

	//PortTypes::CharSequence buffer_out;
    PortTypes::CharSequence *I_in_0(NULL);
    CORBA::UShort I_in_0_length;

    while(continue_processing())
    {
        dataIn_0->getData(I_in_0);
        I_in_0_length = I_in_0->length();

        // Put received data into output FIFO
        if (I_in_0_length > 0) {
        	buf = (char *)(I_in_0->get_buffer());
			ret = bs_output(buf, I_in_0_length);
        }
        dataIn_0->bufferEmptied();
    }
}

/**
 * @details
 */
bool ModemDevice_i::continue_processing()
{
	omni_mutex_lock l(processing_mutex);
	return thread_started;
}

/**
 * @details Receives data from input FIFO from radio, and pushes out to RX component.
 *
 * @author Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
void ModemDevice_i::bs_ProcessData()
{
    DEBUG(3, ModemDevice, "bs_ProcessData() invoked");

    int res;
    char buf[BLOB_LEN8];
	PortTypes::CharSequence_var buffer_var;

    while(bs_continue_processing())
    {
        res = bs_input(buf, BLOB_LEN8);
//        INFO("Received %d bytes\n", res);
        /*insert code here to do work*/
		if (res > 0 ) {
		    //INFO("prepare buffer_var\n");
			buffer_var = new PortTypes::CharSequence(res, res, (CORBA::Char *)buf, 0);

			//INFO("pushPacket\n");
			dataOut_0->pushPacket(buffer_var);
		if (res != BLOB_LEN8)
			ERROR("res != BLOB_LEN8\n");
		}
    }
}

/**
 * @details
 */
bool ModemDevice_i::bs_continue_processing()
{
	omni_mutex_lock ll(bs_processing_mutex);
	return bs_thread_started;
}

/**
 * @details Receives control information from UI component. Control signals that can be
 *			received are a frequency change request, a request to flush the output FIFO buffer,
 *			and a request to flush or recet the input FIFO (although currently not used).
 *
 * @see 	set_fm_fc()
 * @see 	bs_output_flush()
 * @see		ModemPacket
 * 
 * @author  Jeffrey Cuenco <jcuenco@ucsd.edu>
 * @author 	Justyn Bell <jwbell@ucsd.edu>
 */
void ModemDevice_i::ui_ProcessData()
{
    DEBUG(3, ModemDevice, "ProcessData() invoked");

	//PortTypes::CharSequence buffer_out;
    PortTypes::CharSequence *ui_in(NULL);
    CORBA::UShort ui_in_length;
	ModemPacket *mod_pkt;
	
    double cfreq = 0.0;

    while(ui_continue_processing())
    {
    	ui_dataIn->getData(ui_in);
    	ui_in_length = ui_in->length();

    	mod_pkt = (ModemPacket *)(ui_in->get_buffer());

		switch(mod_pkt->hdr) {
		case FREQ_CHANGE:
			cfreq = mod_pkt->freq;
    		INFO("Got new frequency of %5.4f MHz!\n", cfreq / 1e6);
			set_fm_fc(cfreq);
			break;
		case BS_OUTPUT_FLUSH:
			INFO("Triggered bs_output_flush\n");
			bs_output_flush();
			break;
		case BS_INPUT_FLUSH:
			break;
		default:
			INFO("Unknown modem packet header\n");
			break;
		}
    	ui_dataIn->bufferEmptied();
    }
}

/**
 * @details
 */
bool ModemDevice_i::ui_continue_processing()
{
	omni_mutex_lock lll(ui_processing_mutex);
	return ui_thread_started;
}
