/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

//#define _DEBUG
#include <string>
#include <iostream>
#include "OveroVocoderDevice.h"

/* user includes */
#include "p25/config.h"
#include "p25/controller.h"
#include "p25/format/packets.h"
#include "p25/util/logger.h"
#include "p25/util/ticktock.h"
#include "VcwBuffers.h"
#include "host_gpp_intf.h"
#include "cntrl_types.h"


extern "C" {
#include "p25/format/bit_fifo.h"
}

//#define USE_ENCFILE
//#define AUDIO_DEC_DUMP

#define CRYPT_ENCRYPTION        0x0000CCFF
#define CRYPT_DECRYPTION        0x0000CCAA
#define CRYPT_END               0x0000AABB

//-------------------------------------------------------------------------------------
/**
 * @details Vocoder class constructor for OSSIE implementation and Overo board.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
OveroVocoderDevice_i::OveroVocoderDevice_i(const char *uuid, omni_condition *condition) :
    Resource_impl(uuid), component_running(condition)
{
    rx_in_data = new standardInterfaces_i::realChar_p("rx_in");        // Encoded data from rx comp
    tx_out_data = new standardInterfaces_i::realChar_u("tx_out");      // Encoded data  to  tx comp
    rx_out_data = new standardInterfaces_i::realChar_u("rx_out");      // Data  to  rx comp
    tx_in_data = new standardInterfaces_i::realChar_p("tx_in");        // Data from tx comp
    au_dataIn = new standardInterfaces_i::complexShort_p("audio_in");  // Audio from Ossie audio device
    au_dataOut = new standardInterfaces_i::complexShort_u("audio_out");// Audio  to  Ossie audio device
    encrypt_out = new standardInterfaces_i::realChar_u("encrypt_out"); // Data  to  Encryption comp
    encrypt_in = new standardInterfaces_i::realChar_p("encrypt_in");   // Data from encryption comp
    //start();

    // LCM of 64 and 88 is 704, just for extra buffering
    orig_bf = bf_alloc(352);
    enc_bf = bf_alloc(352);

    pthread_mutex_init (&orig_bf_m, NULL);
    pthread_mutex_init (&enc_bf_m, NULL);

    sem_init(&sem_arm_processdata_start, 0, 0);
    sem_init(&sem_arm_processdata_done, 0, 0);
    sem_init(&sem_processdata_start, 0, 0);
    sem_init(&sem_processdata_done, 0, 0);

    hg = new HostGPPInterface ("10.10.10.15", 51718); // Port for sending/receiving vocoded data
    hg_enc = new HostGPPInterface ("10.10.10.15", 51719); // Port for sending/receiving encrypted data
    mController = new Controller(0, NULL, NULL, 2);

    // Instantiate Vocoder (soft vocoder mode)
	INFO("Creating VC object\n");
	tVc = new VcwBuffers();
}

//-------------------------------------------------------------------------------------
/**
 * @details Vocoder class constructor for OSSIE implementation with Overo board.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
OveroVocoderDevice_i::~OveroVocoderDevice_i(void)
{
    delete rx_in_data;
    delete tx_out_data;
    delete rx_out_data;
    delete tx_in_data;
    delete au_dataIn;
    delete au_dataOut;
    delete encrypt_out;
    delete encrypt_in;

    delete hg;
    delete hg_enc;
    delete tVc;
	delete mController;

	bf_free(orig_bf);
    bf_free(enc_bf);
    pthread_mutex_destroy (&orig_bf_m);
    pthread_mutex_destroy (&enc_bf_m);
}

//-------------------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::host_voc_Run( void * data )
{
    ((OveroVocoderDevice_i*)data)->host_voc_ProcessData();
}

//-------------------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::rx_in_Run( void * data )
{
    ((OveroVocoderDevice_i*)data)->rx_in_ProcessData();
}

//-------------------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::tx_in_Run( void * data )
{
    ((OveroVocoderDevice_i*)data)->tx_in_ProcessData();
}

//-------------------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::voc_host_Run( void * data )
{
    ((OveroVocoderDevice_i*)data)->voc_host_ProcessData();
}

//-------------------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::encrypt_out_Run( void * data )
{
    ((OveroVocoderDevice_i*)data)->encrypt_out_ProcessData();
}

//-------------------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::encrypt_in_Run( void * data )
{
    ((OveroVocoderDevice_i*)data)->encrypt_in_ProcessData();
}


//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
CORBA::Object_ptr OveroVocoderDevice_i::getPort( const char* portName ) throw (
    CORBA::SystemException, CF::PortSupplier::UnknownPort)
{
    DEBUG(3, OveroVocoderDevice, "getPort() invoked with " << portName)

    CORBA::Object_var p;

    p = rx_in_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = tx_out_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = rx_out_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = tx_in_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = au_dataIn->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = au_dataOut->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = encrypt_out->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = encrypt_in->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    /*exception*/
    throw CF::PortSupplier::UnknownPort();
}

//-------------------------------------------------------------------------------------
/**
 * @details Initializes buffers for vocoder, starts all threads, sets state variables
 *			for a start-up state, and establishes a connection via two TCP ports to the
 *			Overo board.
 *
 * @author Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::start() throw (CORBA::SystemException,
    CF::Resource::StartError)
{
    DEBUG(3, OveroVocoderDevice, "start() invoked");

    // Instantiate Controller
    mController->start();

    // Initialize Vocoder buffer class
    tVc->initialize();

	end_instream = 1;
	end_outstream = 1;
	end_out_encstream = 1;
	end_in_encstream = 1;
	echanbits = 88;

	mController->set_radio_activity(STANDBY_ACTIVITY);

	/* Initialize TCP connection to GPP for vocoded data */
	INFO("***Initializing TCP connection to GPP...");
	if (hg->gpp_connect() < 0) {
		throw CF::Resource::StartError(CF::CFNOTSET, "Connection to GPP failed!\n");
	}
	printf("done.\n");

	/* Initialize TCP connection to GPP for encryption loop */
	INFO("***Initializing TCP connection to GPP...");
	if (hg_enc->gpp_connect() < 0) {
		throw CF::Resource::StartError(CF::CFNOTSET, "Connection to GPP failed!\n");
	}
	printf("done.\n");

	omni_thread::stacksize(2*1024*1024);

	omni_mutex_lock  l(host_voc_processing_mutex);
	if( false == host_voc_thread_started )
	{
		host_voc_thread_started = true;
		// Create the thread for the writer's processing function
		host_voc_processing_thread = new omni_thread(host_voc_Run, (void *) this);

		// Start the thread containing the writer's processing function
		host_voc_processing_thread->start();
	}

	omni_mutex_lock  ll(rx_in_processing_mutex);
	if( false == rx_in_thread_started )
	{
		rx_in_thread_started = true;
		// Create the thread for the writer's processing function
		rx_in_processing_thread = new omni_thread(rx_in_Run, (void *) this);

		// Start the thread containing the writer's processing function
		rx_in_processing_thread->start();
	}

	omni_mutex_lock  lll(voc_host_processing_mutex);
	if( false == voc_host_thread_started )
	{
		voc_host_thread_started = true;
		// Create the thread for the writer's processing function
		voc_host_processing_thread = new omni_thread(voc_host_Run, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		voc_host_processing_thread->start();

		INFO("Started ARM data processing thread.\n");
	}

	omni_mutex_lock  llll(tx_in_processing_mutex);
	if( false == tx_in_thread_started )
	{
		tx_in_thread_started = true;
		// Create the thread for the writer's processing function
		tx_in_processing_thread = new omni_thread(tx_in_Run, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		tx_in_processing_thread->start();

		INFO("Started tx_in_in port pushPacket processing thread.\n");
	}

	omni_mutex_lock  lllll(encrypt_out_processing_mutex);
	if( false == encrypt_out_thread_started )
	{
		encrypt_out_thread_started = true;
		// Create the thread for the writer's processing function
		encrypt_out_processing_thread = new omni_thread(encrypt_out_Run, (void *) this);

		// Start the thread containing the writer's processing function
		encrypt_out_processing_thread->start();

		INFO("Started encryption loop thread. (out)\n");
	}

	omni_mutex_lock  llllll(encrypt_in_processing_mutex);
	if( false == encrypt_in_thread_started )
	{
		encrypt_in_thread_started = true;
		// Create the thread for the writer's processing function
		encrypt_in_processing_thread = new omni_thread(encrypt_in_Run, (void *) this);

		// Start the thread containing the writer's processing function
		encrypt_in_processing_thread->start();

		INFO("Started encryption loop thread. (in)\n");
	}
}

//-------------------------------------------------------------------------------------
/**
 * @details Stops all threads, and shuts down the connection to the Overo board.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::stop() throw (CORBA::SystemException, CF::Resource::StopError)
{
    DEBUG(3, OveroVocoderDevice, "stop() invoked");

    mController->set_radio_activity(STANDBY_ACTIVITY);

    // Shutdown Vocoder buffer class
    tVc->shutdown();

	omni_mutex_lock l(host_voc_processing_mutex);
	host_voc_thread_started = false;

	omni_mutex_lock ll(rx_in_processing_mutex);
	rx_in_thread_started = false;

	omni_mutex_lock lll(voc_host_processing_mutex);
	voc_host_thread_started = false;

	omni_mutex_lock llll(tx_in_processing_mutex);
	tx_in_thread_started = false;

	omni_mutex_lock lllll(encrypt_out_processing_mutex);
	encrypt_out_thread_started = false;

	omni_mutex_lock llllll(encrypt_in_processing_mutex);
	encrypt_in_thread_started = false;

	INFO("Waiting for end_instream...\n");
	while (!end_instream)
		sched_yield();
	INFO("Received end_instream = 1\n");

	/* Close TCP connection to ARM */
	INFO("***Closing TCP connection to ARM vocoder...");
	hg->gpp_disconnect();
	printf("done.\n");

	/* Close TCP connection to ARM */
	INFO("***Closing TCP connection to ARM encryption loop...");
	hg_enc->gpp_disconnect();
	printf("done.\n);");

	INFO("Waiting for end_encstream...\n");
	while (!(end_out_encstream) && !(end_in_encstream))
		sched_yield();
	INFO("Received end_encstream = 1\n");

}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::releaseObject() throw (CORBA::SystemException,
    CF::LifeCycle::ReleaseError)
{
    DEBUG(3, OveroVocoderDevice, "releaseObject() invoked")

    component_running->signal();
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::initialize() throw (CF::LifeCycle::InitializeError,
    CORBA::SystemException)
{
    DEBUG(3, OveroVocoderDevice, "initialize() invoked")
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::query( CF::Properties & configProperties ) throw (CORBA::SystemException, CF::UnknownProperties)
{
	if( configProperties.length() == 0 )
	{
		configProperties.length( propertySet.length() );
		for( int i = 0; i < propertySet.length(); i++ )
		{
			configProperties[i].id = CORBA::string_dup( propertySet[i].id );
			configProperties[i].value = propertySet[i].value;
		}
		return;
	} else {
		for( int i = 0; i < configProperties.length(); i++ ) {
			for( int j = 0; j < propertySet.length(); j++ ) {
				if( strcmp(configProperties[i].id, propertySet[j].id) == 0 ) {
					configProperties[i].value = propertySet[j].value;
				}
			}
		}
	} // end if-else
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::configure(const CF::Properties& props)
throw (CORBA::SystemException,
    CF::PropertySet::InvalidConfiguration,
    CF::PropertySet::PartialConfiguration)
{
    DEBUG(3, OveroVocoderDevice, "configure() invoked")

    static int init = 0;
    if( init == 0 ) {
        if( props.length() == 0 ) {
            std::cout << "configure called with invalid props.length - " << props.length() << std::endl;
            return;
        }
        propertySet.length(props.length());
        for( int j=0; j < props.length(); j++ ) {
            propertySet[j].id = CORBA::string_dup(props[j].id);
            propertySet[j].value = props[j].value;
        }
        init = 1;
    }

    std::cout << "props length : " << props.length() << std::endl;

    for ( int i = 0; i <props.length(); i++)
    {
        std::cout << "Property id : " << props[i].id << std::endl;

    }
}

//-------------------------------------------------------------------------------------
/**
 * @details	Handles interaction from OSSIE host side Vocoder component to the Overo board
 *			implementation of the IMBE vocoder.  When a start is received from either the 
 *			TX device or RX device, the mode is set and the thread is posted.  It starts
 *			by sending a start control message to the OSSIE Encryption component if encryption
 *			is requested, followed by a start control message to the Overo board depending on what 
 *			state the OSSIE Vocoder component is in (either TX or RX) and the state of 
 *			encryption (either on or off, based on the algorithm ID).
 *			In TX mode a start is sent followed by a packet containing a DATA control
 *			message and a payload of the actual voice codeword data received from the 
 *			Sound_in component. It will continue sending these voice codewords until the 
 *			radio is switched to a different state, and 18 voice code words have been sent
 *			since the vocoder on the Overo board cannot finish without receiving all 18 VCWs.
 *			The thread will then send another control packet with a STOP header to notify 
 *			the Overo that the radio has completed a transmission.
 *			In RX mode, much of the same processes will occer, with only variations in 
 *			what control messages are sent, and where the voice codewords come from.
 *			In the RX case, voice codewords are pulled from the vcw playback buffer which
 *			is based upon a producer<->consumer model, and that is filled with voice 
 *			codeword data received from the OSSIE receiver component instead of an OSSIE
 *			sound device.
 *			
 *
 * @see		IMBE_ENCODE_START
 * @see		IMBE_ENCR_ENCODE_START
 * @see		IMBE_DECODE_START
 * @see		IMBE_DECODE_DECR_START
 * @see 	IMBE_ENCODE_STOP
 * @see		IMBE_DECODE_STOP
 *
 * @author	Justyn Bell	<jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::host_voc_ProcessData()
{
    DEBUG(3, OveroVocoderDevice, "host_voc_ProcessData() invoked")

	// Port variables for TX mode
	PortTypes::ShortSequence *au_in_L(NULL), *au_in_R(NULL);
	CORBA::UShort au_in_L_length, au_in_R_length;

    int len;
    unsigned short vcw_count;

    while(host_voc_continue_processing()) {
    	DBG("Waiting on sem_processdata_start post\n");
    	sem_wait(&sem_processdata_start);

    	ambe_stop_sent = false;

		if ((mController->get_radio_activity() == TX_ACTIVITY)) {
			struct e_host_gpp_pkt pkt;

			// Open a test file for dropping encoded data into
			#ifdef USE_ENCFILE
				testfile_enc = fopen("testfile_enc","w");
				if (testfile_enc == NULL)
					printf("Can't open file to write IMBE-encoded PCM data\n");
			#endif

			// Send IMBE encode start command to Overo
			if (tx_algid == 0x80) {
			DBG("IMBE_ENCODE_START\n");
			} else {
				// If in encrypted mode, send a start to the encryption component
				DBG("IMBE_ENCR_ENCODE_START\n");
				DBG("Sending CRYPTO_PKT_CNTRL_START to encryption device\n");
				sendbuf_vocoder_crypt(CRYPTO_PKT_CNTRL_START);
			}

			pkt.ctrl = ((tx_algid == 0x80) ? IMBE_ENCODE_START : IMBE_ENCR_ENCODE_START);

			if (hg->write_data_payload((char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0) {
				ERROR("Error writing control start to GPP!\n");
			}

			while (mController->get_radio_activity() == TX_ACTIVITY) {
				vcw_count = 0;
				do {
					au_dataIn->getData(au_in_L, au_in_R);
					au_in_L_length = au_in_L->length();
					au_in_R_length = au_in_R->length();

					// pkt.payload is in bytes, not shorts
					if (au_in_L_length <= BUFF_LEN)
						len = au_in_L_length;
					else
						len = BUFF_LEN;

					char *lbuf = (char *)(au_in_L->get_buffer());
					char *rbuf = (char *)(au_in_R->get_buffer());

					for (int i = 0; i < len/2; i++) {
						((short *)pkt.payload)[i*2] = ((short *)lbuf)[i];
						((short *)pkt.payload)[i*2+1] = ((short *)rbuf)[i];
					}

					// Send to TCP socket
					DBG("PCM_DATA %d\n",vcw_count);
					pkt.ctrl = IMBE_DATA;

					if (hg->write_data_payload((char *)&pkt, sizeof(struct e_host_gpp_pkt)) < 0)
						ERROR("ERROR writing payload to gpp\n");

					au_dataIn->bufferEmptied();
					vcw_count++;

				// DSP needs 18 VCWs
				} while (vcw_count < 18);
			}

			// Send IMBE encode stop command to GPP
			DBG("IMBE_ENCODE_STOP\n");
			pkt.ctrl = IMBE_ENCODE_STOP;
			if (hg->write_data_payload((char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0) {
				ERROR("Error writing control stop to GPP!\n");
			}

			ambe_stop_sent = true;

			#ifdef USE_ENCFILE
				fclose(testfile_enc);
			#endif

		} else if ((mController->get_radio_activity() == RX_ACTIVITY)) {
			struct d_host_gpp_pkt pkt;
			char buf[E_BUFF_LEN/2];

			// Open a file previously written to to read from for testing
			#ifdef USE_ENCFILE
				testfile_enc = fopen("testfile_enc","r");
				if (testfile_enc == NULL)
					printf("Can't open file to read IMBE-encoded PCM data\n");
			#endif

			while (rx_algid == 0 && mController->get_radio_activity() == RX_ACTIVITY) {
				DBG("Waiting for received ALGID...\n");
				sleep(1);
			}

			vcw_count = 0;

			if (mController->get_radio_activity() == RX_ACTIVITY) {
				// Send IMBE decode start command to GPP
				pkt.ctrl = ((rx_algid == 0x80) ? IMBE_DECODE_START : IMBE_DECR_DECODE_START);

				if (pkt.ctrl == IMBE_DECODE_START) {
					DBG("IMBE_DECODE_START\n");
				} else if (pkt.ctrl == IMBE_DECR_DECODE_START) {
					// If in encrypted mode, send a start to the OSSIE encryption component
					DBG("IMBE_DECR_DECODE_START\n");
					DBG("Sending CRYPTO_PKT_CNTRL_START to encryption device (rx mode)\n");
					sendbuf_vocoder_crypt(CRYPTO_PKT_CNTRL_START);
				}

				if (hg->write_data_payload((char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0) {
					throw CF::Resource::StartError(CF::CFNOTSET, "Error writing control start to GPP!\n");
				}

				while (mController->get_radio_activity() == RX_ACTIVITY) {
				#ifdef USE_ENCFILE
					if ((res = fread(pkt.payload, sizeof(char), E_BUFF_LEN, testfile_enc)) < 0) {
						ERROR("ERROR reading PCM data from file\n");
						break;
					} else if (res == 0) {
						INFO("EOF reached, file read complete.\n");
						break;
					} else if (res != 288) {
						ERROR("ERROR reading correct amount of data from file, won't be sent to ARM!\n");
						break;
					}

					//INFO("Read %d bytes from file\n (vocoder device)", res);
					usleep(19000);
				#else

					// Get payload from playback buffer.
					int res = tVc->get_from_pb_buf(buf, E_BUFF_LEN/2);
					if (res != E_BUFF_LEN/2) {
						DBG("res = %d, E_BUFF_LEN/2 = %d\n", res, E_BUFF_LEN/2);
						break;
					}

					// Expand encoded data from 1 bit per byte to 1 bit per short as per 
					//	soft vocoder requirements
					for (int i = 0; i < E_BUFF_LEN/2; i++)
						((short *)pkt.payload)[i] = (short)buf[i];
				#endif

					DBG("Sending IMBE_DATA\n");
					pkt.ctrl = IMBE_DATA;
				
					// Write data payload to Overo
					if (hg->write_data_payload((char *)&pkt, sizeof(struct d_host_gpp_pkt)) < 0)
						ERROR("ERROR writing payload to GPP")

					vcw_count++;
				}
			}

			// Send IMBE decode stop command to GPP
			DBG("IMBE_DECODE_STOP\n");
			pkt.ctrl = IMBE_DECODE_STOP;
			if (hg->write_data_payload((char *)&pkt.ctrl, sizeof(pkt.ctrl)) < 0) {
				throw CF::Resource::StopError(CF::CFNOTSET, "Error writing control stop to GPP!\n");
			}

			ambe_stop_sent = true;

			#ifdef USE_ENCFILE
				fclose(testfile_enc);
			#endif
		}
		// Notify other threads that the host->overo thread has completed a transmission
		sem_post(&sem_processdata_done);
    }
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
bool OveroVocoderDevice_i::host_voc_continue_processing()
{
	omni_mutex_lock l(host_voc_processing_mutex);
	return host_voc_thread_started;
}

//-------------------------------------------------------------------------------------
/**
 * @details Upon receiving a start control message from the receiver component,
 *			the data payload will contain the received algorithm ID. This way 
 *			all other threads will know to treat the incoming data as either
 *			encrypted or non encrypted and process it accordingly.
 *			It will also set the radio state of the entire component and start
 *			the threads that handle both sending and receiving data to/from the Overo
 *			GPP. 
 *			When a stop control message is received from the receiver component, 
 *			the buffer that contains the VCW data is flushed, the algorithm ID
 *			is set back to an initialized value of 0, and the thread waits until
 *			the other threads that handle sending/receiving data to/from the Overo
 *			board to complete before switching the radio back into a STANDBY state.
 *			To combat the possible case of receiving a stop control message while the
 *			Vocoder component is already in a standby mode, the code will automatically
 *			send an ACK message back to the receiver component.
 *			When voice codewords are received from the receiver component, they are
 *			simply placed in a buffer for further processing in another thread. 
 *
 * @see VOCODER_PKT_CNTRL_START
 * @see VOCODER_PKT_CNTRL_STOP
 * @see VOCODER_PKT_CNTRL_DATA
 *
 * @author  Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::rx_in_ProcessData()
{
    DEBUG(3, OveroVocoderDevice, "rx_in_ProcessData() invoked")

    // Port variables for RX mode
    PortTypes::CharSequence *rx_data(NULL);
    CORBA::UShort rx_data_length;

    VocoderPacket *v_pkt;

    rx_algid = 0;

    while(rx_in_continue_processing())
    {
		rx_in_data->getData(rx_data);
		rx_data_length = rx_data->length();
		v_pkt = (VocoderPacket *)(rx_data->get_buffer());

		switch(v_pkt->hdr) {
		case VOCODER_PKT_CNTRL_START:
			DBG("Got VOCODER_PKT_CNTRL_START from rx device\n");
			rx_algid = v_pkt->data[0];
			DBG("Received Encryption ALGID: 0x%02X\n", rx_algid);

			if (mController->get_radio_activity() != RX_ACTIVITY) {
				INFO("Setting radio to RX_ACTIVITY\n");
				mController->set_radio_activity(RX_ACTIVITY);

				// Posts host_voc_ProcessData() and voc_host_ProcessData() threads
				sem_post(&sem_arm_processdata_start);
				sem_post(&sem_processdata_start);
			}
			break;

		case VOCODER_PKT_CNTRL_STOP:
			DBG("Got VOCODER_PKT_CNTRL_STOP from rx device\n");
			if (mController->get_radio_activity() == RX_ACTIVITY) {
				INFO("Setting radio to STANDBY_ACTIVITY\n");
				tVc->flush_buffer();
				rx_algid = 0;

				sem_wait(&sem_arm_processdata_done);
				sem_wait(&sem_processdata_done);
				mController->set_radio_activity(STANDBY_ACTIVITY);
			} else {
				VocoderPacket vPkt;
				PortTypes::CharSequence_var buffer_var;
				vPkt.len = 0;
				vPkt.hdr = VOCODER_PKT_CNTRL_ACK;
				buffer_var = new PortTypes::CharSequence(sizeof(vPkt), sizeof(vPkt), (CORBA::Char *)&vPkt, 0);
				rx_out_data->pushPacket(buffer_var);
			}
			break;

		case VOCODER_PKT_CNTRL_DATA:
			tVc->put_in_pb_buf(v_pkt->data, v_pkt->len);
			DBG("Received VOCODER_PKT_CNTRL_DATA from rx device\n");
			break;

		default:
			ERROR("Got an unexpected or wrong header from rx device. Value %x\n", v_pkt->hdr);
			break;
		}
		rx_in_data->bufferEmptied();
	}
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
bool OveroVocoderDevice_i::rx_in_continue_processing()
{
	omni_mutex_lock ll(rx_in_processing_mutex);
	return rx_in_thread_started;
}

//-------------------------------------------------------------------------------------
/**
 * @details Handles interaction of voice codeword data from the Overo software vocoder
 *			to the OSSIE vocoder component. Data received at this point has completed
 *			all vocoder/encryption processing and is ready to be sent either to the TX
 *			component for transmission, or the OSSIE sound card device for playback
 *			depending on the mode of the radio.
 *			It reads data packets via a TCP socket to the Overo board.  If a host 
 *			start control message is received and the radio is in TX mode, it continues
 *			to read the next packet, which is data. In TX mode, the thread converts this data
 *			from one bit per short (as per the IMBE Vocoder requirements) to one bit
 *			per byte (as per the current implementation requirements of the TX component)
 *			and push-packets this data to the OSSIE transmitter component.
 *			Once a stop request has been sent from the transmitter component, the 
 *			thread will continue reading data from the Overo board until it receives a 
 *			host stop stream control message.  At this point it will return to a standby
 *			state and wait for the thread's semaphore to be posted once again by another
 *			start from the transmitter or receiver.
 *			In RX mode, the thread works in an identical fashion except that it push-packets
 *			the data that has been decoded to the Sound_out OSSIE device.
 *
 * @see 	HOST_STREAM_DATA
 * @see 	HOST_START_STREAM
 * @see		HOST_STOP_STREAM
 *
 * @author	Justyn Bell <jwbell@ucsd.edu>
 * @author  Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::voc_host_ProcessData()
{
    DEBUG(3, OveroVocoderDevice, "voc_host_ProcessData() invoked")

	// Port variables for TX mode
    PortTypes::CharSequence_var buffer_var;

	struct e_gpp_host_pkt tx_pkt;
    struct d_gpp_host_pkt rx_pkt;
	VocoderPacket v_pkt;

	char tx_buf[E_BUFF_LEN/2];

    // Port variables for RX mode
    PortTypes::ShortSequence au_out_L, au_out_R;

	int res;

    au_out_L.length(BUFF_LEN/2); //must define length of output
    au_out_R.length(BUFF_LEN/2); //must define length of output

    while(voc_host_continue_processing()) {
    	DBG("Waiting on sem_arm_processdata_start post\n");
    	sem_wait(&sem_arm_processdata_start);
    	DBG("sem_arm_processdata_start posted\n");

		if ((mController->get_radio_activity() == TX_ACTIVITY)) {

			DBG("Waiting for HOST cntrl packet TX\n");

			// Read in control packet to determine state
			tx_pkt.ctrl = hg->read_ctrl_pkt();
			DBG("Read control packet, value %d\n", tx_pkt.ctrl);

			if (tx_pkt.ctrl == H_SOCK_ERROR) {
				ERROR("read_ctrl_pkt() failure!\n");
				break;
			}

			// If the control packet received is not HOST_START_STREAM, go to the next
			// while loop iteration (back to reading next control packet above)
			if (tx_pkt.ctrl != HOST_START_STREAM) {
				DBG("!= HOST_START_STREAM\n");
				continue;
			}

			DBG("HOST_START_STREAM TX\n");

			// Start processing data until stop encode/decode control signal received
			do {
				// Read in control packet to determine state
				DBG("Reading data from ARM\n");
				tx_pkt.ctrl = hg->read_ctrl_pkt();
				DBG("Read tx_pkt.ctrl , value %d\n", tx_pkt.ctrl);
				if (tx_pkt.ctrl == H_SOCK_ERROR) {
					ERROR("read_ctrl_pkt() failure!\n");
					break;
				}

				// If control state is HOST_STREAM_DATA, retrieve data payload
				else if (tx_pkt.ctrl == HOST_STREAM_DATA) {
					DBG("HOST_STREAM_DATA TX\n");
					end_instream = 0;

					// Read in encoded PCM data
					res = hg->read_data_payload(tx_pkt.payload, E_BUFF_LEN);
					if (res == -1) {
						ERROR("read_data_payload() failure!\n");
						break;
					}
					else if (res != E_BUFF_LEN) {
						ERROR("res = %d, E_BUFF_LEN = %d\n", res, E_BUFF_LEN);
						continue;
					}
					// Write to file for debugging
				#ifdef USE_ENCFILE
					if (fwrite(tx_pkt.payload, sizeof(char), res, testfile_enc) < 0)
						printf("ERROR writing encoded data to file \n");

					fflush(testfile_enc);
				#else
					// Compress encoded data from 1 bit per short to 1 bit per byte
					for (int i = 0; i < E_BUFF_LEN/2; i++)
						tx_buf[i] = (char)((short *)tx_pkt.payload)[i];

					// Send data to TX component
					DBG("Writing DATA to TX device\n");
					sendbuf_vocoder_tx(VOCODER_PKT_CNTRL_DATA, tx_buf, E_BUFF_LEN/2);
				#endif
				}
			} while (!((tx_pkt.ctrl == HOST_STOP_STREAM) && ambe_stop_sent));

			end_instream = 1;

			DBG("HOST_STOP_STREAM\n");
			sendbuf_vocoder_tx(VOCODER_PKT_CNTRL_ACK);
			DBG("Sent VOCODER_PKT_CNTRL_ACK to tx device\n");

		} else if ((mController->get_radio_activity() == RX_ACTIVITY)) {
			DBG("Waiting for HOST cntrl packet RX\n");

			// Read in control packet to determine state
			rx_pkt.ctrl = hg->read_ctrl_pkt();
			DBG("Read control packet, value %d\n", rx_pkt.ctrl);

			if (rx_pkt.ctrl == H_SOCK_ERROR) {
				ERROR("read_ctrl_pkt() failure!\n");
				break;
			}

			// If the control packet received is not HOST_START_STREAM, go to the next
			// while loop iteration (back to reading next control packet above)
			if (rx_pkt.ctrl != HOST_START_STREAM)
				continue;

			DBG("HOST_START_STREAM RX\n");

			#ifdef AUDIO_DEC_DUMP
				FILE *testfile_dec = fopen("audio_dec.dat","w");
				if (testfile_dec == NULL)
					printf("Can't open file to write audio data\n");
			#endif

			// Start processing data until stop encode/decode control signal received
			do {
				// Read in control packet to determine state
				rx_pkt.ctrl = hg->read_ctrl_pkt();
				DBG("Read control packet from ARM, value %d\n", rx_pkt.ctrl);
				if (rx_pkt.ctrl == H_SOCK_ERROR) {
					ERROR("read_ctrl_pkt() failure!\n");
					break;
				}

				// If control state is HOST_STREAM_DATA, retrieve data payload
				else if (rx_pkt.ctrl == HOST_STREAM_DATA) {
					DBG("HOST_STREAM_DATA RX\n");
					end_instream = 0;

					// Read in decoded PCM data
					res = hg->read_data_payload(rx_pkt.payload, BUFF_LEN);
					if (res == -1) {
						ERROR("read_data_payload() failure!\n");
						break;
					}
					else if (res != BUFF_LEN) {
						ERROR("res = %d, BUFF_LEN = %d\n", res, BUFF_LEN);
						continue;
					}
					DBG("Read %d characters \n", res);

					#ifdef AUDIO_DEC_DUMP
						// Write data to file for debugging
						if ((res = fwrite(rx_pkt.payload, sizeof(char), BUFF_LEN, testfile_dec)) < 0) {
							ERROR("Error writing PCM data to file\n");
						}
					#endif

					// Push Packet
					for (int i = 0; i < BUFF_LEN/2; i++) {
						au_out_L[i] = ((short *)rx_pkt.payload)[i];
						au_out_R[i] = ((short *)rx_pkt.payload)[i];
					}

					au_dataOut->pushPacket(au_out_L, au_out_R);
				}
			} while (!((rx_pkt.ctrl == HOST_STOP_STREAM) && ambe_stop_sent));

			v_pkt.len = 0;
			v_pkt.hdr = VOCODER_PKT_CNTRL_ACK;

			DBG("Pushing Vocoder ACK to receiver device\n");
			buffer_var = new PortTypes::CharSequence(sizeof(v_pkt), sizeof(v_pkt), (CORBA::Char *)&v_pkt, 0);
			rx_out_data->pushPacket(buffer_var);

			DBG("HOST_STOP_STREAM in RX\n");

			#ifdef AUDIO_DEC_DUMP
				fclose(testfile_dec);
			#endif
			end_instream = 1;
    	}

		sem_post(&sem_arm_processdata_done);
	}
	DBG("END out_ProcessData() thread.\n");
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
bool OveroVocoderDevice_i::voc_host_continue_processing()
{
	omni_mutex_lock lll(voc_host_processing_mutex);
	return voc_host_thread_started;
}

//-------------------------------------------------------------------------------------
/**
 * @details With encryption active, this thread reads data from the Overo vocoder
 *	 		via a TCP socket. It writes packets appended with a data control message
 *			to the OSSIE encryption component until it receives an encryption finished
 *			control message from the Overo GPP. When it does, it writes a stop 
 *			control message to the Encryption component. NOTE: a start control message is
 *			not sent to the encryption component from this thread, instead it is sent from
 *			the host_voc_ProcessData() thread when encryption is active.
 *
 * @see		CRYPTO_PKT_CNTRL_DATA
 * @see		CRYPTO_PKT_CNTRL_STOP
 * @see		CRYPT_END
 * @see		CRYPT_ENCRYPTION
 * @see		CRYPT_DECRYPTION
 *
 * @author
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::encrypt_out_ProcessData()
{
    DEBUG(3, OveroEncryptionDevice, "ProcessData() invoked")

	PortTypes::CharSequence_var buffer_var;

    int mode;
    short chan_buf[E_BUFF_LEN];
    char tmp[88], encrypt_data_out[echanbits/8];;
    int echanbits = 88;
    int echanbytes = echanbits/8;
    int idx, count = 1;

    end_out_encstream = 0;

    while(encrypt_out_continue_processing())
    {
		/* Reads data payload from GPP */
		DBG("Reading mode from ARM...");
		int res = hg_enc->read_data_payload((char *)&mode, sizeof(mode));
		DBG("done! (mode = %d)\n", mode);
		if (res == -1) {
			ERROR("read_data_payload() failure!\n");
			break;
		}
		else if (res != sizeof(mode)) {
			ERROR("res = %d, sizeof(mode) = %d\n", res, sizeof(mode));
		}

		switch (mode) {
			case CRYPT_ENCRYPTION:
			case CRYPT_DECRYPTION:
				res = hg_enc->read_data_payload((char *)&chan_buf, echanbits*sizeof(short));
				if (res == -1) {
					ERROR("read_data_payload() failure!\n");
					break;
				}

				else if (res == echanbits*sizeof(short)) {
					for (int i = 0; i < echanbits; i++)
						tmp[i] = (char)(chan_buf[i]);

					// Pack bytes
					memset(encrypt_data_out, 0, echanbytes);
					for (int i = 0; i < echanbits; i++) {
						idx = echanbits-1-i;
						encrypt_data_out[i/8] = encrypt_data_out[i/8] | tmp[i] << (idx%8);
					}

					sendbuf_vocoder_crypt(CRYPTO_PKT_CNTRL_DATA, encrypt_data_out, echanbytes);
					DBG("Sent voice code word %d\n", count);
					
					if (++count > 18)
						count = 1;
				}
				else
					ERROR("res = %d, echanbits*sizeof(short) = %d\n", res, echanbits*sizeof(short));

				break;
			case CRYPT_END:
				DBG("Sending CRYPTO_PKT_CNTRL_STOP to Encryption Device\n");
				sendbuf_vocoder_crypt(CRYPTO_PKT_CNTRL_STOP);
				break;

			default:
				break;
		}
    }

    end_out_encstream = 1;
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
bool OveroVocoderDevice_i::encrypt_out_continue_processing()
{
	omni_mutex_lock lllll(encrypt_out_processing_mutex);
	return encrypt_out_thread_started;
}

//-------------------------------------------------------------------------------------
/**
 * @details	With encryption active, this thread reads encrypted/decrypted data from
 *			the OSSIE encryption component and writes it back down to the Overo
 *			vocoder for FEC (if in TX mode) or decoding (if in RX mode).
 *			There are no control messages received from the encryption device since
 *			it is assumed that the OSSIE encryption component has processed
 *			all of the data sent to it, and will go into standby mode based on a stop
 *			control message sent to it via the encrypt_out_ProcessData() thread.
 *			The Overo vocoder also needs no control message in this implementation
 *			since it only completes when the amount of data that is sent after
 *			all of its processing is equal to the amount of data it received from the
 *			vocoder component to be processed.
 *
 * @author	Justyn Bell <jwbell@ucsd.edu>
 * @author 	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::encrypt_in_ProcessData()
{
    DEBUG(3, OveroEncryptionDevice, "ProcessData() invoked")

	PortTypes::CharSequence *encrypt_data_in(NULL);
	CORBA::UShort encrypt_data_in_length;

    char *buf, tmp[88];
    short e_chan_buf[176];
    int echanbits = 88;
    int idx;

    end_in_encstream = 0;

	//memset(e_chan_buf, 0, 176*sizeof(short));

    while(encrypt_in_continue_processing())
    {
			// Read data in from encryption device
    		encrypt_in->getData(encrypt_data_in);
            encrypt_data_in_length = encrypt_data_in->length();

            buf = (char *)(encrypt_data_in->get_buffer());

            // Expand bits of encrypted data in to 1 bit per byte (11 bytes -> 88 bytes)
            memset(tmp, 0, echanbits);
            for (int i = 0; i < echanbits; i++) {
            	idx = echanbits-1-i;
            	tmp[i] = (buf[i/8] & (1 << (idx%8))) >> (idx%8);
            }

    		// Expand from 1 bit per byte to 1 bit per short (88 bytes -> 176 bytes)
			for (int i = 0; i < echanbits; i++)
				e_chan_buf[i] = (short)(tmp[i]);

			/* Write data payload to ARM */
			int res = hg_enc->write_data_payload((char *)&e_chan_buf, echanbits*sizeof(short));
			if (res != echanbits*sizeof(short))
				ERROR("Tried to write %d to arm, but only wrote %d\n", echanbits*sizeof(short), res);

			encrypt_in->bufferEmptied();
    }
    end_in_encstream = 1;
}

//-------------------------------------------------------------------------------------
/**
 * @details Receiver class constructor for OSSIE implementation.
 *
 * @author
 */
//-------------------------------------------------------------------------------------
bool OveroVocoderDevice_i::encrypt_in_continue_processing()
{
	omni_mutex_lock llllll(encrypt_in_processing_mutex);
	return encrypt_in_thread_started;
}

//-------------------------------------------------------------------------------------
/**
 * @details Upon receiving a start control message from the transmitter component,
 *			the data payload will contain the received algorithm ID. This way 
 *			all other threads will know to treat the incoming data as either
 *			encrypted or non encrypted and process it accordingly.
 *			It will also set the radio state of the entire component and start
 *			the threads that handle both sending and receiving data to/from the Overo
 *			GPP. 
 *			When a stop control message is received from the transmitter component, 
 *			the algorithm ID is set back to an initialized value of 0, and the thread 
 *			waits until	the other threads that handle sending/receiving data to/from the Overo
 *			board to complete before switching the radio back into a STANDBY state.
 *			To combat the possible case of receiving a stop control message while the
 *			Vocoder component is already in a standby mode, the code will automatically
 *			send an ACK message back to the receiver component.
 *			NOTE: The transmitter does not send voice codeword data, so there is never
 *			a packet received from the transmitter component with a data control message.
 *			Instead, when the transmitter sets the component activity to TX and starts
 *			the host_voc_ProcessData() thread, data is automatically read via push-packets
 *			from the OSSIE Sound_in device. 
 *
 * @see VOCODER_PKT_CNTRL_START
 * @see VOCODER_PKT_CNTRL_STOP
 *
 * @author  Justyn Bell <jwbell@ucsd.edu>
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::tx_in_ProcessData()
{
    DEBUG(3, OveroVocoderDevice, "tx_in_ProcessData() invoked");

    PortTypes::CharSequence *tx_in(NULL);
    CORBA::UShort tx_in_length;

    VocoderPacket *v_pkt;

    // For now, this thread takes care of reading in data that the TX device sends
    // such as headers for starting and stopping.
    while(tx_in_continue_processing())
    {
		tx_in_data->getData(tx_in);
		tx_in_length = tx_in->length();
		v_pkt = (VocoderPacket *)(tx_in->get_buffer());

		switch(v_pkt->hdr) {
			case VOCODER_PKT_CNTRL_START:
				DBG("Received a VOCODER_PKT_CNTRL_START from Tx device\n");
				tx_algid = v_pkt->data[0];
				DBG("Got algid %x\n", tx_algid);

				if (mController->get_radio_activity() != TX_ACTIVITY) {
					INFO("Setting radio to TX_ACTIVITY\n");
					mController->set_radio_activity(TX_ACTIVITY);
					sem_post(&sem_arm_processdata_start);
					sem_post(&sem_processdata_start);
				}

				break;
			case VOCODER_PKT_CNTRL_STOP:
				DBG("Received a VOCODER_PKT_CNTRL STOP from Tx device\n");
				if (mController->get_radio_activity() == TX_ACTIVITY) {
					INFO("Setting radio to STANBY_ACTIVITY\n");
					//tVc->flush_buffer();
					tx_algid = 0;

					mController->set_radio_activity(STANDBY_ACTIVITY);

					sem_wait(&sem_arm_processdata_done);
					sem_wait(&sem_processdata_done);

				} else {
					// Ignore the stop if we're alread in STANDBY_ACTIVITY and send ack back
					sendbuf_vocoder_tx(VOCODER_PKT_CNTRL_ACK);
				}
				break;
			default:
				ERROR("Unknown or wrong header from the Tx device.  Value: %x\n", v_pkt->hdr);
				continue;
			}

        tx_in_data->bufferEmptied();
    }
}

//-------------------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------------------
bool OveroVocoderDevice_i::tx_in_continue_processing()
{
	omni_mutex_lock llll(tx_in_processing_mutex);
	return tx_in_thread_started;
}

//-------------------------------------------------------------------------------------
/**
 * @details Simple function responsible for pushing packets from the vocoder device
 *			to the transmitter device.
 *
 * @param 	hdr		Control message header appended to data packet.
 * @param	data	Data payload to be sent. Defaults to NULL if no data is
 *					specified.
 * @param	len		Length of data in bytes to be sent. Defaults to 0 if no
 *					data is specified.
 *
 * @author	Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::sendbuf_vocoder_tx(short hdr, char *data, int len)
{
	DBG("Sending %d bytes of data to tx device", len);
	VocoderPacket v_pkt;

	v_pkt.hdr = hdr;
	v_pkt.len = ((data == NULL) ? 0 : len);

	memcpy((char *)&v_pkt.data, data, len);
	DBG("Sending cntrl 0x%04X, len %d\n", v_pkt.hdr, v_pkt.len);

	PortTypes::CharSequence_var buffer_var;
	buffer_var = new PortTypes::CharSequence(sizeof(VocoderPacket), sizeof(VocoderPacket),
			(CORBA::Char *)&v_pkt, 0);

	tx_out_data->pushPacket(buffer_var);
}

//-------------------------------------------------------------------------------------
/**
 * @details Simple function responsible for pushing packets from the vocoder device
 *			to the cryptography device.
 *
 * @param 	hdr		Control message header appended to data packet.
 * @param	data	Data payload to be sent. Defaults to NULL if no data is 
 *					specified.
 * @param	len		Length of data in bytes to be sent. Defaults to 0 if no
 *					data is specified.
 *
 * @author	Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------------------
void OveroVocoderDevice_i::sendbuf_vocoder_crypt(short hdr, char *data, int len)
{
	DBG("Sending %d bytes of data to encryption device", len);
	CryptVocPacket cv_pkt;

	cv_pkt.hdr = hdr;
	cv_pkt.len = ((data == NULL) ? 0 : len);

	memcpy((char *)&cv_pkt.data, data, len);
	DBG("Sending cntrl 0x%04X, len %d\n", cv_pkt.hdr, cv_pkt.len);

	PortTypes::CharSequence_var buffer_var;
	buffer_var = new PortTypes::CharSequence(sizeof(CryptVocPacket), sizeof(CryptVocPacket),
			(CORBA::Char *)&cv_pkt, 0);

	encrypt_out->pushPacket(buffer_var);
}
