//-------------------------------------------------------------------------
/**
 * @file 	VcwBuffers.cpp
 * @brief	Contains implementation of the VcwBuffers class
 * @note	none.
 *
 * @author	Cal(it)2
 * @author  Justyn Bell <jwbell@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 OSSIE/Linux Implementation
 * @endverbatim
 */
//-------------------------------------------------------------------------

#include <semaphore.h>
#include <pthread.h>
#include "VcwBuffers.h"
#include "p25/util/logger.h"
#include "p25/constants.h"

extern "C"
{
#include "p25/format/bit_fifo.h"
}

#define VC_PKT_SIZE VC_BLEN/8

/** 
 * @details
 */
VcwBuffers::VcwBuffers(void)
{
}

/** 
 * @details
 */
VcwBuffers::~VcwBuffers(void)
{
}

/** 
 * @details Allocates buffers with appropriate sizes, initializes semaphores and mutexes for
 *			bit FIFO producer<->consumer model.
 * @author Justyn Bell <jwbell@ucsd.edu>
 */
void VcwBuffers::initialize()
{
    pb_bf = bf_alloc(PLAY_BACK_BUF_SIZE);
    pthread_mutex_init (&vc_pb_bf_m, NULL);

    sem_init(&sem_pb_bf_empty, 0, 0);
    sem_init(&sem_pb_bf_notempty, 0, 0);
    sem_init(&sem_pb_bf_flush, 0, 0);
}

/** 
 * @details Frees all buffers, closes all semaphores and mutexes for bit FIFO producer<->consumer model.
 *
 * @author Justyn Bell <jwbell@ucsd.edu>
 */
void VcwBuffers::shutdown()
{
    bf_free(pb_bf);
    pthread_mutex_destroy (&vc_pb_bf_m);

    sem_destroy(&sem_pb_bf_empty);
    sem_destroy(&sem_pb_bf_notempty);
    sem_destroy(&sem_pb_bf_flush);

}

/** 
 * @details Puts voice codewords into pb_bf based on a simple producer consumer model. 
 *			The buffer only accepts less that the normal length of a voice codeword when
 *			a flush is requested.
 *
 * @author Justyn Bell <jwbell@ucsd.edu>
 */
void VcwBuffers::put_in_pb_buf(char *vcw, int len)
{
    int offset = 0;
    int space = 0;
    int res;

    while (len >= VC_BLEN) {
    	pthread_mutex_lock(&vc_pb_bf_m);
    	space = bf_space(pb_bf);
    	pthread_mutex_unlock(&vc_pb_bf_m);
    	while (space < VC_BLEN) {
    		sem_wait(&sem_pb_bf_empty);
    		pthread_mutex_lock(&vc_pb_bf_m);
    		space = bf_space(pb_bf);
    		pthread_mutex_unlock(&vc_pb_bf_m);
    	}

    	pthread_mutex_lock(&vc_pb_bf_m);
    	res = bf_put_n(pb_bf, vcw+offset, VC_BLEN);
		pthread_mutex_unlock(&vc_pb_bf_m);
    	if (res != VC_BLEN)
    		ERROR("Put %d instead of %d", res, VC_BLEN);

    	len -= VC_BLEN;
    	offset += VC_BLEN;

    	sem_post(&sem_pb_bf_notempty);
    }
}

/** 
 * @details Gets voice codewords out of pb_bf based on a simple producer consumer model. 
 *			The buffer only accepts less that the normal length of a voice codeword when
 *			a flush is requested.
 *
 * @author Justyn Bell <jwbell@ucsd.edu>
 */
int VcwBuffers::get_from_pb_buf(char *vcw, int len)
{
	int res;
	int size;

	if (len > VC_BLEN) {
		INFO("Reduced requested length to VC_BLEN = %d!\n",VC_BLEN);
		len = VC_BLEN;
	}

	pthread_mutex_lock(&vc_pb_bf_m);
	size = bf_size(pb_bf);
	pthread_mutex_unlock(&vc_pb_bf_m);
	while (size < len && !pb_bf_accept_less) {
		sem_wait(&sem_pb_bf_notempty);
		pthread_mutex_lock(&vc_pb_bf_m);
		size = bf_size(pb_bf);
		pthread_mutex_unlock(&vc_pb_bf_m);
	}

	pthread_mutex_lock(&vc_pb_bf_m);
	size = bf_size(pb_bf);
	res = bf_get_n(pb_bf,vcw,len);
	pthread_mutex_unlock(&vc_pb_bf_m);
	if (res != len)
		ERROR("Got %d instead of %d", res, len);

	if (pb_bf_accept_less && size == 0) {
		pb_bf_accept_less = 0;
		sem_post(&sem_pb_bf_flush);
	}

//	INFO("Size of pb_bf = %d bytes\n", size);
	sem_post(&sem_pb_bf_empty);

	return res;
}

/** 
 * @details Flushes voice codeword buffers by telling the consumer that it can accept less data than requested,
 *			and tells the producer to wait until the flush is completed to start putting more data in the FIFO. Once the 
 *			buffer size is 0, the producer is posted, and the model continues.
 *
 * @author Justyn Bell <jwbell@ucsd.edu>
 */
void VcwBuffers::flush_buffer()
{
	int size;

	INFO("Entered flush_buffer\n");

	pthread_mutex_lock(&vc_pb_bf_m);
	size = bf_size(pb_bf);
	pthread_mutex_unlock(&vc_pb_bf_m);
	do {
		INFO("Flushing pb_bf\n");
		if (!pb_bf_accept_less)
			pb_bf_accept_less = 1;

		sem_post(&sem_pb_bf_notempty);
		sem_wait(&sem_pb_bf_flush);
		pthread_mutex_lock(&vc_pb_bf_m);
		size = bf_size(pb_bf);
		pthread_mutex_unlock(&vc_pb_bf_m);
	} while (size > 0);

	INFO("Leaving flush_buffer\n");
}


