/*
 * Copyright (c) 2001-2003, David Janssens
 * Copyright (c) 2002-2003, Yannick Verschueren
 * Copyright (c) 2003-2005, Francois Devaux and Antonin Descampe
 * Copyright (c) 2005, Hervé Drolon, FreeImage Team
 * Copyright (c) 2002-2005, Communications and remote sensing Laboratory, Universite catholique de Louvain, Belgium
 * Copyright (c) 2005-2006, Dept. of Electronic and Information Engineering, Universita' degli Studi di Perugia, Italy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
@file rs.h
@brief Functions used to compute Reed-Solomon parity and check of byte arrays

*/

#ifndef _P25_ECC_RS_H
#define _P25_ECC_RS_H

/** Global definitions for Reed-Solomon encoder/decoder
 * Phil Karn KA9Q, September 1996
 *
 * The parameters r_MM and r_KK specify the Reed-Solomon code parameters.
 *
 * Set r_MM to be the size of each code symbol in bits. The Reed-Solomon
 * block size will then be r_NN = 2**M - 1 symbols. Supported values are
 * defined in rs.c.
 *
 * Set r_KK to be the number of data symbols in each block, which must be
 * less than the block size. The code will then be able to correct up
 * to r_NN-r_KK erasures or (r_NN-r_KK)/2 errors, or combinations thereof with
 * each error counting as two erasures.
 */

#define r_MM  6           /* RS code over GF(2**r_MM) - change to suit */

/* Original code */
/*#define r_KK  239*/             /* r_KK = number of information symbols */

#define r_NN ((1 << r_MM) - 1)

#if (r_MM <= 8)
typedef char r_dtype;
#else
typedef unsigned int r_dtype;
#endif

/** Initialization function */
void init_rs(void);

/** These two functions *must* be called in this order (e.g.,
 * by init_rs()) before any encoding/decoding
 */
void generate_gf_rs(void); /* Generate Galois Field */
void gen_poly_rs(void);    /* Generate generator polynomial */

/** Reed-Solomon encoding
 * data[] is the input block, parity symbols are placed in bb[]
 * bb[] may lie past the end of the data, e.g., for (255,223):
 *      encode_rs(&data[0],&data[223]);
 */
int encode_rs(r_dtype data[], r_dtype enc[]);
int encode_rs_36_20(r_dtype data[], r_dtype enc[]);
int encode_rs_24_16(r_dtype data[], r_dtype enc[]);
int encode_rs_24_12(r_dtype data[], r_dtype enc[]);

/** Reed-Solomon erasures-and-errors decoding
 * The received block goes into data[], and a list of zero-origin
 * erasure positions, if any, goes in eras_pos[] with a count in no_eras.
 *
 * The decoder corrects the symbols in place, if possible and returns
 * the number of corrected symbols. If the codeword is illegal or
 * uncorrectible, the data array is unchanged and -1 is returned
 */
int eras_dec_rs(r_dtype data[], int eras_pos[], int no_eras);
int decode_rs_36_20(r_dtype data[], int eras_pos[], int no_eras);
int decode_rs_24_16(r_dtype data[], int eras_pos[], int no_eras);
int decode_rs_24_12(r_dtype data[], int eras_pos[], int no_eras);

/**
Computes the minimum between two integers
@param a first integer to compare
@param b second integer to compare
@return returns the minimum integer between a and b
*/
#ifndef min
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif /* min */

#endif /* _P25_ECC_RS_H_ */
