/****************************************************************************** 
 
     Name:  ham.c 
   Author:  LeCocq 
     Date:  7/28/93 
  History:  6/28/93  New 
            7/28/93  Revised by ALW to compile on HP 
            5/20/09  Modified for Calit2 APCO25 build by Jeff Cuenco

This module contains two functions: 
    hamming_10_encode ( word )    encodes (10, 6,3) shortened Hamming code 
    hamming_15_encode ( word )    encodes (15,11,3) standard  Hamming code 
The functions return a long int value that is 18, 23, or 24 bits long 
for the respective function. 
 
******************************************************************************/ 
 
#include <stdio.h>
#include <math.h>

#include "p25/ecc/ham.h"
 
static long int matrix_10[6]    = {0x20e,0x10d,0x8b,0x47,0x23,0x1c}; 
static int code_word_mask_10[6] = {0x20 ,0x10 ,0x8, 0x4, 0x2, 0x1}; 

static long int h_tr_matrix_10[10] = {0xe,0xd,0xb,0x7,0x3,0xc,0x8,0x4,0x2,0x1};
static int decode_mask_10[10] = {0x200,0x100,0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};
 
long int hamming_10_encode(int code_word_in) 
 
{ 
    long int code_word_out = 0; 
    long int temp_word; 
    int i;
 
    for(i=0;i<6;i++) 
    { 
        temp_word = code_word_in & code_word_mask_10[i]; 
        if(temp_word >= 1) 
            code_word_out = code_word_out ^ matrix_10[i]; 
    } 
 
    return(code_word_out); 
 
} 

int hamming_10_decode(long int code_word_in, int *code_word_out)
{
	int syndrome = 0;
	int temp_word;
	int i;
    int eras = 0;

	for(i=0;i<10;i++)
	{
		temp_word = code_word_in & decode_mask_10[i];
		if (temp_word >= 1)
			syndrome = syndrome ^ h_tr_matrix_10[i];
	}

	if (syndrome != 0) {
            *code_word_out = code_word_in;
			eras = 1;
	} else {
		*code_word_out = code_word_in;
        eras = 0;
    }

	*code_word_out = *code_word_out >> 4;

	return eras;
}
 
int hamming_10_decode_with_correction(long int code_word_in)
{
	int code_word_out = 0;
	int syndrome = 0;
	int temp_word;
	int bit_count = 0;
	int i;
	for(i=0;i<10;i++)
	{
		temp_word = code_word_in & decode_mask_10[i];
		if (temp_word >= 1)
			syndrome = syndrome ^ h_tr_matrix_10[i];
	}

	if (syndrome != 0) {
		bit_count = ((syndrome & 0x8) >> 3) + ((syndrome & 0x4) >> 2) + ((syndrome & 0x2) >> 1) + (syndrome & 0x1);
		printf("\nsyndrome = 0x%x,  bit count = %d\n", syndrome, bit_count);
		if ((bit_count % 2) != 0) { // Odd number of 1's
			// Assume 1 error occurred and decode as such
			printf("Error in decoding, attempting to repair error...\n"); 
			for(i=0; i < 10; i++) {
				if (syndrome == h_tr_matrix_10[i]) {
					printf("Found error location in column %d of parity check matrix, correcting...\n", 10-i-1);
					code_word_out = code_word_in ^ (int)pow(2,10-i-1);
					break;
				}
			}
			if (i == 10) {
				printf("Attempt to repair error failed!\n");
				return -1;
			}
		}
		else {	// Even number of 1's
			printf("Attempt to repair error failed - uncorrectable error pattern detected!\n");
			return -1;
		}
	}
	else
		code_word_out = code_word_in;

	code_word_out = code_word_out >> 4;

	return(code_word_out);
}
 
static long int matrix_15[11] = {0x400f,0x200e,0x100d,0x80c,0x40b,0x20a,0x109, 
                                 0x87,0x46,0x25,0x13}; 
static int code_word_mask_15[11] = {0x400,0x200,0x100,0x80,0x40,0x20,0x10, 
                                    0x08,0x04,0x02,0x01}; 

static long int h_tr_matrix_15[15] = {0xf,0xe,0xd,0xc,0xb,0xa,0x9,0x7,0x6,0x5,0x3,0x8,0x4,0x2,0x1};
static int decode_mask_15[15] = {0x4000,0x2000,0x1000,0x800,0x400,0x200,0x100,0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};
 
long int hamming_15_encode(int code_word_in) 
 
{ 
 
    long int code_word_out = 0; 
    long int temp_word; 
    int i;
 
    for(i=0;i<11;i++) 
    { 
        temp_word = code_word_in & code_word_mask_15[i]; 
        if(temp_word >= 1) 
            code_word_out = code_word_out ^ matrix_15[i]; 
    } 
 
    return(code_word_out); 
 
} 

int hamming_15_decode(long int code_word_in) 
{
	int code_word_out = 0;
	int syndrome = 0;
	int temp_word;
	int i;
	for(i=0;i<15;i++)
	{
		temp_word = code_word_in & decode_mask_15[i];
		if (temp_word >= 1)
			syndrome = syndrome ^ h_tr_matrix_15[i];
	}
	
	if (syndrome != 0) {
		printf("Error in decoding, attempting to repair error...\n"); 
		for(i=0; i < 15; i++) {
			if (syndrome == h_tr_matrix_15[i]) {
				printf("Found error location in column %d of parity check matrix, correcting...\n", 15-i-1);
				code_word_out = code_word_in ^ (int)pow(2,15-i-1);
				break;
			}
		}
		if (i == 15) {
			printf("Attempt to repair error failed!\n");
			return -1;
		}
	}
	else
		code_word_out = code_word_in;
	
	code_word_out = code_word_out >> 4;
	
	return(code_word_out);
}
