/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

//#define _DEBUG
#include <string>
#include <iostream>
#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <fcntl.h>
#include "p25_Receiver.h"

/* user includes */
#include "p25/config.h"
#include "p25/controller.h"
#include "p25/format/packets.h"

extern "C" {
#include "p25/ecc/golay.h"
#include "p25/ecc/cyclic.h"
#include "p25/ecc/ham.h"
#include "p25/ecc/rs.h"
#include "p25/format/bit_fifo.h"
#include "p25/format/bit_manipulate.h"
}

#include "p25/util/logger.h"

#define ALGID_UNENCRYPTED	0x80
#define CRYPTO_START		0xAA
#define CRYPTO_STOP			0xCC

using namespace P25;

//-------------------------------------------------------------------------
/**
 * @details Receiver class constructor for OSSIE implementation.
 *
 * @author
 */
//-------------------------------------------------------------------------
p25_Receiver_i::p25_Receiver_i(const char *uuid, omni_condition *condition) :
    Resource_impl(uuid), component_running(condition)
{
    modem_data = new standardInterfaces_i::realChar_p("mod_in");
    voc_out_data = new standardInterfaces_i::realChar_u("voc_out");
    voc_in_data = new standardInterfaces_i::realChar_p("voc_in");
    ui_dataOut = new standardInterfaces_i::realChar_u("ui_out");
    ui_dataIn = new standardInterfaces_i::realChar_p("ui_in");
    enc_dataOut = new standardInterfaces_i::realChar_u("enc_out");
    enc_dataIn = new standardInterfaces_i::realChar_p("enc_in");
//    start();
}

//-------------------------------------------------------------------------
/**
 * @details Receiver class destructor for OSSIE implementation.
 *
 * @author
 */
//-------------------------------------------------------------------------
p25_Receiver_i::~p25_Receiver_i(void)
{
    delete modem_data;
    delete voc_out_data;
    delete voc_in_data;
    delete ui_dataOut;
    delete ui_dataIn;
    delete enc_dataOut;
    delete enc_dataIn;
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::Run( void * data )
{
    ((p25_Receiver_i*)data)->ProcessData();
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::ui_Run( void * data )
{
    ((p25_Receiver_i*)data)->ui_ProcessData();
}

//-------------------------------------------------------------------------
/**
 * @details Static function for omni thread.
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::rx_Run( void * data )
{
    ((p25_Receiver_i*)data)->run_rx();
}

//-------------------------------------------------------------------------
/**
 * @details Get port data.
 *
 * @author
 */
//-------------------------------------------------------------------------
CORBA::Object_ptr p25_Receiver_i::getPort( const char* portName ) throw (
    CORBA::SystemException, CF::PortSupplier::UnknownPort)
{
    DEBUG(3, p25_Receiver, "getPort() invoked with " << portName)

    CORBA::Object_var p;

    p = modem_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = voc_out_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = voc_in_data->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = ui_dataOut->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = ui_dataIn->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = enc_dataOut->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    p = enc_dataIn->getPort(portName);

    if (!CORBA::is_nil(p))
        return p._retn();

    /*exception*/
    throw CF::PortSupplier::UnknownPort();
}

//-------------------------------------------------------------------------
/**
 * @details	Instantiates Controller, sets stack size, allocates
 *			buffers and starts all threads for Rx component.
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::start() throw (CORBA::SystemException,
    CF::Resource::StartError)
{
    DEBUG(3, p25_Receiver, "start() invoked");

    // Instantiate Controller
    mController = new Controller(0, NULL, NULL, 2);
    mController->start();

    rx_bf = bf_alloc(MAX_FRAME_SIZE);

    omni_thread::stacksize(2*1024*1024);

	omni_mutex_lock  l(processing_mutex);
	if( false == thread_started )
	{
		thread_started = true;
		// Create the thread for the writer's processing function
		processing_thread = new omni_thread(Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		processing_thread->start();

		INFO("Started rx_in port pushPacket processing thread.\n");
	}

	omni_mutex_lock  ll(ui_processing_mutex);
	if( false == ui_thread_started )
	{
		ui_thread_started = true;
		// Create the thread for the writer's processing function
		ui_processing_thread = new omni_thread(ui_Run, (void *) this, omni_thread::PRIORITY_NORMAL);

		// Start the thread containing the writer's processing function
		ui_processing_thread->start();

		INFO("Started ui_in port pushPacket processing thread.\n");
	}

//	omni_mutex_lock  lll(rx_processing_mutex);
	if( false == rx_thread_started )
	{
		rx_thread_started = true;
		// Create the thread for the writer's processing function
		rx_processing_thread = new omni_thread(rx_Run, (void *) this, omni_thread::PRIORITY_HIGH);

		// Start the thread containing the writer's processing function
		rx_processing_thread->start();

		INFO("Started RX main processing thread.\n");
	}
}

//-------------------------------------------------------------------------
/**
 * @details	Stops all threads, frees buffers and cleans up Controller.
 *
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::stop() throw (CORBA::SystemException, CF::Resource::StopError)
{
    DEBUG(3, p25_Receiver, "stop() invoked")
	omni_mutex_lock l(processing_mutex);
	thread_started = false;

	omni_mutex_lock ll(ui_processing_mutex);
	ui_thread_started = false;

//	omni_mutex_lock  lll(rx_processing_mutex);
	rx_thread_started = false;

	bf_free(rx_bf);
}

//-------------------------------------------------------------------------
/**
 * @details
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::releaseObject() throw (CORBA::SystemException,
    CF::LifeCycle::ReleaseError)
{
    DEBUG(3, p25_Receiver, "releaseObject() invoked")

    component_running->signal();
}

//-------------------------------------------------------------------------
/**
 * @details Initializes all semaphores and mutexes for Rx component, and puts
 * 			the component in an initialized state.
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::initialize() throw (CF::LifeCycle::InitializeError,
    CORBA::SystemException)
{
	int result;
	CF::StringSequence initErrSeq;

    DEBUG(3, p25_Receiver, "initialize() invoked");

	result = pthread_mutex_init(&vc_rx_bf_m, NULL);
	if (result != 0) {
		initErrSeq.length(1);
		initErrSeq[0] = CORBA::string_dup("RX pthread_mutex_init() failure\n");
		throw CF::LifeCycle::InitializeError(initErrSeq);
	}

	sem_init(&m_bf_full, 0, 0);
	sem_init(&m_bf_empty, 0, 0);

	sem_init(&sem_ui_rx_ack, 0, 0);
	sem_init(&sem_run_rx, 0, 0);

	rx_stop_requested = false;
}

//-------------------------------------------------------------------------
/**
 * @details
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::query( CF::Properties & configProperties ) throw (CORBA::SystemException, CF::UnknownProperties)
{
	unsigned int i, j;
	if( configProperties.length() == 0 )
	{
		configProperties.length( propertySet.length() );
		for( i = 0; i < propertySet.length(); i++ )
		{
			configProperties[i].id = CORBA::string_dup( propertySet[i].id );
			configProperties[i].value = propertySet[i].value;
		}
		return;
	} else {
		for( i = 0; i < configProperties.length(); i++ ) {
			for( j = 0; j < propertySet.length(); j++ ) {
				if( strcmp(configProperties[i].id, propertySet[j].id) == 0 ) {
					configProperties[i].value = propertySet[j].value;
				}
			}
		}
	} // end if-else
}

//-------------------------------------------------------------------------
/**
 * @details
 * @author
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::configure(const CF::Properties& props)
throw (CORBA::SystemException,
    CF::PropertySet::InvalidConfiguration,
    CF::PropertySet::PartialConfiguration)
{
    DEBUG(3, p25_Receiver, "configure() invoked")

    static int init = 0;
	unsigned int i, j;

    if( init == 0 ) {
        if( props.length() == 0 ) {
            std::cout << "configure called with invalid props.length - " << props.length() << std::endl;
            return;
        }
        propertySet.length(props.length());
        for( j=0; j < props.length(); j++ ) {
            propertySet[j].id = CORBA::string_dup(props[j].id);
            propertySet[j].value = props[j].value;
        }
        init = 1;
    }

    std::cout << "props length : " << props.length() << std::endl;

    for ( i = 0; i <props.length(); i++)
    {
        std::cout << "Property id : " << props[i].id << std::endl;

    }
}

//-------------------------------------------------------------------------
/**
 * @details Reads data from modem and puts into a fifo buffer if
 * 			in Rx mode.
 *
 * @author		Jeffrey Cuenco <jcuenco@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::ProcessData()
{
    DEBUG(3, Transmitter, "ProcessData() invoked");

    PortTypes::CharSequence *I_in_0(NULL);
    CORBA::UShort I_in_0_length;

	int res, len, bf_isfull, bufsize;
	char *buf;

    while(continue_processing())
    {
        modem_data->getData(I_in_0);
        I_in_0_length = I_in_0->length();

        if (I_in_0_length > 0) {
        	/* get buffer and buffer length */
        	buf = (char *)(I_in_0->get_buffer());
			len = I_in_0_length;

        	/* write code to enqueue in buffer */
            if (mController->get_radio_activity() == RX_ACTIVITY) {
            	// Block on semaphore if buffer is full or does not have enough space to
            	// store all of the incoming payload into the buffer.
            	pthread_mutex_lock(&vc_rx_bf_m);
            	bf_isfull = bf_full(rx_bf) || (len > bf_space(rx_bf));
            	pthread_mutex_unlock(&vc_rx_bf_m);

            	while (bf_isfull) {
            	    //INFO("bf_isfull, waiting\n");
            	    if (sem_wait(&m_bf_full) < 0) {
						ERROR("error in sem_wait\n");
						continue;
            	    }
            	    pthread_mutex_lock(&vc_rx_bf_m);
            	    bf_isfull = bf_full(rx_bf) || (len > bf_space(rx_bf));
            	    pthread_mutex_unlock(&vc_rx_bf_m);
            	}
            	//INFO("!bf_isfull, continuing\n");

            	// Put payload into buffer.
            	pthread_mutex_lock(&vc_rx_bf_m);
            	res = bf_put_n(rx_bf,buf,len);
            	bufsize = bf_size(rx_bf);
            	pthread_mutex_unlock(&vc_rx_bf_m);

            	// Notify the consumer that the buffer is no longer empty
            	// provided a nonzero payload was inserted into the buffer.
            	if (bufsize > 0) {
            	    //INFO("bf_isempty, posting\n");
            	    if (sem_post(&m_bf_empty) < 0) {
						ERROR("error in sem_post\n");
            	    }
            	}
            }
        }
        modem_data->bufferEmptied();
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool p25_Receiver_i::continue_processing()
{
	omni_mutex_lock l(processing_mutex);
	return thread_started;
}

//-------------------------------------------------------------------------
/**
 * @details	Reads packetized control data from UI component.  If header is START,
 * 			set Rx component into active mode.
 * 			If header is STOP, set Rx component to standby mode, stop getting
 * 			data from Modem component, wait for the Rx processing thread to
 * 			complete and send an ACK to the UI component to notify the UI
 * 			that the Rx component has finished processing and is in standby.
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::ui_ProcessData()
{
    DEBUG(3, Receiver, "ProcessData() invoked");

    PortTypes::CharSequence *I_in_1(NULL);
    CORBA::UShort I_in_1_length;

    PortTypes::CharSequence_var buffer_var;

	UIControlPacket *ui_pkt;

    while(ui_continue_processing())
    {
        ui_dataIn->getData(I_in_1);
        I_in_1_length = I_in_1->length();

        if (I_in_1_length == sizeof(UIControlPacket)) {
			ui_pkt = (UIControlPacket *)I_in_1->get_buffer();
			memcpy((char *)&m_ui,(char *)&ui_pkt->ui,sizeof(UInterface));
			DBG("Got Stuff from UI. EncState is %d\n", m_ui.rxEncState);

			switch (ui_pkt->hdr) {
			case START:
				if (mController->get_radio_activity() == STANDBY_ACTIVITY) {
					INFO("Switching Receiver Device to RX_ACTIVITY\n");
					mController->set_radio_activity(RX_ACTIVITY);
					sem_post(&sem_run_rx);
				}
				break;

			case STOP:
				if ((mController->get_radio_activity() == RX_ACTIVITY)) {
					INFO("Switching Receiver Device to STANDBY_ACTIVITY\n")
					mController->set_radio_activity(STANDBY_ACTIVITY);

					// post semaphore to release get_bits
					rx_stop_requested = true;
					sem_post(&m_bf_empty);

					sem_wait(&sem_ui_rx_ack);
					rx_stop_requested = false;
				}
				// Send ack to UI
				DBG("Sending ACK to UI device\n");
				sendstate_rx_ui(ACK);
				break;

			default:
				ERROR("Unkown UIControlPacket Header\n");
				break;
			}
        }
        ui_dataIn->bufferEmptied();
    }
}
//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author
 */
//-------------------------------------------------------------------------
bool p25_Receiver_i::ui_continue_processing()
{
	omni_mutex_lock ll(ui_processing_mutex);
	return ui_thread_started;
}


//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::frameIsMine(void)
{
    /*TODO: Check NAC before proceeding to decode frame*/
    if (m_ui.nac == mr_ui.nac || m_ui.nac == P25_NO_NAC_MODE)
        return 1;

    if (!memcmp(mr_ui.dstId,m_ui.unitId,sizeof(mr_ui.dstId)) ||
        !memcmp(mr_ui.dstId, EVERYONE_UID,sizeof(mr_ui.dstId))){
        if ((mr_ui.tgId == m_ui.tgId) || (mr_ui.tgId == EVERYONE_TGID)){
            return 1;
        }
    }

    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::play_vcw(char* vcw, int len)
{
    if (frameIsMine()) {
    	sendbuf_rx_vox(VOCODER_PKT_CNTRL_DATA, vcw, len);
		sendstate_rx_ui(SCAN_HALT);
    } else {
        printf("Frame is not mine!!\n");
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::nid_dec(char* nid_cw, char* duid)
{
    mr_ui.nac = bi2de(nid_cw,12);
    *duid = bi2de(nid_cw+12,4);

    sendstate_rx_ui(UPDATE);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::lsd_dec(char* lsd_cw, char* lsd)
{
    int i;
    char decw;

    // get and decode lsd
    for (i = 0; i < 2; i++) {
        decw = cyclic_16_8_decode(bi2de(lsd_cw+(i*16),16));
        lsd[i] = decw & 0xff;
    }
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::is_lc_protected(struct bit_fifo *lc)
{
    char bits[8];
	int p;

    memset(bits, 0, sizeof(bits));
    bf_get_n(lc,bits,1);
    p = bi2de(bits,1);
	
	DBG("P = %d\n", p);
	return p;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::lc_deformat(struct bit_fifo *lc)
{
    char bits[32];
    int lco, /*p,*/ sf;
    unsigned int i;

    //memset(bits, 0, 1);
    //bf_get_n(lc,bits,1);
    //p = bi2de(bits,1);
    //if (p == 0) {
    	memset(bits, 0, 1);
        bf_get_n(lc,bits,1);
        sf = bi2de(bits,1);

        memset(bits, 0, 6);
        bf_get_n(lc,bits,6);
        lco = bi2de(bits,6);

        if (lco == LC_GRP_V_CH_USR) {
        	DBG("lco == LC_GRP_V_CH_USR\n");

            // MFID
            bf_get_n(lc,bits,8);
            mr_ui.mfId = bi2de(bits,8);

            // Service Options
            bf_get_n(lc,bits,8);

            // Reserved
            bf_get_n(lc,bits,8);

            // Group Address
            bf_get_n(lc,bits,16);
            mr_ui.tgId = (short)bi2de(bits,16);

            // Source Address
            for (i = 0; i < sizeof(mr_ui.unitId); i++) {
                bf_get_n(lc,bits,8);
                mr_ui.unitId[i] = bi2de(bits,8);
				DBG("unitid[%d] = %d\n",i,mr_ui.unitId[i]);  
            }
        } else if (lco == LC_UU_V_CH_USR) {
        	DBG("lco == LC_UU_V_CH_USR\n");

            // MFID
        	memset(bits, 0, 8);
            bf_get_n(lc,bits,8);
            mr_ui.mfId = bi2de(bits,8);

            // Service Options
            memset(bits, 0, 8);
            bf_get_n(lc,bits,8);

            // Target Address
            for (i = 0; i < sizeof(mr_ui.dstId); i++) {
            	memset(bits, 0, 8);
                bf_get_n(lc,bits,8);
                mr_ui.dstId[i] = bi2de(bits,8);
            }

            // Source Address
            for (i = 0; i < sizeof(mr_ui.unitId); i++) {
            	memset(bits, 0, 8);
                bf_get_n(lc,bits,8);
                mr_ui.unitId[i] = bi2de(bits,8);
            }
        } else if (lco == LC_TELE_INT_V_CH_USR) {

        } else if (lco == LC_SOURCE_ID_EXT) {

        } else if (lco == LC_UU_V_CH_USR_EXT) {

        } else if (lco == LC_CALL_TRM_CAN) {

        } else if (lco == LC_STS_Q) {

        } else if (lco == LC_STA_UPDT) {

        } else if (lco == LC_MSG_UPDT) {

        } else if (lco == LC_CALL_ALRT) {

        } else if (lco == LC_EXT_FNCT_CMD) {

        } else if (lco == LC_CH_ID_UPDT) {

        } else {

        }
    //}

    sendstate_rx_ui(UPDATE);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::es_deformat(struct bit_fifo *es)
{
	char bits[16];

    for (int i = 0; i < (int)sizeof(m_ui.mi); i++) {
    	bf_get_n(es,bits,8);
    	mr_ui.mi[i] = bi2de(bits,8);
    }
    //INFO("mi = 0x%02X%02X%02X%02X%02X%02X%02X%02X%02X\n", mr_ui.mi[0],
    //		mr_ui.mi[1], mr_ui.mi[2], mr_ui.mi[3], mr_ui.mi[4], mr_ui.mi[5],
    //		mr_ui.mi[6], mr_ui.mi[7], mr_ui.mi[8]);

	bf_get_n(es,bits,8);

	if (m_ui.rxEncState)
		mr_ui.algId = m_ui.algId;
	else
		mr_ui.algId = bi2de(bits,8);

	DBG("algid = 0x%02X\n", mr_ui.algId);

	bf_get_n(es,bits,16);
	mr_ui.keyId = bi2de(bits,16);
	//INFO("keyid = 0x%04X\n\n", mr_ui.keyId);

    sendstate_rx_ui(UPDATE);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::lcw_dec(char* lcw_cw, char *lc)
{
    int i,decw,eras,eras_count,temp;
    struct bit_fifo *bf;
    char dec_bits[24];
    int   dec_eras[24];
    char bits[32];

    bf = bf_alloc(HEXBIT*12);

    // get bits and send to hamming decoder
    eras_count = 0;
    for (i = 0; i < 24; i++) {
        eras = hamming_10_decode(bi2de(lcw_cw+(i*10),10), &decw);
        dec_bits[i] = decw & 0x3f;
        dec_eras[i] = eras;
        if (eras == 1) eras_count++;
    }

    reverse_bits(dec_bits,24);
    for (i = 0; i < 24/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[24-i-1];
        dec_eras[24-i-1] = temp;
    }
    decode_rs_24_12(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,24); // we only need the first 12 bits

    for (i = 0; i < 12; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    // Extract from bit fifo to lc byte buffer
    bf_export(bf, lc, 12*HEXBIT);

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::esw_dec(char* esw_cw, char *es)
{
    int decw,eras,eras_count,temp;
    unsigned int i;
    struct bit_fifo *bf;
    char dec_bits[24];
    int   dec_eras[24];
    char bits[32];

    bf = bf_alloc(HEXBIT*16);

    // get bits and send to hamming decoder
    eras_count = 0;
    for (i = 0; i < 24; i++) {
        eras = hamming_10_decode(bi2de(esw_cw+(i*10),10), &decw);
        dec_bits[i] = decw & 0x3f;
        dec_eras[i] = eras;
        if (eras == 1) eras_count++;
    }

    reverse_bits(dec_bits,24);
    for (i = 0; i < 24/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[24-i-1];
        dec_eras[24-i-1] = temp;
    }
    decode_rs_24_16(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,24); // we only need the first 16 bits

    for (i = 0; i < 16; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    // Extract from bit fifo to es byte buffer
    bf_export(bf, es, 16*HEXBIT);

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::hdu_dec(char* hdu, char *mi, char *algid, char *kid, char *tgid)
{
    int eras,eras_count,temp;
    unsigned int i;
    long decw;
    struct bit_fifo *bf;
    char dec_bits[36];
    int   dec_eras[36];
    char bits[32];

    bf = bf_alloc(HEXBIT*20);

    // get bits and send to hamming decoder
    eras_count = 0;
    for (i = 0; i < 36; i++) {
        decw = bi2de(hdu+(i*18),18);
        eras = golay_18_6_decode(&decw);
        dec_bits[i] = decw & 0x3f;
        dec_eras[i] = eras;
        if (eras == 1) eras_count++;
    }

    reverse_bits(dec_bits,36);
    for (i = 0; i < 36/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[36 -i-1];
        dec_eras[36-i-1] = temp;
    }
    decode_rs_36_20(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,36); // reverse all 36 bits but we only use 20 bits

    for (i = 0; i < 20; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

    for (i = 0; i < sizeof(mr_ui.mi); i++) {
        bf_get_n(bf,bits,8);
        mr_ui.mi[i] = bi2de(bits,8);
        memcpy(mi+8*i, bits, 8);
    }

    bf_get_n(bf,bits,8);
    mr_ui.mfId = bi2de(bits,8);

    bf_get_n(bf,bits,8);
    mr_ui.algId = bi2de(bits,8);
    memcpy(algid, bits, 8);

    bf_get_n(bf,bits,16);
    mr_ui.keyId = bi2de(bits,16);
    memcpy(kid, bits, 16);

    bf_get_n(bf,bits,16);
    mr_ui.tgId = bi2de(bits,16);
    memcpy(tgid, bits, 16);

    mr_ui.ptt = 1;

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::ldu_cw_split(char* ldu, int len, char* vc, char* lsd, char* lc)
{
    int i,vc_offset;
    struct bit_fifo *bf;

    bf = bf_alloc(len);
    bf_put_n(bf,ldu,len);

    bf_get_n(bf,vc,VC_BLEN*2);

    for (i = 0; i < (LCW_BLEN/40); i++) {
        vc_offset = (VC_BLEN*2)+(i*VC_BLEN);
        bf_get_n(bf,lc+(i*40),40);
        bf_get_n(bf,vc+vc_offset,VC_BLEN);
    }

    bf_get_n(bf,lsd,LSD_BLEN);
    bf_get_n(bf,vc+vc_offset+VC_BLEN,VC_BLEN);
    bf_free(bf);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::ldu_dec(char* ldu, char *vc, char *lc_es, char *lsd, char type)
{
    int len,res;

    char lsd_cw[LSD_BLEN];
    char ldu_ci[LCW_BLEN]; // LDU control Information (LCW or ESW)

    len = LDU_VC_BLEN;
    ldu_cw_split(ldu,len,vc,lsd_cw,ldu_ci);
    lsd_dec(lsd_cw,lsd);

    switch (type) {
        case DUID_LDU1:
            res = lcw_dec(ldu_ci, lc_es);
            if (res < 0)
                ERROR("LCW decode failed at frame %d: %d\n", frame_id, res);
    	    break;
        case DUID_LDU2:
    	    res = esw_dec(ldu_ci, lc_es);
            if (res < 0)
                ERROR("ESW decode failed at frame %d: %d\n", frame_id, res);
    	    break;
    }

    mr_ui.ptt = 1;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::tdu_dec(char* tdu)
{
	mr_ui.ptt = 0;
	sendstate_rx_ui(UPDATE);
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::tdu_lcw_dec(char* tdu)
{
    int i,j,eras,eras_count,temp;
    long decw;
    struct bit_fifo *bf;
    char dec_bits[24];
    int dec_eras[24];
    char bits[32];

    bf = bf_alloc(HEXBIT*12);

    // get bits and send to hamming decoder
    eras_count = 0;
    j = 0;
    for (i = 0; i < 12; i++) {
        decw = bi2de(tdu+(i*24),24);
        eras = golay_24_12_decode(&decw);
        dec_bits[j] = (decw >> HEXBIT) & 0x3f;
        dec_bits[j+1] = decw  & 0x3f;
        dec_eras[j] = eras;
        dec_eras[j+1] = eras;
        j += 2;
        if (eras == 1) eras_count += 2;
    }

    reverse_bits(dec_bits,24);
    for (i = 0; i < 24/2; i++) {
        temp = dec_eras[i];
        dec_eras[i] = dec_eras[24-i-1];
        dec_eras[24-i-1] = temp;
    }
    decode_rs_24_12(dec_bits,dec_eras,eras_count);
    reverse_bits(dec_bits,24); //reverse all 24 bits but we use only the first 12

    for (i = 0; i < 12; i++) {
        de2bi(dec_bits[i],bits,HEXBIT);
        bf_put_n(bf,bits,HEXBIT);
    }

	if (is_lc_protected(bf)) {
		// Should decrypt lc here but since this is TDU, we'll skip it
	} else {
    	lc_deformat(bf);
	}

	mr_ui.ptt = 0;
	sendstate_rx_ui(UPDATE);

    bf_free(bf);
    return 0;
}

//-------------------------------------------------------------------------
/**
 * @details
 * @param[in] bits Buffer containing the FS bits to check
 * @return 	Integer value representing success or failure of the FS check
 * @retval 	The value returned is zero or greater (positive) when the number
 * 			of bit differences in the received frame sync is less than the
 * 			FS_ERROR_THRESHOLD and is less than zero (negative) otherwise.
 * @see 	FS_ERROR_THRESHOLD
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::check_fs(char* bits)
{
    int threshold = FS_ERROR_THRESHOLD;
    int i,sum;

    sum = 0;
    for (i = 0; i < FS_BLEN; i++) {
    	// Sum up all the differences between received FS bits
    	// and expected FSbits
        sum += abs(fs[i] - bits[i]);
    }

    // 
    return threshold - sum;
}

//-------------------------------------------------------------------------
/**
 * @details Gets the specified number of bits from the receiver bit stream
 * 			with the status bits extracted. rx_bit_count tracks the number
 * 			of data bits read since Status bits were last encountered.
 * @param[in] bits The buffer to store the requested bits.
 * @param[in] len The number of bits to read.
 * @return The number of bits read.
 * @see rx_bit_count
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::get_bits(char* bits, int len)
{
    int res;
    int bits_read; // Count of total bits read so far
    int bits_to_get;
    char ss[2];

    bits_read = 0;
    while((bits_read < len) && !rx_stop_requested) {

    	// Calculate number of bits to read to make sure that
    	// Status bits are not included with the data
        bits_to_get = 70 - rx_bit_count;
        if (bits_to_get > (len - bits_read))
            bits_to_get = len - bits_read;

	    res = bs_input(bits+bits_read, bits_to_get);
	    bits_read += res;
	    rx_bit_count += res;

	    // Extract Status Bits from the stream.
	    // Currently the status bits are discarded.
        if (rx_bit_count == 70) {
    	    //INFO("rx_bit_count = 70, reading status symbol\n");

    	    res = bs_input(ss,SS_BLEN);
    	    if (res != SS_BLEN) {
    			ERROR("Did not retrieve status symbol!  res = %d\n", res);
    	    }

    	    rx_bit_count = 0;
        }
    }

    return bits_read;
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//------------------------------------------------------------------------
void p25_Receiver_i::rx_initialize()
{
    rx_bit_count = 0;
    frame_id = 0;
    memset((char* )&mr_ui,0,sizeof(mr_ui));
}

//-------------------------------------------------------------------------
/**
 * @details
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::rx_shutdown()
{
	PortTypes::CharSequence_var buffer_var;
	PortTypes::CharSequence *rx_data_in(NULL);
	//VocoderPacket vPkt;
	VocoderPacket *pvPkt;
	CryptoPacket *pcPkt;

	// Send Stop Control Packet to Encryption Device
	if (encrypted_frame_mode) {
		// Send Encryption Device a CRYPTOC_STOP message
		DBG("Sending Crypto Cntrl Stop\n");
		sendbuf_rx_crypto(CRYPTO_PKT_CNTRL_STOP);
	}

	// Send STOP message to Vocoder Device to cleanup its state.
	DBG("Sending VOCODER_PKT_CNTRL_STOP message to Vocoder Device\n");
	sendbuf_rx_vox(VOCODER_PKT_CNTRL_STOP);

	// Wait for ACKs to be received from Vocoder and Encryption Devices
	do {
		voc_in_data->getData(rx_data_in);
		pvPkt = (VocoderPacket *)rx_data_in->get_buffer();
	} while (pvPkt->hdr != VOCODER_PKT_CNTRL_ACK);
	voc_in_data->bufferEmptied();
	DBG("Got VOCODR_PKT_CNTRL_ACK from Vocoder device\n");

	if (encrypted_frame_mode) {
		do {
			enc_dataIn->getData(rx_data_in);
			pcPkt = (CryptoPacket *)rx_data_in->get_buffer();
		} while (pcPkt->hdr != CRYPTO_PKT_CNTRL_ACK);
		enc_dataIn->bufferEmptied();
		DBG("Got CRYPTO_PKT_CNTRL_ACK from Encryption device\n");
	}

	if (mController->get_radio_activity() != RX_ACTIVITY) {
		// Tell ui_processData that ack has been received
		sem_post(&sem_ui_rx_ack);
	}
}

//-------------------------------------------------------------------------
/**
 * @details The main receiver process. It decodes one frame at a time by
 * 			checking for a valid Frame Sync. If a valid frame sync is found
 * 			it goes on to decode the Network Identifier (NID) which contains
 * 			the Data Unit ID (DUID) that identifies the frame. Upon successful
 * 			decoding and extraction of the DUID, the appropriate decode
 * 			function for the frame type is called to decode the frame. At
 * 			the end of a full frame decode, the UInterface is updated with
 * 			parameters extracted from the frame.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::run_rx()
{
    int     res;
    char    duid;
    char 	prev_duid;
    unsigned char	prev_algid;
    char    bits[MAX_FRAME_SIZE];

    struct 	bit_fifo *bf;
    char	lc[9];
    char	lc_orig[9];
    char	es[12];

    char 	vc[VC_BLEN*9];
    char 	lsd[2];

    // The following variables are used for decryption purposes
    char	mi[72];			// Message Indicator
    char	algid[8];		// Algorithm ID
    char	kid[16];		// Key ID
    char	tgid[16];		// Talk Group ID

	PortTypes::CharSequence_var buffer_var;

	CryptoPacket cPkt;
	UIControlPacket ui_pkt;

	bool	lc_protected;

    bf = bf_alloc(HEXBIT*16);
    /**
     * @todo: 	Change while(true) to depend on radio state so this thread
     * 			can be shutdown.
     */
    // While radio is in the running state
    while (mController->get_radio_state() == RUNNING_STATE)
    {
    	// and the current radio activity is receive activity
        if (mController->get_radio_activity() == RX_ACTIVITY) {
        	// Initialize the Receiver State variables
            rx_initialize();
            DBG("Receiver started....\n");

            encrypted_frame_mode = false;
            prev_duid = -1;
            prev_algid = 0;

            while((mController->get_radio_activity() == RX_ACTIVITY)) {
                frame_id++;

                DBG("prev_algid = %x\n",prev_algid);

                DBG("Waiting for FS_BLEN bits\n");
				DBG("Rx bit count = %d\n", rx_bit_count);
                res = get_bits(bits, FS_BLEN);
                DBG("Got some data from Baseband [res = %d]\n", res);
                if (res < FS_BLEN) {
                    ERROR("Invalid FS Length read!!\n");
                    continue;
                }

                DBG("Checking FS\n");
                res = check_fs(bits);
                if (res < 0 ) {
                    ERROR("check FS failed at frame %d: %d\n", frame_id, res);
					rx_bit_count = 0;
					//sendstate_rx_ui(FLUSH);
                    continue;
                }

                res = get_bits(bits, NID_BLEN);
                if (res < NID_BLEN) {
                    ERROR("Invalid NID Length read!!\n");
                    continue;
                }

                DBG("Decoding NID\n");
                res = nid_dec(bits, &duid);
                if (res < 0) {
                    ERROR("decode NID failed at frame %d: %d\n", frame_id, res);
                    continue;
                }

                switch (duid) {
                    case DUID_HDU:
                        DBG("Decoding HDU\n");

                        if (prev_duid != -1 && (prev_duid != DUID_LDU2)) {
                        	DBG("Got HDU when expecting LDU2. Forcing fake LDU2\n");
                        	duid = DUID_TDU_LCW;
                        	memset(vc,0,sizeof(vc));
                        	play_vcw(vc, sizeof(vc));
                        }


                        res = get_bits(bits, HCW_BLEN+HDU_PAD_BLEN);
                        if (res == HCW_BLEN+HDU_PAD_BLEN) {
                            res = hdu_dec(bits, mi, algid, kid, tgid);
                            if (res < 0) {
                                ERROR("HDU decode failed at frame %d: %d\n",
                                    frame_id, res);
                            }
                        	// Send Received ALGID to Vocoder Device to set appropriate encryption/decryption states.
							DBG("Sending received ALGID to Vocoder Device\n");
							sendbuf_rx_vox(VOCODER_PKT_CNTRL_START, (char *)&mr_ui.algId, 1);

							//if (prev_algid != -1 && (mr_ui.algId != prev_algid))
								//duid = DUID_TDU_LCW;

							prev_algid = mr_ui.algId;
							printf("prev_algid = %x\n",prev_algid);

                            // If ALGID field is != ALGID_UNENCRYPTED, then data is encrypted
                            if (mr_ui.algId != ALGID_UNENCRYPTED) {
                            	// Set state flag
                            	encrypted_frame_mode = true;

                            	DBG("Sending Crypto Cntrl Start\n");
                            	sendbuf_rx_crypto(CRYPTO_PKT_CNTRL_START);

                            	// Send received MI to crypto process
                            	int idx;
                            	unsigned char packed_mi[8];
                            	memset(packed_mi, 0, sizeof(packed_mi));
                            	for (int i = 0; i < 64; i++) {
                            			idx = 64-1-i;
                            			packed_mi[i/8] = packed_mi[i/8] | (mi[i] << (idx%8));
                            	}
								DBG("Sending MI_HDU to encryption device\n");
								sendbuf_rx_crypto(CRYPTO_PKT_MI_HDU, (char *)packed_mi, sizeof(packed_mi));
                            } else
                            	encrypted_frame_mode = false;

                        } else {
                            ERROR("Invalid HDU Length read!!\n");
                        }
                        break;

                    case DUID_TDU:
                      DBG("Decoding TDU\n");

                        if (prev_duid == DUID_LDU1) {
                        	DBG("Previous duid == LDU1, Forcing fake LDU2 bits\n");

//							INFO("\nlsd2 [encrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);

							// Send Received ALGID to Vocoder Device to set appropriate encryption/decryption states.
							DBG("Sending received ALGID to Vocoder Device\n");
							sendbuf_rx_vox(VOCODER_PKT_CNTRL_START, (char *)&mr_ui.algId, 1);

							// If ALGID field is != ALGID_UNENCRYPTED, then data is encrypted
							if (mr_ui.algId != ALGID_UNENCRYPTED) {
								// Set state flag
								if (!encrypted_frame_mode) {
									encrypted_frame_mode = true;
									DBG("Sending Crypto Cntrl Start\n");
									sendbuf_rx_crypto(CRYPTO_PKT_CNTRL_START);
								}
								// Send LSD to crypto
								DBG("LDU2: Sending LSD for decoding\n");
								sendbuf_rx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));

								// Send received MI to crypto process
								DBG("Sending MI_LDU to encryption device\n");
								sendbuf_rx_crypto(CRYPTO_PKT_MI_LDU, (char *)mr_ui.mi, sizeof(mr_ui.mi)-1);
							} else
								encrypted_frame_mode = false;

							// Send VCW to vocoder
							play_vcw(vc, sizeof(vc));

							if (encrypted_frame_mode) {
								do {
									DBG("Waiting for LSD2\n");
									recvbuf_rx_crypto((char *)&cPkt, sizeof(cPkt));
								} while (cPkt.hdr != CRYPTO_PKT_LSD);
								memcpy(lsd,cPkt.data,cPkt.len);
								DBG("Got LSD2\n");
							}
//							INFO("\nlsd2 [decrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);
                        }

                        res = get_bits(bits, TDU_PAD_BLEN);
                        if (res == TDU_PAD_BLEN) {
                            tdu_dec(bits);
                        }

        				// Send SCAN_CONTINUE to UI
        				DBG("Sending SCAN_CONTINUE to UI device\n");
        				sendstate_rx_ui(SCAN_CONTINUE);

                        break;

                    case DUID_TDU_LCW:
                        DBG("Decoding TDU_LCW\n");

                    	if (prev_duid == DUID_LDU1) {
                      	DBG("Previous frame == LDU1, Forcing fake LDU2 bits\n");

//							INFO("\nlsd2 [encrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);

                    		// Send Received ALGID to Vocoder Device to set appropriate encryption/decryption states.
							DBG("Sending received ALGID to Vocoder Device\n");
							sendbuf_rx_vox(VOCODER_PKT_CNTRL_START, (char *)&mr_ui.algId, 1);

                    		// If ALGID field is != ALGID_UNENCRYPTED, then data is encrypted
                    		if (mr_ui.algId != ALGID_UNENCRYPTED) {
                    			// Set state flag
                    			if (!encrypted_frame_mode) {
                    				encrypted_frame_mode = true;
                    				DBG("Sending Crypto Cntrl Start\n");
                    				sendbuf_rx_crypto(CRYPTO_PKT_CNTRL_START);
                    			}
                    			// Send LSD to crypto
                    			DBG("LDU2: Sending LSD for decoding\n");
                    			sendbuf_rx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));

                    			// Send received MI to crypto process
								DBG("Sending MI_LDU to encryption device\n");
                    			sendbuf_rx_crypto(CRYPTO_PKT_MI_LDU, (char *) mr_ui.mi, sizeof(mr_ui.mi)-1);
                    		} else
                    			encrypted_frame_mode = false;

                    		// Send VCW to vocoder
                    		play_vcw(vc, sizeof(vc));

                    		if (encrypted_frame_mode) {
                    			do {
                    				recvbuf_rx_crypto((char *)&cPkt, sizeof(cPkt));
                    			} while (cPkt.hdr != CRYPTO_PKT_LSD);
                    			memcpy(lsd,cPkt.data,cPkt.len);
								DBG("Got LSD2\n");
                    		}
//							INFO("\nlsd2 [decrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);
                    	}

                        res = get_bits(bits, TDU_LC_BLEN+TDU_LCW_PAD_BLEN);
                        if (res == TDU_LC_BLEN+TDU_LCW_PAD_BLEN) {
                            res = tdu_lcw_dec(bits);
                            if (res < 0) {
                                ERROR("TDU_LCW decode failed at frame %d: %d\n",
                                frame_id, res);
                            }
                        } else {
                            ERROR("Invalid TDU_LCW Length read!!\n");
                        }

        				// Send SCAN_CONTINUE to UI
        				DBG("Sending SCAN_CONTINUE to UI device\n");
        				sendstate_rx_ui(SCAN_CONTINUE);

                        break;

                    case DUID_LDU1:
                    	// Late entries should only occur in LDU2 frames because
                    	// the AlgId is required to determine which mode to set
                    	// the vocoder/encryption devices.
                        DBG("Decoding LDU1\n");
                        res = get_bits(bits, LDU_VC_BLEN);

                        if (prev_duid != DUID_HDU && prev_duid != DUID_LDU2)
                        	continue;

                        if (res == LDU_VC_BLEN) {
                        	memset(lc, 0, sizeof(lc));
                            ldu_dec(bits, vc, lc, lsd, duid);

							bf_reset(bf);
                            bf_import(bf, lc, 1);
							if (!(lc_protected = is_lc_protected(bf)))
								memcpy(lc_orig, lc, sizeof(lc));

							if (encrypted_frame_mode) {
								// Send received LC to crypto process
								DBG("Sending LC to encryption device\n");
								sendbuf_rx_crypto(CRYPTO_PKT_LC, lc, sizeof(lc)-1);

								// Send received LC to crypto process
								DBG("Sending LSD to encryption device\n");
								sendbuf_rx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));
							}

                            // Send VCW to vocoder
                            play_vcw(vc, sizeof(vc));

							if (encrypted_frame_mode) {
								// Receive decrypted LC from crypto process
								do {
									recvbuf_rx_crypto((char *)&cPkt, sizeof(cPkt));
								} while (cPkt.hdr != CRYPTO_PKT_LC);
								memcpy(lc,cPkt.data,cPkt.len);
								DBG("Got LC\n");

								// Receive decrypted LSD from crypto process
								do {
									recvbuf_rx_crypto((char *)&cPkt, sizeof(cPkt));
								} while (cPkt.hdr != CRYPTO_PKT_LSD);
								memcpy(lsd,cPkt.data,cPkt.len);
								DBG("Got LSD1\n");

//								INFO("\nlsd1 [decrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);
							}
							// Extract received fields from LC
							bf_reset(bf);

							// If lc was not encrypted, ignore decrypted lc
							if (!lc_protected)
								memcpy(lc,lc_orig,sizeof(lc_orig));

							// import lc into bit fifo and remove first bit since lc_deformat does not look at the protected bit
                            bf_import(bf,lc,HEXBIT*12);
							bf_get_n(bf,lc_orig,1);
                            lc_deformat(bf);

                        } else {
                            ERROR("length not equal at frame %d\n", frame_id);
                        }
                        break;
                    case DUID_LDU2:
                        DBG("Decoding LDU2\n");
                        res = get_bits(bits, LDU_VC_BLEN);
                        DBG("Got LDU2 bits\n");

                        if (res == LDU_VC_BLEN) {
                            ldu_dec(bits, vc, es, lsd, duid);

//                          INFO("\nlsd2 [encrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);

                            // Extract received fields from ES
							bf_reset(bf);
                            bf_import(bf, es, HEXBIT*16);
                            es_deformat(bf);

                        	// Send Received ALGID to Vocoder Device to set appropriate encryption/decryption states.
							DBG("Sending received ALGID to Vocoder Device\n");
							sendbuf_rx_vox(VOCODER_PKT_CNTRL_START, (char *)&mr_ui.algId, 1);

							if (prev_algid != 0 && (mr_ui.algId != prev_algid)) {
								printf("prev_algid = %x, mr_algid = %x", prev_algid, mr_ui.algId);
								DBG("\n\n !!!!!!!!!!!!!!ALGID switched in the middle of a transmission!!!!!!!!!!!\n\n");
								duid = DUID_TDU_LCW;
								mr_ui.algId = prev_algid;
							}

							prev_algid = mr_ui.algId;

							// If ALGID field is != ALGID_UNENCRYPTED, then data is encrypted
							if (mr_ui.algId != ALGID_UNENCRYPTED) {
								// Set state flag
								if (!encrypted_frame_mode) {
									encrypted_frame_mode = true;
								   	sendbuf_rx_crypto(CRYPTO_PKT_CNTRL_START);
								}
								if (prev_duid == DUID_HDU || prev_duid == DUID_LDU1)
									sendbuf_rx_crypto(CRYPTO_PKT_LSD, lsd, sizeof(lsd));

								sendbuf_rx_crypto(CRYPTO_PKT_MI_LDU, (char *)mr_ui.mi, sizeof(mr_ui.mi)-1);
							} else
								encrypted_frame_mode = false;

							if (prev_duid == DUID_HDU || prev_duid == DUID_LDU1)
								play_vcw(vc, sizeof(vc));

                            if (encrypted_frame_mode) {
                            	if (prev_duid == DUID_HDU || prev_duid == DUID_LDU1) {
									do {
										recvbuf_rx_crypto((char *)&cPkt, sizeof(cPkt));
									} while (cPkt.hdr != CRYPTO_PKT_LSD);
									memcpy(lsd,cPkt.data,cPkt.len);
									DBG("Got LSD2\n");
                            	}
                            }

                            DBG("\nlsd2 [decrypted] = 0x%02X%02X\n", (unsigned char)lsd[0], (unsigned char)lsd[1]);
                        } else {
                            ERROR("length not equal at frame %d\n", frame_id);
                        }
                        break;

                    case DUID_PDU:
                        //INFO("PDU duid not supported yet at frame %d\n", duid);
                        break;

                    case DUID_TSB:
                        //INFO("TSB duid not supported yet at frame %d\n", duid);
                        break;

                    default:
                        ERROR("duid %d error at frame %d\n", duid, frame_id);
                        break;
                }

                prev_duid = duid;
                // update user interface object
                sendstate_rx_ui(UPDATE);

				if (duid == DUID_TDU || duid == DUID_TDU_LCW)
					break;
            }

            rx_shutdown();
            DBG("Finished Receiving....\n");
        } else {
        	// wait for START from UI component

        	sem_wait(&sem_run_rx);
        }
    }

    bf_free(bf);
    INFO("Quitting receiver....\n");
}

//-------------------------------------------------------------------------
/**
 * @details Gets data from buffer based on a producer-consumer semaphore model.
 * @param buf	Data buffer.
 * @param len	Length of data.
 * @return		Amount actually read from buffer.
 *
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::bs_input(char *buf, unsigned int len)
{
    int res, bf_isempty, bufspace;
    //INFO("bs_input()\n");
	
	bufspace = 0;

    // Block on semaphore if buffer is empty.
    pthread_mutex_lock(&vc_rx_bf_m);
    bf_isempty = bf_empty(rx_bf) || ((unsigned int)bf_size(rx_bf) < len);
    pthread_mutex_unlock(&vc_rx_bf_m);

    while (bf_isempty && !rx_stop_requested) {
		if (sem_wait(&m_bf_empty) < 0) {
			ERROR("error in sem_wait\n");
			continue;
		}

		pthread_mutex_lock(&vc_rx_bf_m);
		bf_isempty = bf_empty(rx_bf) || ((unsigned int)bf_size(rx_bf) < len);
		pthread_mutex_unlock(&vc_rx_bf_m);
    }

    // Retrieve requested data from buffer.
    res = 0;
    if (len > 0) {
		pthread_mutex_lock(&vc_rx_bf_m);
		res = bf_get_n(rx_bf, buf, len);
		bufspace = bf_space(rx_bf);
		pthread_mutex_unlock(&vc_rx_bf_m);
    }

    // Notify the producer that the buffer is no longer full
    // provided a nonzero datasize was retrieved from the buffer.
    if (bufspace > 0) {
		if (sem_post(&m_bf_full) < 0)
			ERROR("error in sem_post\n");
    }

    return res;
}


//-------------------------------------------------------------------------
/**
 * @details		Sends control information and data to Cryptography component.
 * @param hdr	Header of packet.
 * @param data	Defaults to NULL if only control message needs to be sent.
 * 				Contains data based on header otherwise.
 * @param len	Defaults to 0 if only header control needs to be sent.
 * 				Contains length of data otherwise.
 * @author		Justyn Bell <jwbell@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::sendbuf_rx_crypto(short hdr, char *data, int len)
{
	DBG("Sending payload of size %d bytes to be decrypted...\n", len);
	CryptoPacket c_pkt;
	int CPSize = sizeof(CryptoPacket);

	c_pkt.hdr = hdr;
	c_pkt.len = ((data == NULL) ? 0 : len);

	memcpy((char *)&c_pkt.data, data, len);
	DBG("Sending cntrl 0x%04X, len %d\n", c_pkt.hdr, c_pkt.len);

	PortTypes::CharSequence_var buffer_var;
	buffer_var = new PortTypes::CharSequence(CPSize, CPSize, (CORBA::Char *)&c_pkt, 0);
	enc_dataOut->pushPacket(buffer_var);
}

//-------------------------------------------------------------------------
/**
 * @details		Receives data from Cryptography component
 * @param buf	Buffer where data is stored
 * @param len	Length of buffer
 *
 * @author		Justyn Bell <jwbell@ucsd.edu>
 */
//-------------------------------------------------------------------------
int p25_Receiver_i::recvbuf_rx_crypto(char *buf, int len)
{
    PortTypes::CharSequence *enc_data(NULL);
    CORBA::UShort enc_data_length;

    int res;
    char *data;

    enc_dataIn->getData(enc_data);
    enc_data_length = enc_data->length();

    data = (char *)(enc_data->get_buffer());

    if (enc_data_length != len) {
    	ERROR("Houston, we have a problem:  enc_data_length = %d, len = %d\n", enc_data_length, len);
    	res = 0;
    }
    else {
    	memcpy(buf, data, len);
    	res = len;
    }
    enc_dataIn->bufferEmptied();

    return res;
}

//-------------------------------------------------------------------------
/**
 * @details		Sends control information and data to Vocoder component.
 * @param hdr	Header of packet.
 * @param data	Defaults to NULL if only control message needs to be sent.
 * 				Contains data based on header otherwise.
 * @param len	Defaults to 0 if only header control needs to be sent.
 * 				Contains length of data otherwise.
 * @author		Justyn Bell <jwbell@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::sendbuf_rx_vox(short hdr, char *data, int len)
{
	PortTypes::CharSequence_var buffer_var;
	VocoderPacket v_pkt;

	v_pkt.hdr = hdr;
	v_pkt.len = ((data == NULL) ? 0 : len);
	memcpy((char *)&v_pkt.data, data, len);
	//v_pkt.data[0] = data;

	buffer_var = new PortTypes::CharSequence(sizeof(VocoderPacket),
			sizeof(VocoderPacket), (CORBA::Char *)&v_pkt, 0);
	voc_out_data->pushPacket(buffer_var);
}

//-------------------------------------------------------------------------
/**
 * @details		Sends control information and data to UI component.
 * @param hdr	Header of packet.

 * @author		Justyn Bell <jwbell@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
void p25_Receiver_i::sendstate_rx_ui(short hdr)
{
	PortTypes::CharSequence_var buffer_var;
	UIControlPacket ui_pkt;

	ui_pkt.hdr = hdr;
	if (hdr == UPDATE)
		memcpy((char *)&ui_pkt.ui, (char *)&mr_ui, sizeof(ui_pkt.ui));

	buffer_var = new PortTypes::CharSequence(sizeof(UIControlPacket),
			sizeof(UIControlPacket), (CORBA::Char *)&ui_pkt, 0);
	ui_dataOut->pushPacket(buffer_var);
}

//-------------------------------------------------------------------------
/**
 * @details		Sends control information and data to UI component.
 * @param hdr	Header of packet. If DATA, user interface is updated.

 * @author		Justyn Bell <jwbell@ucsd.edu>
 *
 */
//-------------------------------------------------------------------------
/*
void p25_Receiver_i::r_ui_update()
{
	PortTypes::CharSequence_var buffer_var;
	UIControlPacket uPkt;

	uPkt.hdr = DATA;
	memcpy((char *)&uPkt.ui, (char *)&mr_ui, sizeof(uPkt.ui));

	buffer_var = new PortTypes::CharSequence(sizeof(uPkt), sizeof(uPkt), (CORBA::Char *)&uPkt, 0);
	ui_dataOut->pushPacket(buffer_var);
}
*/
