/****************************************************************************
Copyright 2007 Virginia Polytechnic Institute and State University
This file is part of the OSSIE __COMP_NAME__.
OSSIE __COMP_NAME__ is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
"(at your option) any later version
OSSIE __COMP_NAME__ is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with OSSIE __COMP_NAME__; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

/**
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 */

#ifndef P25_TRANSMITTER_IMPL_H
#define P25_TRANSMITTER_IMPL_H

#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "ossie/cf.h"
#include "ossie/PortTypes.h"
#include "ossie/Resource_impl.h"
#include "ossie/debug.h"

#include "standardinterfaces/realChar.h"
#include "standardinterfaces/realChar_u.h"
#include "standardinterfaces/realChar_p.h"

/* user includes */
#include "p25/config.h"
#include "p25/controller.h"
#include "p25/format/packets.h"

extern "C" {
    #include "p25/format/bit_fifo.h"
}



class p25_Transmitter_i : public virtual Resource_impl
{
  public:
    /** @brief Initializing constructor */
    p25_Transmitter_i(const char *uuid, omni_condition *sem);

    /** @brief Destructor */
    ~p25_Transmitter_i(void);

    /** @brief Static function for omni thread */
    static void Run( void * data );

    /** @brief Static function for omni thread */
    static void ui_Run( void * data );

    /** @brief Static function for omni thread */
    static void tx_Run( void * data );

    /** @brief */
    void start() throw (CF::Resource::StartError, CORBA::SystemException);

    /** @brief */
    void stop() throw (CF::Resource::StopError, CORBA::SystemException);

    /** @brief */
    CORBA::Object_ptr getPort( const char* portName )
        throw (CF::PortSupplier::UnknownPort, CORBA::SystemException);

    /** @brief */
    void releaseObject() throw (CF::LifeCycle::ReleaseError, CORBA::SystemException);

    /** @brief */
    void initialize() throw (CF::LifeCycle::InitializeError, CORBA::SystemException);

	/** @brief Query properties from .prf.xml */
	void query(CF::Properties&)
		throw (CORBA::SystemException,
				CF::UnknownProperties);

    /** @brief Configures properties read from .prf.xml */
    void configure(const CF::Properties&)
        throw (CORBA::SystemException,
            CF::PropertySet::InvalidConfiguration,
            CF::PropertySet::PartialConfiguration);


	/** @brief Encodes reused P25 frame components. */
	void	enc_speedup();

	/**
	 * @brief 	Prepares the parts of the Link Control for the specified
	 * 			Link Control Option (LCO).
	 */
	void	lc_format(struct bit_fifo* lc, int lco);

	/** @brief Get one Voice Code Word (VCW) from the Vocoder. */
	void	get_vcw(char* vcw);

	/** @brief Puts the specified bits in the bit stream. */
	void	send_to_bitstream(char* bits, int len);

	/**
	 * @brief 	Encodes the Link Control Word (LCW).
	 */
	void	lcw_enc(struct bit_fifo* lc);

	/** @brief Encodes the Encryption Synchronization Word. */
	void	esw_enc(void);

	/** @brief Encodes the Network Identifier (NID). */
	void	nid_enc(char* nid, char duid, char par_bit);

	/** @brief Encodes the Header Data Unit (HDU) frame. */
	void	hdu_enc();

	/**
	 * @brief 	Encodes the Logical Data Unit (LDU) frames which carry
	 * 			the voice co_processingdewords.
	 */
	void	ldu_enc(char type, char *lsd);

	/** @brief Encodes the Termination Data Unit (TDU) frame. */
	void	tdu_enc(char type);

	/** @brief Prepares the Transmitter for transmission. */
	void	tx_initialize();

	/** @brief Begin encoding and transmitting P25 frames. */
	void	run();

	/** @brief Gracefully ends transmissions and shuts down the Transmitter. */
	void	tx_shutdown();


  private:
	/** @brief Disallow default constructor */
    p25_Transmitter_i();

    /** @brief Disallow copy constructor */
    p25_Transmitter_i(p25_Transmitter_i&);

    /** @brief Main signal processing method */
    void ProcessData();
	bool continue_processing();
	volatile bool thread_started;
	omni_mutex processing_mutex;

	/** @brief UI signal processing method */
    void ui_ProcessData();
	bool ui_continue_processing();
	volatile bool ui_thread_started;
	omni_mutex ui_processing_mutex;

	/** @brief Tx signal processing method */
	volatile bool tx_thread_started;
	omni_mutex tx_processing_mutex;

    omni_condition *component_running;  ///< for component shutdown
    omni_thread *processing_thread;     ///< for component writer function
    omni_thread *ui_processing_thread;     ///< for component writer function
    omni_thread *tx_processing_thread;     ///< for component writer function


    /** @brief List of components provides and uses ports */
	standardInterfaces_i::realChar_u *modem_data;
	standardInterfaces_i::realChar_p *voc_in_data;
	standardInterfaces_i::realChar_u *voc_out_data;
	standardInterfaces_i::realChar_p *ui_dataIn;
	standardInterfaces_i::realChar_u *ui_dataOut;
	standardInterfaces_i::realChar_u *enc_dataOut;
	standardInterfaces_i::realChar_p *enc_dataIn;

	/**
	 * @brief Variable shared for synchronization of User Interface changes.
	 */
    //UInterface m_ui;
    UIControlPacket m_ui;


	/** @brief	Pre-encoded NID for LDU1 frames.*/
	char	ldu1_nid[NID_BLEN];

	/** @brief	Pre-encoded NID for LDU2 frames.*/
	char	ldu2_nid[NID_BLEN];

	/** @brief	Pre-encoded link control bytes for LDU1 frames.*/
	char	lcw[LCW_BLEN];

	/** @brief	Pre-encoded encryption sync bytes for LDU1 frames.*/
	char	esw[ESW_BLEN];

	/** @brief	Holds status symbols.*/
	char	stat_sym[SS_BLEN];

	/** @brief	Counts the number of bits put on the outgoing channel.*/
	int		bitCount;

	/** @brief	Path to UNIX domain socket for communication with modem.*/
	int		tx_sock;

	/** @brief Link control */
	char lc[9];

	/** @brief	Pointer to Controller object */
	Controller*		mController;

	/** @brief Tx buffer variables */
	struct bit_fifo 	*tx_bf;
	pthread_mutex_t		vc_tx_bf_m;
	sem_t 				m_bf_full;
	sem_t				m_bf_empty;

	/** @brief Tx state machine semaphores */
	sem_t 				sem_run_tx;
	sem_t				sem_ui_tx_ack;
	sem_t				sem_voc_ack;
	sem_t				sem_enc_ack;

	/** @brief Stop requested to Tx component from the UI component */
	bool tx_stop_requested;

	/** @brief Encrypted or non-encrypted frame mode */
	bool encrypted_frame_mode;

	/** @brief Gets voice code word data from buffer */
	int receive_vc_data(char *buf, int len = VC_BLEN);

	/** @brief Sends data from Tx component to Modem component */
	int bs_output(char *buf, unsigned int len);

	/** @brief Sends data from Tx component to Cryptography component */
	int sendbuf_tx_crypto(short hdr, char *data = NULL, int len = 0);

	/** @brief Receives data from Cryptography component */
	int recvbuf_tx_crypto(char *buf, int len);

	/** @brief Sends control data from Tx component to Vocoder component */
	int sendstate_tx_voc(short hdr, char data = '0');

};
#endif

