/**
 * ============================================================================
 * @file	main.c
 * @brief	This file contains implementation of the main processing threads
 * 			that drive the ARM application.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */

/* ----------------------------- Linux Specific ---------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pthread.h>
#include <semaphore.h>

/* ---------------------------- DSP/BIOS Link ------------------------------ */
#include <dsplink.h>

/* ---------------------------- Application specific ----------------------- */
#include "common.h"
#include "arm_ambe.h"
#include "ambe_intf.h"
#include "bs_packet.h"
#include "log_arm.h"

#include "gpp_host_intf.h"

#if defined (__cplusplus)
extern "C" {
#endif /* defined (__cplusplus) */

/**
 * ----------------------------------------------------------------------------
 * @name	opts
 * @brief	Struct to hold options for DVSI AMBE library.
 * ----------------------------------------------------------------------------
 */
PC_OPTIONS opts;

/**
 * ----------------------------------------------------------------------------
 * @name	ao
 * @brief	Struct to hold options for encoding/decoding voice data.
 * @see		opts
 * ----------------------------------------------------------------------------
 */
AMBE_OPTIONS ao;

/**
 * ----------------------------------------------------------------------------
 * @name	arm_mode
 * @brief	Tracks the current operating mode. Values are ARM_NORMAL_MODE or
 * 			ARM_TEST_MODE
 * ----------------------------------------------------------------------------
 */
short arm_mode;

/**
 * ----------------------------------------------------------------------------
 * @name	arm_running
 * @brief	Flag to track the current running state of ARM processor.
 * 			Values are 1 or 0.
 * 			Accessed by all ARMAMBE_* threads.
 * ----------------------------------------------------------------------------
 */
char arm_running;

/**
 * ----------------------------------------------------------------------------
 * @name	dsp_ack_stop
 * @brief	Flag to indicate that DSP has sent an ACK to a stop request message
 * 			Values are 1 or 0.
 * 			Accessed by ARMAMBE_ReceiveDspTask
 * ----------------------------------------------------------------------------
 */
char dsp_ack_stop;

/**
 * ----------------------------------------------------------------------------
 * @name	sem_processDspArm
 * @brief	Semaphore used by tskProcessDspArm
 * ----------------------------------------------------------------------------
 */
sem_t sem_processDspArm;

/**
 * ----------------------------------------------------------------------------
 * @name	sem_processCryptography
 * @brief	Semaphore used by tskProcessCryptography
 * ----------------------------------------------------------------------------
 */
sem_t sem_processCryptography;

/**
 * ----------------------------------------------------------------------------
 * @func	process_file
 * @brief	Processes test file input.
 * @note	For testing purposes only
 * @arg		argc
 * 				argument count
 * 			argv
 * 				array of arguments
 * @author	DVSI
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
int process_file(int argc, char *argv[]) {
	DSP_STATUS status;

	// Extract options from argv
	get_options(&opts, argc, argv);

	// Copy options from PC_OPTIONS struct to AMBE_OPTIONS struct that gets
	// sent to DSP
	ao.verbose		= opts.verbose;
	ao.run_encoder	= opts.run_encoder;
	ao.run_decoder	= opts.run_decoder;
	ao.mode			= opts.mode;
	ao.fec_mode		= opts.fec_mode;			
	ao.rep_mode		= opts.rep_mode;
	ao.enc_ws		= opts.enc_ws;
	ao.dec_ws		= opts.dec_ws;
	ao.vad_level	= opts.vad_level;
	ao.nsprsn_enabled	= opts.nsprsn_enabled;
	ao.tone_det_enabled	= opts.tone_det_enabled;
	ao.tone_on		= opts.tone_on;
	ao.tone_amp		= opts.tone_amp;
	ao.tone_index	= opts.tone_index;
	ao.sd_bits		= opts.sd_bits;
	ao.sd_dist		= opts.sd_dist;
	ao.hd_errs		= opts.hd_errs;
	ao.dtx			= opts.dtx;
	ao.bit_steal	= opts.bit_steal;
	ao.cnirate		= opts.cnirate;
	ao.agc_on		= opts.agc_on;
	ao.use_fec		= opts.use_fec ;
	ao.arm_mode		= arm_mode;
	ao.encryption		= 1;

	DBG("ARM MODE = %d\n",ao.arm_mode);
	DBG("run_encoder = %d, run_decoder = %d\n", ao.run_encoder, ao.run_decoder);

	// Wake DSP<->ARM processing task
	sem_post(&sem_processDspArm);

	// If encryption enabled wake encryption processing task
	if (ao.encryption)
		sem_post(&sem_processCryptography);

	// Send Start control message;
	ARMAMBE_SendBuffer((char *)&ao, sizeof(ao), 1);	

	// Start the Test sending process. Sending process runs in a loop until
	// an EOF is reached then it returns.
	process_test_input(&opts);

	// Send Stop control message [ao->run_encoder = 0 and ao->run_decoder = 0];
	ao.run_encoder = 0;
	ao.run_decoder = 0;
	ARMAMBE_SendBuffer((char *)&ao, sizeof(ao), 1);	

	// Waits for the DSP to send a control message which serves as an ACK.
	while (!dsp_ack_stop)
		sched_yield();

	// reset ack flag
	dsp_ack_stop = 0;

	// close opened files
	ambe_close_files(&opts);
	return (0);
}

/**
 * ----------------------------------------------------------------------------
 * @func	process_inputfile
 * @brief	Processes test file input.
 * @note	For testing purposes only
 * @arg		inputfile
 * 				file to read commandline parameters from
 * @author	DVSI
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
int process_inputfile(char * inputfile) {
	char line[200];
	FILE *fpvector;
	char *argv[10];	// allow up to 10 arguments
	int argc;

	fpvector = fopen(inputfile, "r" );

	if (!fpvector) {
		ERROR("couldn't open file %s\n", inputfile);
		exit(0);
	}

	while( getnextline( fpvector, line ) ) {
		char **argvp = &argv[0];
		char *lp = line;
		char c;

		if (line[0] == '#') {
			INFO("(comment)=%s\n", line+1 );
			//this is helpful for debugging
			if (!strcmp("#done", line )) {
				break;
				exit(0);
			}
		} else {
			// now ignores anything followed by ; on the command line
			INFO("COMMAND LINE [ %s ]\n", line );
			*argvp++ = "arm_ambe";
			argc = 1;
			c = *lp++;
			while ((c != 0) && (c != ';')) {
				*argvp++ = lp-1;
				while (!isspace(c) && (c != 0) && (c != ';'))
					c = *lp++;

				lp--;
				*lp++ = 0;
				argc++;

				while (isspace(c))
					c = *lp++;

				//end-of-word
			}
			process_file( argc, argv );
		}
	}
	INFO("Finished\n");

	return( 0 );
}

/**
 * ----------------------------------------------------------------------------
 * @func	tskProcessDspArm
 * @details	Thread (task) for handling the processing of cryptography voice
 * 			data. It calls the appropriate handling routine based on the
 * 			current mode of the ARM. (normal or test mode)
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
void *tskProcessDspArm (void *arg) 
{
	DBG("Starting tskProcessDspArm\n") ;

	while (arm_running) {

		sem_wait(&sem_processDspArm);

		// Exit if ARM has been shutdown before thread was awoken
		if (!arm_running)
			break;
		
		//INFO("arm_mode = %d, ao.encode = %d, ao.decode = %d\n",arm_mode,ao.run_encoder,ao.run_decoder);
		switch (arm_mode) {
		case ARM_NORMAL_MODE:
			// Receive tasks should be handled in a loop that only
			// kicks out when the number of expected receive bytes
			// is equal to the actual received bytes or 
			// ARMAMBE_ReceiveBuffer() returns 0
			gh_intf_output(&ao); 
			break;
		case ARM_TEST_MODE:
			process_test_output(&opts);
			break;
		}
		INFO("DSP has finished Processing\n");
	}

	INFO("Ending tskProcessDspArm\n") ;

	return (void *) 0;
}

/**
 * ----------------------------------------------------------------------------
 * @func	tskProcessCryptography
 * @details	Thread (task) for handling the processing of cryptography related
 * 			voice data. It calls the appropriate handling routine based on the
 * 			current mode of the ARM. (normal or test mode)
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
void *tskProcessCryptography (void *arg) 
{
	DBG("Starting tskProcessCryptography\n") ;

	while (arm_running) {

		sem_wait(&sem_processCryptography);

		// Exit if ARM has been shutdown before thread was awoken
		if (!arm_running)
			break;

		INFO("arm_mode = %d, ao.encryption = %d\n",arm_mode,ao.encryption);
		INFO("In encryption Mode\n");

		switch (arm_mode) {
			case ARM_NORMAL_MODE:
				// Receive tasks should be handled in a loop that only
				// kicks out when the number of expected receive bytes
				// is equal to the actual received bytes or 
				// ARMAMBE_ReceiveBuffer() returns 0
				gh_intf_cryptography(&opts);	
				break;
			case ARM_TEST_MODE:
				process_test_cryptography(&opts);
				break;
		}
		DBG("ARM has finished Processing Cryptography\n");
	}

	INFO("Ending tskProcessCryptography\n") ;

	return (void *) 0;
}

/**
 * ----------------------------------------------------------------------------
 * @func	tskProcessCryptLoop
 * @brief	Monitors the Cryptography Connection between HOST and ARM
 * @author	Jeff Cuenco <jcuenco@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
void *tskProcessCryptLoop (void *arg) 
{
	DBG("Starting tskProcessCryptLoop\n") ;

	while (arm_running) {

		switch (arm_mode) {
		case ARM_NORMAL_MODE:
			// Receive tasks should be handled in a loop that only
			// kicks out when the number of expected receive bytes
			// is equal to the actual received bytes or
			// ARMAMBE_ReceiveBuffer() returns 0
			gh_intf_crypt_loop(*((int *)arg));
			break;
		case ARM_TEST_MODE:
			break;
		}
		DBG("ARM has finished Processing Cryptography Loop\n");
	}

	DBG("Ending tskProcessCryptLoop\n") ;

	return (void *) 0;
}


/**
 * ----------------------------------------------------------------------------
 * @func	main
 * @brief	Entry point for the application
 * @details	Creates, Initializes and Instantiates threads to process data
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
int main (int argc, char * argv[])
{
	Char8 * dspExecutable = NULL;
	Char8 * strInputFile = NULL;

	Uint32 bufferSize = CHNL_DATA_BUFFER_SZ;
	Uint8 processorId = 0;
	DSP_STATUS status = DSP_SOK ;

	pthread_attr_t attr;
	pthread_t thr_process_dsp_arm;
	pthread_t thr_process_cryptography;
	pthread_t thr_process_crypt_loop;

	int res;

	// Ports to listen on for data/cryptography related data from HOST
	int port, c_port;

	if (argc < 4 || argc > 5) {
		INFO("Usage : %s <absolute path of DSP executable> "
				"<TCP server stream port> "	
				"[absolute path of input file]\n",
				argv[0]) ;
	} else {
		INFO("=============== Sample Application : GPPAMBE ==========\n") ;

		// Get DSP executable from command line param
		dspExecutable = argv[1];

		// Initialize semaphores
		sem_init(&sem_processDspArm, 0, 0);	
		sem_init(&sem_processCryptography, 0, 0);	

		// Initialize flags
		dsp_ack_stop = 0;
		arm_running = 1;
		arm_mode = ARM_NORMAL_MODE;

		// Read in TCP listening ports from command-line parameters
		sscanf(argv[2], "%d", &port);
		if (port < 1025 || port > 65534) {
			ERROR("ERROR, not a valid port ");
			return 0;
		}		
		sscanf(argv[3], "%d", &c_port);
		if (c_port < 1025 || c_port > 65534) {
			ERROR("ERROR, not a valid port ");
			return 0;
		}		

		if (argc == 5) {
			strInputFile = argv[4];
			arm_mode = ARM_TEST_MODE;
		}

		if (MAX_PROCESSORS > 0) {
			status = ARMAMBE_Main (processorId, bufferSize);

			// Pass the DSP executable file name to ARMAMBE_Create.
			if (DSP_SUCCEEDED (status)) {
				status = ARMAMBE_Create (dspExecutable) ;

				INFO("==== Executing sample for DSP processor Id %d ====\n", processorId) ;
				status = ARMAMBE_Execute();

				if (DSP_SUCCEEDED (status)) {
					pthread_attr_init(&attr);
					pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

					// Thread to process data between DSP and ARM
					res = pthread_create(&thr_process_dsp_arm, &attr, tskProcessDspArm, (void *)0);

					// Thread to process cryptography data between DSP and ARM
					res = pthread_create(&thr_process_cryptography, &attr, tskProcessCryptography, (void *)0);

					// Thread to monitor connection from Host machine to ARM
					res = pthread_create(&thr_process_crypt_loop, &attr, tskProcessCryptLoop, (void *)&c_port);

					switch (arm_mode) {
					case ARM_NORMAL_MODE:
						gh_intf_loop(port);
						break;
					case ARM_TEST_MODE:
						INFO("====== Start Processing input file!=====\n");
						process_inputfile(strInputFile);
						INFO("====== Process input file done!=====\n");
						break;
					}

					// Shutdown sequence
					INFO("Intiating Shutdown sequence\n");
					arm_running = 0;

					pthread_join(thr_process_dsp_arm, NULL);
					pthread_join(thr_process_cryptography, NULL);
					pthread_join(thr_process_crypt_loop, NULL);
				}

				INFO("==== Finished Executing sample for DSP processor Id %d ====\n", processorId) ;

				// Perform cleanup operation.
				ARMAMBE_Delete(processorId);
			}
		}
		INFO ("====================================================\n") ;
	}
	return 0;
}

#if defined (__cplusplus)
}
#endif /* defined (__cplusplus) */
