/**
 * ============================================================================
 * @file	ambe_intf.c
 * @brief	This file contains implementation of the processing functions for
 * 			the AMBE library. This file contains modifications from the test
 * 			file provided by DVSI
 * @note	none.
 *
 * @author	DVSI
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */

/******************************************************************************
 * DVSI CONFIDENTIAL PROPRIETARY
 ******************************************************************************
 * DVSI AMBE+2 Vocoder
 * (c) Copyright, Digital Voice Systems Inc., 2001, 2002
 * All Rights Reserved
 *
 * DVSI claims certain rights, including patent rights under
 * U.S. Patents #6,199,037 B1, #6,161,089, #5,870,045, #5,826,222,
 * #5,754,974, #5,715,365, #5,701,390, #5,649,050, #5,630,011,
 * #5,247,579, and #5,226,084, and under other U.S. and foreign
 * patents and patents pending, in the AMBE^(TM) speech coding
 * technology embodied in this software.  Any use of this software
 * or technology requires a separate written license from DVSI.
 *
 * AMBE is a registered trademark of Digital Voice Systems, Inc.
 * AMBE+ and AMBE+2 are trademarks of Digital Voice Systems, Inc.
 ******************************************************************************
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ambe_intf.h"
#include "dprintf.h"
#include "tskDspAmbeIn.h"
#include "tskDspAmbeOut.h"
#include "log_dsp.h"

unsigned long int max_enc_time = 0L;
unsigned long int max_fecenc_time = 0L;
unsigned long int max_fecdec_time = 0L;
unsigned long int max_dec_time = 0L;

int max_enc_stack = 0;
int max_fecenc_stack = 0;
int max_fecdec_stack = 0;
int max_dec_stack = 0;

ENCSTATE_T *est;
DECSTATE_T *dst;
FECSTATE_T efec;
FECSTATE_T *dfec;

ESTS ests;
DSTS dsts;

ENCSTATE_T f_est, s_est;
DECSTATE_T f_dst, s_dst;
FECSTATE_T f_dfec, s_dfec;


#ifdef MEASURE_STACK_SIZE
int stack_count;
#endif


static short speech_buf[SPEECH_BUF_SZ];
static short chan_buf[CHAN_BUF_SZ];
static short crypt_sent_frame_count;
static short crypt_recv_frame_count;

extern char stop_processing;

/*
 * sets the PCC bits in CSR to enable the cache
 */
void enable_cache( void ) {
	unsigned int csr = CSR;
	csr &= 0xffffff1f;	//clear the PCC bits
	csr |= 0x00000040;	//set the PCC bits to 010
	CSR = csr;
}

int clear_stack( void ) {
	char value_on_stack;
	char *ptr = &value_on_stack - 10;

	/*
	 * sets the remainder of the stack to 0xaa
	 */
	while (ptr >= (char *)_stack)
		*ptr-- = 0xaa;

	return (int) ((&value_on_stack - (char *)_stack));
}

/*
 * returns the number of stack locations that are unused.  Unused bytes
 * are equal to zero since they were cleared using clear_stack.
 */
int get_stack_size( int stack_before ) {
	char *ptr = (char *)_stack;
	int stack_after;
	while (*ptr++ == (char) 0xaa);
	stack_after = ptr - (char *)_stack - 1;
	return stack_before - stack_after;
}

/*
 * use this function to compute the maximum of all measurements.
 */
int get_stack_size_max( int stack_before, int max ) {
	int used = get_stack_size( stack_before );
	if (used > max)
		max = used;
	return max;
}

static short * fxmemcpy( short *out, const short *in, short len ) {
	short i;

	for ( i = 0; i < len; i++ )
		out[i] = in[i];

	return ( out + len );
}

static short * fxmemset( short *out, short val, short len ) {
	short i;

	for ( i = 0; i < len; i++ )
		out[i] = val;

	return ( out + len );
}

void init_ests( ESTS *ests ) {
	ests->cni = 0;
	ests->vad_frames  = 0;
	ests->dtx_frames  = 0;
	ests->tone_frames = 0;
	ests->reset_frames = 0;
	ests->frames = 0;
}

void init_dsts( DSTS *dsts ) {
	dsts->data_invalid_frames = 0;
	dsts->erasure_frames			= 0;
	dsts->silence_frames = 0;
	dsts->tone_frames = 0;
	dsts->reset_frames = 0;
	dsts->frames = 0;
}

/*
 * set encoder channel data output to all ones to signal dtx off
 */
static void signal_dtx_off( short *cbuf, short len ) {
	short i;
	for ( i = 0; i < len; i++ )
		cbuf[i] = 1;
}

/*
 * handle discontionuous transmission and comfort noise update
 */
static void dtx_update( short *cni_count, short acni_frames, short *cbuf, short status, short totalbits ) {
	short cni = *cni_count;

	/*
	 * If VOICE_ACTIVE_FLAG is set the frame must be transmitted
	 * to the decoder.  Note that VOICE_ACTIVE_FLAG is true for
	 * the following frames:
	 * 1.  The frame is a voice frame.
	 * 2.  The frame is the first silence frame after a voice burst.
	 * 3.  The encoder determined that the SID frame needs to be
	 * transmitted because of changes in the background noise.
	 */
	if (status & AMBE_VOICE_ACTIVE_FLAG ) {
		cni = acni_frames;
	} else  {
		/*
		 * if VOICE_ACTIVE_FLAG is not set then it is not neccessary
		 * to transmit the frame.  The code below causes the unrequired
		 * frames to be transmitted at a regular interval.  Every
		 * (acni_frames+1)th frame is transmitted.  For instance:
		 * if acni_frames=0 then all unrequired SID frames are transmitted.  
		 * if acni_frames=1 then every other unrequired SID frame is transmitted
		 * if acni_frames=2 then every third unrequired SID frame is transmitted
		 * ...
		 * acni_frames < 0 is a special case.  If acni_frames < 0 then
		 * unrequired SID frames are never transmitted.  Note that for this
		 * case cni is always equal to acni_frames.
		 */
		if ( cni != 0 ) {
			if ( cni > 0 )
				cni--;
			/*
			 * signal no channel data (DTX off)
			 */
			signal_dtx_off( cbuf, totalbits );
		} else {
			/* 
			 * reset count for next update (in frames) 
			 */
			cni = acni_frames;
		}
	}

	*cni_count = cni;
}


/*
 * look for all ones in the decoder channel data input as dtx off signal
 */
static short check_dtx( short *cbuf, short len, short sd_bits ) {
	short i;
	short allones = 1;

	for ( i = 0; i < len; i++ )
		allones &= (cbuf[i] >> (sd_bits-1));

	return ( allones );
}

static short check_erasure( short *cbuf, short len ) {
	short allzeros = 0;
	short i;

	for ( i = 0; i < len; i++ )
		allzeros |= cbuf[i];

	return (allzeros == 0) ? 1 : 0;
}

/*
 * initialize the encoder
 */
static void init_encoder( ENCSTATE_T *est, FECSTATE_T *efec, ESTS *ests, AMBE_OPTIONS *ao ) {
	ambe_init_enc( est, ao->mode, 1 );
	ambe_init_fec( efec, ao->fec_mode );	

	init_ests( ests );

	{
		if ( ao->tone_on )
			ests->cmode = ambe_tone_xmt( est, ao->tone_index, ao->tone_amp );
		else
			ests->cmode = 0;

		if ( ao->dtx )
			ests->cmode |= AMBE_DTX_ENABLE_FLAG;

		if ( ao->nsprsn_enabled )
			ests->cmode |= AMBE_NS_ENABLE_FLAG;

		if ( ao->tone_det_enabled )
			ests->cmode |= AMBE_TONE_DET_ENABLE_FLAG;

		if ( ao->agc_on)
			ests->cmode |= AMBE_AGC_ON_FLAG;
	}
}

/*
 * initialize the decoder
 */
static void init_decoder( DECSTATE_T *dst, FECSTATE_T *dfec, DSTS *dsts, AMBE_OPTIONS *o ) {
	ambe_init_dec( dst, (o->rep_mode >= 0) ? o->rep_mode : o->mode );
	ambe_init_fec( dfec, o->fec_mode );

	init_dsts( dsts );
}

#ifdef BUILD_REPEATER
static void init_rep_coder( REPENCSTATE_T *est, REPDECSTATE_T *dst, AMBE_OPTIONS *o ) {
	ambe_init_enc_rep( est, o->rep_mode );
	ambe_init_dec_rep( dst, o->mode );
}
#endif

int imbe_voice_enc( AMBE_OPTIONS * ao) {
	short ws;
	short ws0;

	DBG("Encoding\n");

	// Get input speech from ARM. This is a blocking call.
	// If data length returned is 0 or less, then the DSP received a control pkt
	// from the ARM and we must exit the encode function
	if (TSKDSPAMBE_IN_getBuffer((char *)&speech_buf, ao->enc_ws*sizeof(short)) <= 0)
		return -1;

	DBG("Received speech data .....\n");

	ws  = ao->enc_ws;
	ws0 = ws >> 1;
	ws -= ws0;

	CLEAR_STACK();
	start_timer();

	DBG("ambe_voice_enc 1 .....\n");
	// input the 1st speech subframe
	ambe_voice_enc( chan_buf, ao->bit_steal, speech_buf, ws0, ests.cmode, 0, APCO_NS_GAIN_NOM, est );

	DBG("ambe_voice_enc 2 .....\n");
	// input the 2nd speech subframe
	ests.status = ambe_voice_enc( chan_buf, ao->bit_steal, speech_buf + ws0, ws, ests.cmode, 1, APCO_NS_GAIN_NOM, est );

	max_enc_time = get_timer_max( max_enc_time );
	GET_STACK_SIZE_MAX( max_enc_stack );
	CLEAR_STACK();

	DBG("Encoding done .....\n");

	return 0;
}

void imbe_fec_enc( AMBE_OPTIONS * ao) {
	short echanbits;

#ifdef BUILD_REPEATER
	if ( ao->rep_mode >= 0 ) {
		FECINFO fecinfo;
		fxmemset( (short *)(&fecinfo), 0, sizeof(fecinfo)/sizeof(short) );
		ambe_set_rep_fecinfo( dstrep, &fecinfo );
		ambe_voice_rep( chan_buf, ao->rep_mode, ao->bit_steal, estrep, chan_buf, ao->mode, ao->bit_steal, dstrep );
		ambe_set_fec_mode( efec, ao->rep_mode );
	}
#endif

	start_timer();

	if ( ao->use_fec ) {
		DBG("Fec Encode .....\n");
		echanbits = ambe_fec_enc( chan_buf, chan_buf, ao->bit_steal, &efec );
	}

	max_fecenc_time = get_timer_max( max_fecenc_time );
	GET_STACK_SIZE_MAX( max_fecenc_stack );

	if ( ests.status & AMBE_VAD_FLAG )
		ests.vad_frames++;

	if ( !(ests.status & AMBE_VOICE_ACTIVE_FLAG) )
		ests.dtx_frames++;

	if ( ests.status & AMBE_DTMF_FRAME_FLAG ) {
		ests.tone_frames++;
		ambe_get_tone_det_info( &ests.info, est );
	} else {
		ests.info.idx = AMBE_INVALID_TONE;
		ests.info.amp = -90;
	}

	if ( ao->verbose > 1 )
		INFO2( "\n Type: %d Tone: %d",
				ests.status & (AMBE_VAD_FLAG | AMBE_VOICE_ACTIVE_FLAG | AMBE_DTMF_FRAME_FLAG ), (ests.status & AMBE_DTMF_FRAME_FLAG) ? ests.info.idx : -1 );

	if ( ao->dtx )
		dtx_update( &ests.cni, ao->cnirate, chan_buf, ests.status, echanbits );

	// If in ARM_TEST_MODE and run_decoder option is set, this block
	// is skipped since the decoder function will get the encoded bits
	// directly from chan_buf. In ARM_NORMAL_MODE, run_decoder cannot
	// be set since we're already encoding, so this block is executed.
	if (!ao->run_decoder) {
		// Send encoded bits back to GPP
		DBG1("Sending Encoded bits to GPP .....[%d bytes]\n",echanbits*sizeof(short));
		TSKDSPAMBE_OUT_sendBuffer((char *)&chan_buf, echanbits*sizeof(short), 0);
	}
}

int imbe_voice_dec( AMBE_OPTIONS * ao) {
	short ws;
	short ws1;

	ws  = ao->dec_ws;
	ws1 = ws >> 1;
	ws -= ws1;

	GET_STACK_SIZE_MAX( max_fecdec_stack );
	CLEAR_STACK();
	start_timer();

	DBG("ambe_voice_dec 1 .....\n");
	dsts.status = ambe_voice_dec( speech_buf, ws, chan_buf, ao->bit_steal, dsts.cmode, 0, dst );

	DBG("ambe_voice_dec 2 .....\n");
	ambe_voice_dec( speech_buf + ws, ws1, chan_buf, ao->bit_steal, dsts.cmode, 1, dst );

	max_dec_time = get_timer_max( max_dec_time );
	GET_STACK_SIZE_MAX( max_dec_stack );

	if ( !(dsts.status & AMBE_VOICE_ACTIVE_FLAG) )
		dsts.silence_frames++;

	if ( dsts.status & AMBE_DATA_INVALID_FLAG )
		dsts.data_invalid_frames++;

	if ( dsts.status & AMBE_DTMF_FRAME_FLAG ) {
		dsts.tone_frames++;
		ambe_get_tone_rcv_info( &dsts.info, dst );
	} else {
		dsts.info.idx = AMBE_INVALID_TONE;
		dsts.info.amp = -90;
	}

	DBG1("Sending Decoded bits to GPP .....[%d bytes]\n",ao->dec_ws*sizeof(short));
	TSKDSPAMBE_OUT_sendBuffer((char *)&speech_buf, ao->dec_ws*sizeof(short), 0);

	return 0;
}

int imbe_fec_dec( AMBE_OPTIONS * ao, short * dttlbits) {
	//FECINFO fecinfo;
	short _dttlbits;

	{
		if ( ao->tone_on )
			dsts.cmode  = ambe_tone_gen( dst, ao->tone_index,ao->tone_amp );
		else
			dsts.cmode  = 0;
	}

	if ( ao->rep_mode >= 0 ) {
#ifdef BUILD_REPEATER
		if ( ambe_get_rep_dec_mode(dstrep) == APCO_MODE ) {
			_dttlbits  = APCO_SRCBITS;
			if ( ao->use_fec )
				_dttlbits += APCO_FECBITS;
		} else  {
			_dttlbits  = APCO_HR_SRCBITS_MAX;
			if ( ao->use_fec )
				_dttlbits += APCO_HR_FECBITS;
		}
#endif 
	} else {
		if ( ambe_get_dec_mode(dst) == APCO_MODE ) {
			_dttlbits  = APCO_SRCBITS;
			if ( ao->use_fec )
				_dttlbits += APCO_FECBITS;
		} else {
			_dttlbits  = APCO_HR_SRCBITS_MAX;
			if ( ao->use_fec )
				_dttlbits += APCO_HR_FECBITS;
		}
	}

	// If in ARM_TEST_MODE and run_encoder option is set, this block
	// is skipped since the encoded bits are already in chan_buf. In
	// ARM_NORMAL_MODE, run_encoder cannot be set since we're already
	// decoding, so this block is executed.
	if (!ao->run_encoder) {
		// Get input bits from ARM. This is a blocking call.
		// If data length returned is 0 or less, then the DSP received a control pkt
		// from the ARM and we must exit the decode function
		if (TSKDSPAMBE_IN_getBuffer((char *)&chan_buf, _dttlbits*sizeof(short)) <= 0)
			return -1;
	}

	DBG("Received encoded bits .....\n");

	if ( check_dtx( chan_buf, _dttlbits, ao->sd_bits ) )
		dsts.cmode |= AMBE_CNI_FRAME_FLAG;

	if ( check_erasure( chan_buf, _dttlbits ) )
		dsts.cmode |= AMBE_LOST_FRAME_FLAG;

	CLEAR_STACK();
	start_timer();

	if ( ao->use_fec ) {
		DBG("Fec Decode .....\n");
		ambe_fec_dec( chan_buf, ao->bit_steal, chan_buf, ao->sd_bits, dsts.cmode, dfec );
		_dttlbits -= (ambe_get_dec_mode(dst) == APCO_MODE) ? APCO_FECBITS : APCO_HR_FECBITS;
	}

//	ambe_get_fecinfo( &fecinfo, dfec );
//	ambe_set_fecinfo( dst, &fecinfo );

#ifdef BUILD_REPEATER
	if ( ao->rep_mode >= 0 ) {
		ambe_set_rep_fecinfo( dstrep, &fecinfo );
		ambe_voice_rep( chan_buf, ao->rep_mode, ao->bit_steal, estrep, chan_buf, ao->mode, ao->bit_steal, dstrep );
	}
#endif

	*dttlbits = _dttlbits;
	
	max_fecdec_time = get_timer_max( max_fecdec_time );

	return 0;
}

int process_encode_request( AMBE_OPTIONS * ao ) {
	short echanbits;
	int ret;

	est = &f_est;
	//INFO2("$$$$$ SWITCH STATES: est is switched back to f=%p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);

#ifdef BUILD_REPEATER
	estrep = &f_estrep;
#endif

#ifndef OVERALL_MAXTIME
	max_enc_time = max_fecenc_time = 0;
#endif

	if (!stop_processing) {
		//INFO2("$$$$$ IMBE_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
		ret = imbe_voice_enc(ao);
		if (ret < 0) { 
			if (ao->encryption)
				stop_processing = 1;
			else
				return 1;
		}
	}

	if ( ao->encryption ) {
		DBG1("Encryption ... [%d]\n",stop_processing);

		if (!stop_processing) {
			echanbits = (ambe_get_enc_mode(est) == APCO_MODE) ? APCO_SRCBITS : APCO_HR_SRCBITS_MAX;
			TSKDSPAMBE_OUT_sendCryptBuffer((char *)&chan_buf, echanbits*sizeof(short), CRYPT_ENCRYPTION);
			crypt_sent_frame_count++;
			DBG1("################## crypt_sent_frame_count = %d\n\n",crypt_sent_frame_count);
	
			// add counter here to check if we've sent 10 IMBE frames.
			// If so, we're in LDU2 and we need to send the 11th IMBE frame in order for cryptography to continue
			switch (crypt_sent_frame_count) {
				case 1:
				case 10:
					//INFO2("$$$$$ SAVE STATES: copy %p into %p\n\n",est,&s_est);
					memcpy(&s_est,est,sizeof(s_est));			
					//memcpy(&s_estrep,estrep,sizeof(s_estrep));			

					//INFO2("$$$$$ IMBE_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
					ret = imbe_voice_enc(ao);
					if (ret < 0) {
						stop_processing = 1;
						break;
					}

					TSKDSPAMBE_OUT_sendCryptBuffer((char *)&chan_buf, echanbits*sizeof(short), CRYPT_ENCRYPTION);
					crypt_sent_frame_count++;
					DBG1("################## crypt_sent_frame_count = %d\n\n",crypt_sent_frame_count);
					break;
				case 9:
				case 18:
					TSKDSPAMBE_IN_getCryptBuffer((char *)&chan_buf, echanbits*sizeof(short));
					crypt_recv_frame_count++;
					DBG1("################## crypt_recv_frame_count = %d\n\n",crypt_recv_frame_count);

					//est = &s_est;
					//INFO2("$$$$$ SWITCH STATES: est is now pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);

					//estrep = &s_estrep;
					//INFO2("$$$$$ FEC_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
					imbe_fec_enc(ao);

					//INFO2("$$$$$ SAVE STATES: copy %p into %p\n\n",&f_est,&s_est);
					//memcpy(&s_est,&f_est,sizeof(s_est));			
					//memcpy(&s_estrep,&f_estrep,sizeof(s_estrep));			
					break;	
			}
		}

		if (crypt_sent_frame_count != crypt_recv_frame_count) {
			TSKDSPAMBE_IN_getCryptBuffer((char *)&chan_buf, echanbits*sizeof(short));
			crypt_recv_frame_count++;
			DBG2("################## sent = %d, crypt_recv_frame_count = %d\n\n",crypt_sent_frame_count, crypt_recv_frame_count);
			
			// If stop_processing is set, unset because we have one last piece of data to process
			stop_processing = 0;

			//est = &s_est;
			//INFO2("$$$$$ SWITCH STATES: est is now pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
			//estrep = &s_estrep;
		}

		if (crypt_sent_frame_count == 18) {
			crypt_sent_frame_count = 0;
			crypt_recv_frame_count = 0;
		}

		DBG("Encryption done .....\n");
	}


	if (!stop_processing)
		//INFO2("$$$$$ FEC_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
		imbe_fec_enc(ao);

	ests.frames++;

	if (ao->encryption) {
		//INFO2("$$$$$ SAVE STATES: copy %p into %p\n\n",&f_est,&s_est);
		memcpy(&s_est,&f_est,sizeof(s_est));			
		//memcpy(&s_estrep,&f_estrep,sizeof(s_estrep));			
	}
	

#ifdef PRINT_TIMING
	INFO2("##### timing (cycles/sec) [encode]: max_enc_time=%d, max_fecenc_time=%d\n", max_enc_time*50, max_fecenc_time*50);
	INFO1("##### timing (cycles/sec) [total]: max_total_enc_time=%d\n",max_enc_time*50+ max_fecenc_time*50);

#endif
#ifdef MEASURE_STACK_SIZE
	INFO2("$$$$$ stack size (bytes) [encode]: max_enc_stack=%d, max_fecenc_stack=%d\n", max_enc_stack, max_fecenc_stack);
#endif

	if (stop_processing)
		return 1;

	return 0;
}

int process_decode_request( AMBE_OPTIONS * ao ) {
	static FECINFO f_fecinfo;
	static FECINFO s_fecinfo;
	FECINFO *fecinfo;

	short dttlbits;
	int ret;

	dst = &f_dst;
	dfec = &f_dfec;
	fecinfo = &f_fecinfo;

#ifdef BUILD_REPEATER
	dstrep = &f_dstrep;
#endif

#ifndef OVERALL_MAXTIME
	max_dec_time = max_fecdec_time = 0;
#endif

	DBG("Decoding\n");

	if (!stop_processing) {
		//INFO2("$$$$$ IMBE_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
		ret = imbe_fec_dec(ao, &dttlbits);
		if (ret < 0) { 
			if (ao->encryption)
				stop_processing = 1;
			else
				return 1;
		} else {
			ambe_get_fecinfo( fecinfo, dfec );
		}
	}

	if ( ao->encryption ) {
		DBG1("Decryption ... [%d]\n",stop_processing);

		if (!stop_processing) {
			TSKDSPAMBE_OUT_sendCryptBuffer((char *)&chan_buf, dttlbits*sizeof(short), CRYPT_DECRYPTION);
			crypt_sent_frame_count++;
			DBG1("################## crypt_sent_frame_count = %d\n\n",crypt_sent_frame_count);
	
			// add counter here to check if we've sent 10 IMBE frames.
			// If so, we're in LDU2 and we need to send the 11th IMBE frame in order for cryptography to continue
			switch (crypt_sent_frame_count) {
				case 1:
				case 10:
					//INFO2("$$$$$ SAVE STATES: copy %p into %p\n\n",est,&s_est);
					//memcpy(&s_dst,dst,sizeof(s_dst));
					memcpy(&s_fecinfo,fecinfo,sizeof(s_fecinfo));			

					//INFO2("$$$$$ IMBE_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
					ret = imbe_fec_dec(ao, &dttlbits);
					if (ret < 0) {
						stop_processing = 1;
						break;
					} else {
						ambe_get_fecinfo( fecinfo, dfec );
					}

					TSKDSPAMBE_OUT_sendCryptBuffer((char *)&chan_buf, dttlbits*sizeof(short), CRYPT_DECRYPTION);
					crypt_sent_frame_count++;
					DBG1("################## crypt_sent_frame_count = %d\n\n",crypt_sent_frame_count);
					break;
				case 9:
				case 18:
					TSKDSPAMBE_IN_getCryptBuffer((char *)&chan_buf, dttlbits*sizeof(short));
					crypt_recv_frame_count++;
					DBG1("################## crypt_recv_frame_count = %d\n\n",crypt_recv_frame_count);

					//INFO2("$$$$$ FEC_ENC: est is pointing to %p, sent_frame_count = %d\n\n",est,crypt_sent_frame_count);
					fecinfo = &s_fecinfo;
					ambe_set_fecinfo( dst, fecinfo );
					imbe_voice_dec(ao);
					
					memcpy(&s_fecinfo, &f_fecinfo, sizeof(s_fecinfo));
					break;	
			}
		}

		if (crypt_sent_frame_count != crypt_recv_frame_count) {
			TSKDSPAMBE_IN_getCryptBuffer((char *)&chan_buf, dttlbits*sizeof(short));
			crypt_recv_frame_count++;
			DBG2("################## sent = %d, crypt_recv_frame_count = %d\n\n",crypt_sent_frame_count, crypt_recv_frame_count);
			
			// If stop_processing is set, unset because we have one last piece of data to process
			stop_processing = 0;

			fecinfo = &s_fecinfo;
		}

		if (crypt_sent_frame_count == 18) {
			crypt_sent_frame_count = 0;
			crypt_recv_frame_count = 0;
		}

		DBG("Decryption done .....\n");
	}


	if (!stop_processing) {
		ambe_set_fecinfo( dst, fecinfo );
		imbe_voice_dec(ao);
	}

	dsts.frames++;

	if (ao->encryption) {
		//INFO2("$$$$$ SAVE STATES: copy %p into %p\n\n",&f_est,&s_est);
		memcpy(&s_fecinfo,&f_fecinfo,sizeof(s_fecinfo));			
		//memcpy(&s_estrep,&f_estrep,sizeof(s_estrep));			
	}

#ifdef PRINT_TIMING
	INFO2("##### timing (cycles/sec) [decode]: max_fecdec_time=%d, max_dec_time=%d\n", max_fecdec_time*50, max_dec_time*50);
	INFO1("##### timing (cycles/sec) [total]: max_total_decode_time=%d\n", max_fecdec_time*50+ max_dec_time*50);

#endif
#ifdef MEASURE_STACK_SIZE
	INFO2("$$$$$ stack size (bytes) [decode]: max_fecdec_stack=%d, max_dec_stack=%d\n", max_fecdec_stack, max_dec_stack);
#endif

	if (stop_processing)
		return 1;
	else
		return 0;
}


void init_coders(AMBE_OPTIONS * ao) 
{
	init_encoder(&f_est, &efec, &ests, ao);
	init_decoder(&f_dst, &f_dfec, &dsts, ao);

	//init_encoder(&s_est, &s_efec, &s_ests, ao);
	//init_decoder(&s_dst, &s_dfec, &s_dsts, ao);

#ifdef BUILD_REPEATER
	if ( ao->rep_mode >= 0 ) {
		init_rep_coder( &f_estrep, &f_dstrep, ao);
		//init_rep_coder( &s_estrep, &s_dstrep, ao);
	}
#endif

	memset(speech_buf, 0, SPEECH_BUF_SZ);
	memset(chan_buf, 0, CHAN_BUF_SZ);

	crypt_sent_frame_count = 0;
	crypt_recv_frame_count = 0;
	stop_processing = 0;
}
