/**
 * ============================================================================
 * @file	main.c
 * @brief	This file contains implementation of the main processing threads
 * 			that drive the ARM application.
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */


/* ---------------------------- DSP/BIOS Headers --------------------------- */
#include <std.h>
#include <log.h>
#include <sys.h>
#include <sio.h>
#include <tsk.h>
#include <gio.h>
#include <pool.h>

/* ---------------------------- DSP/BIOS Link ------------------------------ */
#include <dsplink.h>
#include <failure.h>

#if defined (CHNL_PCPY_LINK)
#include <buf_pool.h>
#endif /* if defined (CHNL_PCPY_LINK) */

/* ---------------------------- Application specific ----------------------- */
#include "common.h"
#include "ambe_intf.h"
#include "tskDspAmbeIn.h"
#include "tskDspAmbeOut.h"
#include "dsp_ambe_config.h"
#include "dprintf.h"
#include "log_dsp.h"


/**
 * ============================================================================
 * @const	FILEID
 * @desc	FILEID is used by SET_FAILURE_REASON macro.
 * ============================================================================
 */
#define FILEID              FID_APP_C


/**
 * ============================================================================
 * @name	xferBufSize
 * @desc	Size of the buffer size for TSK based dsp_ambeback.
 * ============================================================================
 */
Uint32 xferBufSize ;


/**
 * ============================================================================
 * @name	trace
 * @desc	trace LOG_Obj used to do LOG_printf
 * ============================================================================
 */
extern LOG_Obj trace;

/**
 * ============================================================================
 * @func	atoi
 * @desc	Converts character string to integer value.
 * @arg	str
 *              Pointer to character string.
 * @ret	<integer value>
 *              Converted integer value.
 * @enter	None
 * @leave	None
 * @see	None
 * ============================================================================
 */
extern int atoi(const char *str);


/**
 * ----------------------------------------------------------------------------
 * @func	tskDspAmbeIn
 * @desc	Task for TSK based TSKDSPAMBE application.
 * @arg	None
 * @ret	None
 * @enter	None
 * @leave	None
 * @see	None
 * ----------------------------------------------------------------------------
 */
static Int tskDspAmbeIn () ;

/**
 * ----------------------------------------------------------------------------
 * @func	tskDspAmbeOut
 * @desc	Task for TSK based TSKDSPAMBE application.
 * @arg	None
 * @ret	None
 * @enter	None
 * @leave	None
 * @see	None
 * ----------------------------------------------------------------------------
 */
static Int tskDspAmbeOut () ;

/**
 * ----------------------------------------------------------------------------
 * @func	tskDspAmbeProcess
 * @desc	Task for TSK based TSKDSPAMBE application.
 * @arg	None
 * @ret	None
 * @enter	None
 * @leave	None
 * @see	None
 * ----------------------------------------------------------------------------
 */
static Int tskDspAmbeProcess () ;

short inputTskRunning = 0;
short outputTskRunning = 0;
char stop_processing = 0;

/**
 * ============================================================================
 * @func	main
 * @desc	Entry function.
 * @modif	None
 * ============================================================================
 */
Void main(Int argc, Char *argv[])
{
	/* TSK based dsp_ambeback */
	TSK_Handle tskDspAmbeInTask;
	TSK_Handle tskDspAmbeOutTask;
	TSK_Handle tskDspAmbeProcessTask;
	TSK_Attrs attrs;
	Int status ;

	/* Initialize DSP/BIOS LINK. */
	DSPLINK_init () ;

	status = dprintf_init_dsp();
	if (status != SYS_OK)
		return;

	/* Transfer size for dsp_ambeback given by GPP side */
	xferBufSize = DSPLINK_ALIGN ((CHNL_DATA_BUFFER_SZ / DSP_MAUSIZE),
			DSPLINK_BUF_ALIGN) ;

	// Dsp Receive Task
	attrs = TSK_ATTRS;
	attrs.name = "Dsp Receive Task";
	attrs.priority = TSK_MAXPRI-1;
	attrs.stacksize = 0x2000;
	tskDspAmbeInTask = TSK_create(tskDspAmbeIn, &attrs);
	if (tskDspAmbeInTask == NULL) {
		ERROR("Create tskDspAmbeIn: Failed.\n");
		return;
	}

	// Dsp Send Task
	attrs = TSK_ATTRS;
	attrs.name = "Dsp Send Task";
	attrs.priority = TSK_MAXPRI-1;
	attrs.stacksize = 0x2000;
	tskDspAmbeOutTask = TSK_create(tskDspAmbeOut, &attrs);
	if (tskDspAmbeOutTask == NULL) {
		ERROR("Create tskDspAmbeOut: Failed.\n");
		return;
	}

	// Dsp Process Task
	attrs = TSK_ATTRS;
	attrs.name = "Dsp Process Task";
	attrs.priority = TSK_MAXPRI-1;
	attrs.stacksize = 0x2000;
	tskDspAmbeProcessTask = TSK_create(tskDspAmbeProcess, &attrs);
	if (tskDspAmbeProcessTask == NULL) {
		ERROR("Create tskDspAmbeProcess: Failed.\n");
		return;
	}

	return;
}


/**
 * ----------------------------------------------------------------------------
 * @func	tskDspAmbeIn
 * @desc	Input Task for TSK based TSKDSPAMBE application.
 * @modif	None
 * ----------------------------------------------------------------------------
 */
static Int tskDspAmbeIn()
{
	Int status = SYS_OK;
	TSKDSPAMBE_IN_TransferInfo *info;

	DBG("Started tskDspAmbeIn\n");

	/* Create Phase */
	status = TSKDSPAMBE_IN_create (&info);

	/* Execute Phase */
	if (status == SYS_OK) {
		status = TSKDSPAMBE_IN_execute (info);
		if (status != SYS_OK) {
			SET_FAILURE_REASON(status);
			ERROR1("TSKDSPAMBE_IN_execute failed [status = 0x%x]", status);
		}
	}

	/* Delete Phase */
	status = TSKDSPAMBE_IN_delete (info);
	if (status != SYS_OK) {
		SET_FAILURE_REASON(status);
		ERROR1("TSKDSPAMBE_IN_delete failed [status = 0x%x]", status);
	}

	DBG("Ending tskDspAmbeIn\n");

	return status ;
}

/**
 * ----------------------------------------------------------------------------
 * @func	tskDspAmbeOut
 * @desc	Output Task for TSK based TSKDSPAMBE application.
 * @modif	None
 * ----------------------------------------------------------------------------
 */
static Int tskDspAmbeOut()
{
	Int status = SYS_OK;
	TSKDSPAMBE_OUT_TransferInfo *info;

	DBG("Starting tskDspAmbeOut\n");

	/* Create Phase */
	status = TSKDSPAMBE_OUT_create (&info);

	/* Execute Phase */
	if (status == SYS_OK) {
		status = TSKDSPAMBE_OUT_execute (info);
		if (status != SYS_OK) {
			SET_FAILURE_REASON(status);
			ERROR1("TSKDSPAMBE_OUT_execute failed [status = 0x%x]", status);
		}
	}

	/* Delete Phase */
	status = TSKDSPAMBE_OUT_delete (info);
	if (status != SYS_OK) {
		SET_FAILURE_REASON(status);
		ERROR1("TSKDSPAMBE_OUT_delete failed [status = 0x%x]", status);
	}

	DBG("Ending tskDspAmbeOut\n");

	return status ;
}

/**
 * ----------------------------------------------------------------------------
 * @func	tskDspAmbeProcess
 * @desc	Process Task for TSK based TSKDSPAMBE application.
 * @modif	None
 * ----------------------------------------------------------------------------
 */
static Int tskDspAmbeProcess()
{
	Int status = SYS_OK;
	AMBE_OPTIONS ao;
	int no_decode;
	int finished;
	int ret;

	DBG("Starting tskDspAmbeProcess\n");

	DBG("Waiting for input and output tasks to get started\n");
	while(!inputTskRunning && !outputTskRunning)
		TSK_yield();

	DBG("All tasks started\n");
	while (status == SYS_OK) {
		DBG("Waiting form Control Message\n");
		ret = TSKDSPAMBE_IN_getBuffer((char *)&ao, sizeof(AMBE_OPTIONS));
		if (ret != sizeof(AMBE_OPTIONS))
			continue;

		DBG2("ao.encoder = %d, ao.decoder = %d\n", ao.run_encoder, ao.run_decoder);
		init_coders(&ao);

		switch (ao.arm_mode) {
		case ARM_NORMAL_MODE:
			// In AMBE_NORMAL_MODE the encoder and decoder should never
			// be run back to back. This is because of the half duplex
			// nature of the radio implementation which only allows one
			// operation type (encode or decode) at any given time
			if(ao.run_encoder && ao.run_decoder) {
				INFO("In ARM_NORMAL_MODE, both run_encoder and run_decoder should not be set.\n");
				ao.run_decoder = 0;
			}

			finished = 0;

			TSKDSPAMBE_IN_setArmMode(ARM_NORMAL_MODE_RUN);
			while (TSKDSPAMBE_IN_getArmMode() == ARM_NORMAL_MODE_RUN && !finished) {
				if (ao.run_encoder)
					finished = process_encode_request(&ao);
				else
					finished = process_decode_request(&ao);
			}
			break;
		case ARM_TEST_MODE:
			// In AMBE_TEST_MODE the encoder and decoder can be run back
			// to back allowing the test program to compare synthesized
			// speech to a previously save synthesized speech file

			no_decode = 0;

			TSKDSPAMBE_IN_setArmMode(ARM_TEST_MODE_RUN);
			while (TSKDSPAMBE_IN_getArmMode() == ARM_TEST_MODE_RUN && !no_decode) {
				if (ao.run_encoder)
					no_decode = process_encode_request(&ao);

				if (ao.run_decoder && !no_decode)
					no_decode = process_decode_request(&ao);

				DBG("*\n");
			}
			break;
		default:
			INFO("ARM_UNKOWN_MODE");
			break;
		}

		if (ao.encryption) {
			// wrap up encryption
			stop_processing = 1;
			
			// Note may not work for TEST_MODE since we can run both encoder and
			// decoder in one loop (-cmp)
			if (ao.run_encoder)
				process_encode_request(&ao);
			else
				process_decode_request(&ao);

			// Send CRYPT_STOP control packet
			TSKDSPAMBE_OUT_sendCryptBuffer((char *)&ao, 1, CRYPT_END);
		}

		// Send ACK_STOP Control packet
		ao.run_encoder = 0;
		ao.run_decoder = 0;
		TSKDSPAMBE_OUT_sendBuffer((char *)&ao, sizeof(AMBE_OPTIONS), 1);
	}

	DBG("Ending tskDspAmbeProcess\n");

	return status ;
}
