/**
 * ============================================================================
 * @file	tskDspAmbeIn.h
 * @brief	This file contains function definitions and parameters for the
 * 			tskDspAmbeIn.c
 * @note	none.
 *
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */


#if !defined (TSKDSPAMBE_IN_)
#define TSKDSPAMBE_IN_

#include <sio.h>

#if defined (__cplusplus)
extern "C" {
#endif /* defined (__cplusplus) */


/** ============================================================================
 *  @const  MAX_BUFFERS
 *  @desc   Size of buffers array.
 *  ============================================================================
 */
#define MAX_BUFFERS            100

/** ============================================================================
 *  @const  TSK_NUM_BUFFERS
 *  @desc   Number of buffers to allocate and use for each channel.
 *  ============================================================================
 */
#define TSK_NUM_BUFFERS          2


/** ============================================================================
 *  @name   TSKDSPAMBE_IN_TransferInfo
 *  @desc   Structure used to keep various information needed by various phases
 *          of the application.
 *  @field  inputStream
 *              Handle to input stream.
 *  @field  outputStream
 *              Handle to output stream.
 *  @field  numTransfers
 *              Number of transfers to be done.
 *  @field  bufferSize
 *              Buffer size information.
 *  @field  receivedSize
 *              Actual length of the buffer that was received.
 *  @field  numBuffers
 *              Number of valid buffers in "buffers" array.
 *  @field  buffers
 *              Array of buffer pointers.
 *  ============================================================================
 */
typedef struct TSKDSPAMBE_IN_TransferInfo_tag {
    SIO_Handle inStream;
    SIO_Handle inCryptStream;
    Uint32     bufferSize;
    Uint16     receivedSize;
    Uint16     numBuffers;
    Uint16     numChnlBuffers;
    Char *     buffers [MAX_BUFFERS];
} TSKDSPAMBE_IN_TransferInfo;


/** ============================================================================
 *  @func   TSKDSPAMBE_IN_create
 *  @desc   Create phase function of TSKDSPAMBE_IN application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_create (TSKDSPAMBE_IN_TransferInfo ** transferInfo);

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_execute
 *  @desc   Excecute phase function of TSKDSPAMBE_IN application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_execute (TSKDSPAMBE_IN_TransferInfo * transferInfo);

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_getBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_getBuffer(Char * buf, Int len);

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_getCryptBuffer
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_getCryptBuffer(Char * buf, Int len);

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_setArmMode
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Void TSKDSPAMBE_IN_setArmMode(Int activity);

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_getArmMode
 *  @desc   Execute phase function for the TSKDSPAMBE_IN application. Application
 *          receives the data from the input channel and sends the same data
 *          back on output channel. Channel numbers can be configured through
 *          header file.
 *  @modif  None.
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_getArmMode();

/** ============================================================================
 *  @func   TSKDSPAMBE_IN_delete
 *  @desc   Delete phase function of TSKDSPAMBE_IN application.
 *  @arg    transferInfo
 *              Information for transfer.
 *  @ret    SYS_OK
 *              Successful operation.
 *          SYS_EBADIO
 *              Failure occured while doing IO.
 *  @enter  None
 *  @leave  None
 *  @see    None
 *  ============================================================================
 */
Int TSKDSPAMBE_IN_delete (TSKDSPAMBE_IN_TransferInfo * transferInfo);


#if defined (__cplusplus)
}
#endif /* defined (__cplusplus) */


#endif /* !defined (TSKDSPAMBE_IN_) */
