/**
 * ============================================================================
 * @file	dprintf.h
 * @brief	This file contains definition for real time print outs from DSP
 * 			called dprintf.
 * @note	none.
 *
 * @author	DVSI
 * @author	Cal(it)2
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 *
 * @verbatim
 * CAL(IT)2 CONTROL INFORMATION
 *
 * Copyright ©2009-2011, The Regents of the University of California
 * All Rights Reserved
 * See UCSD-license.h for more information
 *
 * Project Name:    APCO-P25 Linux Implementation
 * @endverbatim
 * ============================================================================
 */


/* ---------------------------- DSP/BIOS Headers --------------------------- */
#include <std.h>
#include <sys.h>
#include <sem.h>
#include <bcache.h>

/* ---------------------------- DSP/BIOS LINK Headers ---------------------- */
#include <dsplink.h>
#include <notify.h>

/* ---------------------------- Application specific ----------------------- */
#include "dprintf.h"

// Semaphore for synchronization
static SEM_Handle dprintf_sema; 
static SEM_Handle dprintf_semb;

// Cached aligned buffer for data transfer
#pragma DATA_ALIGN(printf_buffer, 128); 

static char printf_buffer[DPRINTF_BUFFERSZ]; 

/**
 * ----------------------------------------------------------------------------
 * @func	dprintf
 * @brief	outputs messages to ARM host processor.
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Void dprintf (const char * format, ...) {
	int n; 
	va_list args;

	SEM_pendBinary (dprintf_semb, SYS_FOREVER);
	
	va_start (args, format); 
	n = vsprintf (printf_buffer, format, args); 
	va_end (args); 

	if (n<=0) return; 

	// Writeback caches
	BCACHE_wb (printf_buffer, n, 1); 

	// Notify GPP
	NOTIFY_notify (ID_GPP, DPRINTF_IPS_ID, DPRINTF_IPS_EVENTNO, (Uint32) printf_buffer); 

	// Wait for GPP acknowledge
	SEM_pendBinary (dprintf_sema, SYS_FOREVER);

	SEM_postBinary (dprintf_semb);
} 

/**
 * ----------------------------------------------------------------------------
 * @func	dprintf_dsp_callback
 * @brief	callback function for the DSP BIOS NOTIFY mechanism.
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
static Void dprintf_dsp_callback (Uint32 eventNo, Ptr arg, Ptr info) {
	(Void) eventNo;
	(Void) arg;
	(Void) info;

  	// Acknowledge HOST is done with reading the print buffer:
  	SEM_postBinary (dprintf_sema); 
} 


/**
 * ----------------------------------------------------------------------------
 * @func	dprintf_init_dsp
 * @brief	Initialize the DSP side of dprintf.
 * @author	Anthony Nwokafor <aanwokafor@ucsd.edu>
 * ----------------------------------------------------------------------------
 */
Int dprintf_init_dsp () {
    Int status = SYS_OK;

	// Create semaphores
	dprintf_sema = SEM_create (0, NULL); 
	dprintf_semb = SEM_create (1, NULL); 

	// Register callback
  	status = NOTIFY_register (ID_GPP, DPRINTF_IPS_ID, DPRINTF_IPS_EVENTNO, 
		(FnNotifyCbck) dprintf_dsp_callback, (Uint32) 0); 

	return status;
}

