#ifndef BYTE_FIFO_H
#define BYTE_FIFO_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "byte_fifo.h"
#include "dprintf.h"

/* 
 * bf_alloc - create a byte fifo
 *
 * size: expected size in bytes of the fifo, the allocated
 * 		 fifo size might be larger
 *
 * return pointer to a byte_fifo structure, or NULL if failed
 */		 
struct byte_fifo *bf_alloc(int size)
{
	int nbytes;
	struct byte_fifo *bf;

	if (size == 0)
		size = BF_SIZE_DFT;
	if (size > BF_SIZE_MAX)
		size = BF_SIZE_MAX;

	//nbytes = size/8 + 1;
	nbytes = size + 1;
	bf = (struct byte_fifo *) malloc(sizeof(struct byte_fifo) + nbytes);
	if (bf == NULL) {
		dprintf("alloc byte fifo failed, size = %d\n",size);
		return NULL;
	}

	bf->byte_size = nbytes;
	bf->head_index = 0;
	bf->tail_index = bf->byte_size - 1;
	bf->bytes = (char *) (bf + 1);

	dprintf("allocated fifo, size = %d\n",size);
	return bf;
}

/* 
 * bf_free - destroy a byte fifo
 *
 * bf: the byte fifo to be freed
 */		 
void bf_free(struct byte_fifo *bf)
{
	free(bf);
}

/* 
 * bf_reset - reset a byte fifo
 *
 * bf: the byte fifo to be reset
 */		 
void bf_reset(struct byte_fifo *bf)
{
	bf->head_index = 0;
	bf->tail_index = bf->byte_size - 1;
}

#define incr_index_offset(idx, nbytes)	\
{					\
	(idx) = ((idx)+1) % (nbytes);	\
}

/*
 * bf_empty - return 1 if byte fifo is empty, otherwise
 * 			  return 0
 */
int bf_empty(struct byte_fifo *bf)
{
	int index;

	index = bf->tail_index;

	incr_index_offset(index, bf->byte_size);
	return index == bf->head_index;
}

/*
 * bf_full - return 1 if byte fifo is full, otherwise 
 * 			 return 0
 */
int bf_full(struct byte_fifo *bf)
{
	return bf->tail_index == bf->head_index;
}

/*
 * bf_size - return the number of bytes in a byte fifo
 */
int bf_size(struct byte_fifo *bf)
{
	int size;
	size = (bf->head_index - bf->tail_index) -1;
	return (size >= 0) ? size : size + bf->byte_size; 
}

/*
 * bf_space - return the number of bytes can be put in the byte fifo
 */
int bf_space(struct byte_fifo *bf)
{
	return (bf->byte_size - 1) - bf_size(bf);
}

void __bf_put(struct byte_fifo *bf, char b)
{
	bf->bytes[bf->head_index] = b;
	incr_index_offset(bf->head_index, bf->byte_size);
}

char __bf_get(struct byte_fifo *bf)
{
	incr_index_offset(bf->tail_index, bf->byte_size);
	return bf->bytes[bf->tail_index];
}

/* 
 * bf_put_n - put n bytes in a byte fifo
 *
 * bf:   the destination byte fifo
 * bytes: bytes array
 * n: 	 number of bytes to be put into fifo
 *
 * return the number of bytes put into the fifo
 */
int bf_put_n(struct byte_fifo *bf, char *bytes, int n)
{
	int i;

	for (i = 0; i < n && !bf_full(bf); i++)
		__bf_put(bf, bytes[i]);
	if (i != n)
		dprintf("n = %d, i = %d\n", n, i);
	return i;	
}

/* 
 * bf_get_n - get n bytes from a byte fifo
 *
 * bf:   the source byte fifo
 * bytes: bytes array
 * n: 	 number of bytes to get from fifo
 * 
 * return the number of bytes gotten from the fifo
 */
int bf_get_n(struct byte_fifo *bf, char *bytes, int n)
{
	int i;

	for (i = 0; i < n && !bf_empty(bf); i++)
		bytes[i] = __bf_get(bf);
	if (i != n)
		dprintf("n = %d, i = %d\n", n, i);
	return i;	
}

/*
 * bf_dump - dump the contents of a byte fifo
 */
void bf_dump(struct byte_fifo *bf)
{
	struct byte_fifo bf1;
	char b;
	int id;

	memcpy(&bf1, bf, sizeof(bf1));
	dprintf("size=%d, space=%d", bf_size(bf), bf_space(bf));
	dprintf(" bs=%d, hi=%d, ti=%d", bf->byte_size,
		bf->head_index, bf->tail_index);
	id = 0;
	while (!bf_empty(&bf1)) {
		if (id % 16 == 0)
			dprintf("\n%06d  ", id);
		else if (id % 8 ==0)
			dprintf("  ");

		b = __bf_get(&bf1);
		dprintf("% 4d", b);
		id++;
	}
	dprintf("\n");
}

#endif //BIT_FIFO_H
